/* vi: set sw=4 ts=4: */
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <getopt.h>
#include "busybox.h"
#ifdef BB_LOCALE_SUPPORT
#include <locale.h>
#endif

int been_there_done_that = 0; /* Also used in applets.c */
const char *applet_name;

#ifdef BB_FEATURE_INSTALLER
/* 
 * directory table
 *		this should be consistent w/ the enum, busybox.h::Location,
 *		or else...
 */
static char* install_dir[] = {
	"/",
	"/bin",
	"/sbin",
	"/usr/bin",
	"/usr/sbin",
};

/* abstract link() */
typedef int (*__link_f)(const char *, const char *);

/* 
 * Where in the filesystem is this busybox?
 * [return]
 *		malloc'd string w/ full pathname of busybox's location
 *		NULL on failure
 */
static char *busybox_fullpath()
{
	return xreadlink("/proc/self/exe");
}

/* create (sym)links for each applet */
static void install_links(const char *busybox, int use_symbolic_links)
{
	__link_f Link = link;

	char *fpc;
	int i;
	int rc;

	if (use_symbolic_links) 
		Link = symlink;

	for (i = 0; applets[i].name != NULL; i++) {
		fpc = concat_path_file(
			install_dir[applets[i].location], applets[i].name);
		rc = Link(busybox, fpc);
		if (rc!=0 && errno!=EEXIST) {
			perror_msg("%s", fpc);
		}
		free(fpc);
	}
}

#endif /* BB_FEATURE_INSTALLER */

/*
  Warning:  This function allocates a variable and passes it back to the calling
  function.  Be sure to free the variable there.

  Parses an input line and creates argc,argv information.
*/
static int parse_input_line(char *line, char ***args)
{
	int num_args = 0;
	int max_args = 10;
	char **argv = xmalloc(sizeof (char *) * max_args);
	char *pline;
	int new_delim = 1;

	pline = line;

	/*
	   Go through the line looking for tokens.  Tokens are characters
	   delineated by blanks or tabs.  This matches quotes so they are just
	   a single token.
	 */
	while (*pline) {
		switch (*pline) {
			/*  White space.  If it's at the beginning of the
			   line, then just remove it.  If we've started
			   parsing real words, then it is a token
			   delimeter.
			 */
		case ' ':
		case '\t':
			if (pline == line) {
				while (*pline && (*pline < 'a')) {
					pline++;
				}
			} else {
				*pline = '\0';
				new_delim = 1;
				pline++;
			}
			break;

			/*  Match quotes.  */
		case '\"':
			argv[num_args] = ++pline;
			num_args++;

			// Match quotes
			while (*pline && (*pline != '\"')) {
				pline++;
			}
			if (pline)			// get past second "
				*pline++ = '\0';

			break;

			/* Handle real non-whitespace characters */
		default:
			if (new_delim) {
				argv[num_args] = pline;
				num_args++;
			}
			new_delim = 0;
			pline++;
			break;
		}

		/*  Check if we have to bump up the number of arguments.  */
		if (num_args == max_args) {
			max_args += 10;
			argv = (char **) xrealloc(argv, sizeof (char *) * max_args);
		}

	}

	*args = argv;
	return num_args;
}

int flag_config_file;

int main(int argc, char **argv)
{
	const char *s;
	char *line = NULL;
	FILE *file = NULL;

	/*  Check to see if we have a command line: */
	/*   busybox <filename>                     */
	if (argc == 2) {
		/*  Were we called with "./busybox" ? */
		char *name = strrchr(argv[0], '/');
		if (name == NULL) {
			name = argv[0];
		} else {
			name++;
		}
		if (strncmp("busybox", name, 7) == 0) {
			/*  Is the second argument a filename we can open? */
			file = fopen(argv[argc - 1], "r");
			if (file != NULL) {
				flag_config_file = 1;
				line = get_line_from_file(file);
			}
		}
	}

	do {

		/*  If we're parsing a config file, we have to create
		   argc and argv entries for our applets.
		   This is a simple parser.  

		   It cleans up leading white space (tabs or spaces).
		   It matches " characters for echo.
		   It expects a typical max of 10 arguments, but will
		   realloc if it finds more arguments.
		 */
		if (flag_config_file) {
			int num_args = 0;
			char **args = NULL;

			/* Check for comments and blank lines */
			chomp(line);
			if ((*line == '#') || (*line == '\0')) {
				free(line);
				continue;
			}

			/*  parse_input_line allocates args, so be sure 
			   to free it.
			 */
			num_args = parse_input_line(line, &args);

			/*  Did we find anything on this line?  */
			if (num_args) {
				applet_name = args[0];

				if (applet_name[0] == '-')
					applet_name++;

				/*  Strip path from name.  */
				for (s = applet_name; *s != '\0';) {
					if (*s++ == '/')
						applet_name = s;
				}
				run_applet_by_name(applet_name, num_args, args);
			}
			free(args);
			free(line);

		}
		/*
		   Original code, any non-script invocation will go through here.
		 */
		else {
			applet_name = argv[0];

			if (applet_name[0] == '-')
				applet_name++;

			for (s = applet_name; *s != '\0';) {
				if (*s++ == '/')
					applet_name = s;
			}
			run_applet_by_name(applet_name, argc, argv);
			error_msg_and_die("if script, then run like:  . ./foo");
		}

#ifdef BB_LOCALE_SUPPORT
#ifdef BB_INIT
		if (getpid() != 1)		/* Do not set locale for `init' */
#endif
		{
			setlocale(LC_ALL, "");
		}
#endif

	} while (flag_config_file && (line = get_line_from_file(file)));

	if (file) {
		fclose(file);
	}
}


int busybox_main(int argc, char **argv)
{
	int col = 0, len, i;

#ifdef BB_FEATURE_INSTALLER	
	/* 
	 * This style of argument parsing doesn't scale well 
	 * in the event that busybox starts wanting more --options.
	 * If someone has a cleaner approach, by all means implement it.
	 */
	if (argc > 1 && (strcmp(argv[1], "--install") == 0)) {
		int use_symbolic_links = 0;
		int rc = 0;
		char *busybox;

		/* to use symlinks, or not to use symlinks... */
		if (argc > 2) {
			if ((strcmp(argv[2], "-s") == 0)) { 
				use_symbolic_links = 1; 
			}
		}

		/* link */
		busybox = busybox_fullpath();
		if (busybox) {
			install_links(busybox, use_symbolic_links);
			free(busybox);
		} else {
			rc = 1;
		}
		return rc;
	}
#endif /* BB_FEATURE_INSTALLER */

	argc--;

	/* If we've already been here once, exit now */
	if (been_there_done_that == 1 || argc < 1) {
		const struct BB_applet *a = applets;

		fprintf(stderr, "%s\n\n"
				"Usage: busybox [function] [arguments]...\n"
				"   or: [function] [arguments]...\n\n"
				"\tBusyBox is a multi-call binary that combines many common Unix\n"
				"\tutilities into a single executable.  Most people will create a\n"
				"\tlink to busybox for each function they wish to use, and BusyBox\n"
				"\twill act like whatever it was invoked as.\n" 
				"\nCurrently defined functions:\n", full_version);

		while (a->name != 0) {
			col +=
				fprintf(stderr, "%s%s", ((col == 0) ? "\t" : ", "),
						(a++)->name);
			if (col > 60 && a->name != 0) {
				fprintf(stderr, ",\n");
				col = 0;
			}
		}
		fprintf(stderr, "\n\n");
		exit(0);
	}

	/* Flag that we've been here already */
	been_there_done_that = 1;
	
	/* Move the command line down a notch */
	len = argv[argc] + strlen(argv[argc]) - argv[1];
	memmove(argv[0], argv[1], len);
	memset(argv[0] + len, 0, argv[1] - argv[0]);

	/* Fix up the argv pointers */
	len = argv[1] - argv[0];
	memmove(argv, argv + 1, sizeof(char *) * (argc + 1));
	for (i = 0; i < argc; i++)
		argv[i] -= len;

	return (main(argc, argv));
}

/*
Local Variables:
c-file-style: "linux"
c-basic-offset: 4
tab-width: 4
End:
*/
