/***************************************************************
 *             Copyright (C) 2003 by SerComm Corp.
 *                    All Rights Reserved.
 *
 *      Use of this software is restricted to the terms and
 *      conditions of SerComm's software license agreement.
 *
 *                        www.sercomm.com
 ****************************************************************/
/* orinoco_isa.c:
   Driver for Prism II devices which would be driven by ISA interface.
   It's a H/W specific driver. It may not meet general user's requrest.
   It's based on the original Prism II drivers: hermes & orinoco.

   Written 2002 by Vincent Kuo.

   Changelog:
   
 	Vincent Kuo		: V1.03 (October, 2002)
 			: We want to prevent the following situation:
 			: Generating 100% utilization traffic to our device for
 			: a long period. Original Code will die in 5 Sec. Now
 			: we could last almost one hour! :-) But still DIE :-(
			: Using the same method in DaviCom DM9000 driver:
			: Set an extra timer to check "dev->trans_start".
	Vincent Kuo		: V1.03 (September, 2002)
			: Remove the unecessary defines (previous version).
	Vincent Kuo		: V1.02 (August, 2002)
			: For Network Camera specific H/W.
			: Setup GIOs and External-IRQs (now move to kernel)
	Vincent Kuo		: V1.01 (July, 2002)
			: For Network Camera specific H/W.
			: ORINOCO_INIT_IOADDR 0x0400:0000 ~ 0x0400:07FF
			: ORINOCO_IOADDR      0x0400:0800 ~ 0x0400:0BFF
			: ORINOCO_IRQ         DSC21_IRQ_EXB0   (GIO0=EXT INT0)
			: DSC21 default EtherNET Card: 0230:0000 (CS8900_BASE)
	Vincent Kuo		: V1.00 (July, 2002)
			: ORINOCO_INIT_DATA "0x45" --> "0x01"
			: request_irq( ..SA_INTERRUPT..) --> 0
			  This difference doesn't matter. SA_INTERRUPT is for 
			  fast interrupt. IRQ won't disable while ENTER.
			: kfree_s( ...) --> kfree( ...)
			: add "orinoco_isa_reset" function
	
	Vincent Kuo		: init version (June, 2002)

 */

static const char isa_version[] = "orinoco_isa.c:v1.00 06/07/2002  Vincent Kuo (vincent_kuo@sercomm.com)";

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>	/* request/free_irq */
#include <asm/irq.h>		/* enable/disable_irq */
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/netdevice.h>
#include <linux/ioport.h>	/* check/request/release_region */
#include <asm/io.h>		/* inb/outb */
#include <asm/system.h>
#include <linux/wireless.h>
#include <linux/timer.h>	/* add_timer */

#include "porting.h"		/* vincent: add */
#include "hermes.h"
#include "orinoco.h"

#ifdef CONFIG_ARCH_DSC21
#include <asm/arch/irqs.h>	/* DSC21_IRQ_EXB0 */
#define WLAN_INIT_IOBASE        0x04000800       /* 0400:0000~07FF + 03E0 */
#define WLAN_IOBASE             0x04000800       /* 0400:0800~0BFF */
#define ORINOCO_IRQ		DSC21_IRQ_EXB0 /* External IRQ=GIO0 */
#define ORINOCO_IOADDR		WLAN_IOBASE
#define ORINOCO_IOEXTENT	0x0040         /* 0BFF-07FF=0400 */
#define ORINOCO_INIT_IOADDR	WLAN_INIT_IOBASE + 0x000003E0
#define ORINOCO_INIT_IOEXTENT   0x0010         /* 07FF-0000=07FF */
#else
#define ORINOCO_IRQ		5
#define ORINOCO_IOADDR		0x330   /* PCMCIA: 0x0100~0x013F (use 0x40) */
#define ORINOCO_IOEXTENT	0x0040  /* Intersil (P.33): 00~3E (64) */

#define ORINOCO_INIT_IOADDR	0x3E0 /* MUST set this I/O to init the card */
#define ORINOCO_INIT_IOEXTENT   16     /* 1 byte */
#endif /* CONFIG_ARCH_DSC21 */

#define ORINOCO_INIT_DATA	0x01  /* bit-1 HIGH: Change to I/O mode */
#define ORINOCO_RESET_DATA	0x80  /* bit-7 HIGH */

#ifdef CONFIG_ARCH_DSC21
#include <linux/gio.h>		/* All Stuff Here */
#endif /* CONFIG_ARCH_DSC21 */

/*
 * This is the root of the linked list of the ISA drivers.
 * It is use to verify that we don't reuse the same base address
 * for two differents drivers and to make the cleanup when
 * removing the module. (We only support one card!!)
 */
static struct dldwd_priv * orinoco_list = (struct dldwd_priv *) NULL;

static int orinoco_isa_config(struct device *dev);
static int orinoco_isa_open(struct device *dev);
static int orinoco_isa_close(struct device *dev);
static void orinoco_isa_interrupt(int irq, void *dev_id, struct pt_regs *regs);

				/* ---- LED Functions (Start) ---- */
#define WLAN_LED_INTERVAL 5	/* HZ = 100 */
#define WLAN_LED_ON  0
#define WLAN_LED_OFF 1
static struct timer_list orinoco_timer1; /* LED */
static u8 timer_status = WLAN_LED_ON; /* Normal=High (Active=ON) */

static void orinoco_timer1_out(unsigned long ptr)
{
	GIOSetPortBitSet(bit31_16, (1<<GIO_BIT_30)); /* LED_ON */
	timer_status = WLAN_LED_ON; /* Flag Low */
}

static void orinoco_timer1_set(void)
{
	if (timer_status != WLAN_LED_OFF) {
		GIOSetPortBitClear(bit31_16, (1<<GIO_BIT_30)); /* LED_OFF */
		init_timer(&orinoco_timer1);
		orinoco_timer1.function = orinoco_timer1_out;
		orinoco_timer1.data = 0;
		orinoco_timer1.expires = jiffies + WLAN_LED_INTERVAL;
		add_timer(&orinoco_timer1);
		timer_status = WLAN_LED_OFF; /* Flag High */
	} /* End-If: preventing re-entrance (need blink) */
}
				/* ---- LED Functions (Stop)  ---- */

				/* ---- Tx-Timeout Functions (Start) ---- */
#define WLAN_TXCHECK_INTERVAL   (4*HZ)+50	/* HZ = 100 */
#define WLAN_TXCHECK_TIMEOUT    (4*HZ)		/* original tx_timout=4*HZ */
static struct timer_list orinoco_timer2; 	/* TX-Timeout */

/* vincent: 2002/10/22
 * Because the continueous hacking our device (100% utilization) will cause
 * the Linux network system unusable (I don't know why.).
 * Most of the time, we will remove the driver and then insert it again.
 * So I will do a H/W reset and redo the initial step.
 */
void orinoco_restart_init(struct device *dev)
{
	dldwd_priv_t *priv = dev->priv;

	/* 1. H/W RF Reset: */
				/* copy from orinoco_isa_hw_reset */
	GIOSetPortBitSet(bit31_16, (1<<GIO_BIT_24));
	udelay(50);		/* delay 50 uSec */
	GIOSetPortBitClear(bit31_16, (1<<GIO_BIT_24));
	dbugprintk("Reset-RF Complete!\n");
	udelay(30000);		/* delay 30 mSec */
			/* InterSil: RESET */
	GIOSetPortBitClear(bit15_0, (1<<GIO_BIT_11));
	udelay(5);		/* 5 uSec */
	outb_p(ORINOCO_RESET_DATA, ORINOCO_INIT_IOADDR);	/* 0x80 */
	udelay(30);
	GIOSetPortBitSet(bit15_0, (1<<GIO_BIT_11));
			/* InterSil: IO Mode */
	GIOSetPortBitClear(bit15_0, (1<<GIO_BIT_11));
	udelay(5);		/* 5 uSec */
	outb_p(ORINOCO_INIT_DATA, ORINOCO_INIT_IOADDR);	/* 0x01 */
	GIOSetPortBitSet(bit15_0, (1<<GIO_BIT_11));

	/* 2. S/W Registers Reset: */
				/* FIXME: Is this enough? */
	dldwd_reset(priv);
}

static void orinoco_timer2_set(struct device *dev); /* timer function */

static void orinoco_timer2_out(unsigned long ptr)
{
	struct device *dev = (struct device *)ptr;

	if (dev->trans_start && ((jiffies - dev->trans_start) > WLAN_TXCHECK_TIMEOUT)) {
		disable_irq(dev->irq);	/* deactive it */
		netif_stop_queue(dev);
		dev->trans_start = 0; /* re-init */
#if 0				/* vincent: H/W Reset or S/W reset */
		orinoco_restart_init(dev);
		netif_wake_queue(dev);
#else
		dldwd_tx_timeout(dev);
#endif 				/* vincent: Which way is better??? */
		enable_irq(dev->irq);	/* active it */
	}
	orinoco_timer2_set(dev);
}

static void orinoco_timer2_set(struct device *dev)
{
	init_timer(&orinoco_timer2);
	orinoco_timer2.function = orinoco_timer2_out;
	orinoco_timer2.data = (unsigned long)dev;
	orinoco_timer2.expires = jiffies + WLAN_TXCHECK_INTERVAL;
	add_timer(&orinoco_timer2);
}
				/* ---- Tx-Timeout Functions (Stop)  ---- */

static int orinoco_isa_open(struct device *dev)
{
	dldwd_priv_t *priv = orinoco_list;
	int err;

	funcprintk("orinoco_isa_open\n");
	dev->trans_start = 0;	/* init timeout flag */
	netif_device_attach(dev);
	MOD_INC_USE_COUNT;
	err = dldwd_reset(priv);
	if (err) {
		myprintk(KERN_ERR "%s: dldwd_reset failed in orinoco_isa_open()", dev->name);
		orinoco_isa_close(dev);
	} else {
		netif_mark_up(dev);
		netif_start_queue(dev);
		netif_carrier_on(dev);
		orinoco_timer2_set(dev);	/* vincent: Check TX-Timeout */
	}

	return err;
}

static int orinoco_isa_close(struct device *dev)
{
	dldwd_priv_t *priv = orinoco_list;

	funcprintk("orinoco_isa_close\n");
	del_timer(&orinoco_timer2); /* vincent: stop checking Tx-Timer */
	netif_carrier_off(dev);
	netif_stop_queue(dev);
	netif_mark_down(dev);
	dldwd_shutdown(priv);
	MOD_DEC_USE_COUNT;
	return 0;
}

static void orinoco_isa_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	dldwd_priv_t *priv = orinoco_list;

	funcprintk("orinoco_isa_interrupt\n");
  	dldwd_interrupt(irq, priv, regs);
	orinoco_timer1_set();	/* Set LED Blink Once */
}

/*
 * As we know, we need to set 0x3e0 with bit7 high to initialize the card.
 * Is it necessary to restore the original value? (read/save/write/restore?)
 */
static int orinoco_isa_register(struct device *dev)
{
	funcprintk("orinoco_isa_register\n");
				/* allocate the H/W IO (ISA): 0x330 */
	if (check_region(ORINOCO_IOADDR, ORINOCO_IOEXTENT) < 0) {
		myprintk(KERN_ERR "orinoco_isa: unable to get I/O %d\n", ORINOCO_IOADDR);
		return -ENXIO;
	}
	request_region(ORINOCO_IOADDR, ORINOCO_IOEXTENT, "orinoco_isa");

#ifndef CONFIG_ARCH_DSC21	/* IOADDR == INIT_IOADDR */
				/* IO for "init card": 0x3E0 */
	if (check_region(ORINOCO_INIT_IOADDR, ORINOCO_INIT_IOEXTENT) < 0) {
		myprintk(KERN_ERR "orinoco_isa: unable to get I/O %d\n", ORINOCO_INIT_IOADDR);
		return -ENXIO;
	}
	request_region(ORINOCO_INIT_IOADDR, ORINOCO_INIT_IOEXTENT, "orinoco_isa");
#endif /* CONFIG_ARCH_DSC21 */

	return 0;
}

static void orinoco_isa_unregister(struct device *dev)
{
	funcprintk("orinoco_isa_unregister\n");

	free_irq(dev->irq, dev);
#ifndef CONFIG_ARCH_DSC21	/* IOADDR == INIT_IOADDR */
	release_region(ORINOCO_INIT_IOADDR, ORINOCO_INIT_IOEXTENT);
#endif /* CONFIG_ARCH_DSC21 */
	release_region(ORINOCO_IOADDR, ORINOCO_IOEXTENT);
}

#ifdef CONFIG_ARCH_DSC21
static int orinoco_isa_hw_reset(void)
{
	unsigned short cur_register;
	unsigned short out_register;

	/* For ARM system: */
				/* Check PCMCIA Card */
	GIOSetPortBitSet(bit31_16, (1<<GIO_BIT_17));
	cur_register = GIOGetPortBitSet(bit31_16);
	GIOSetPortBitSet(bit31_16, (1<<GIO_BIT_18));
	out_register = GIOGetPortBitSet(bit31_16);
	if ((cur_register&(1 << GIO_BIT_17))||(out_register&(1 << GIO_BIT_18))) {
		dbugprintk("PCMCIA Card has not inserted!\n");
		GIOSetPortBitClear(bit31_16, (1<<GIO_BIT_30)); /* LED_OFF */
	} else {
		dbugprintk("PCMCIA Card has inserted!\n");
				/* DO "Reset RF" */
		GIOSetPortBitSet(bit31_16, (1<<GIO_BIT_24));
		udelay(50);		/* delay 50 uSec */
		GIOSetPortBitClear(bit31_16, (1<<GIO_BIT_24));
		dbugprintk("Reset-RF Complete!\n");
		udelay(30000);		/* delay 30 mSec */
		GIOSetPortBitSet(bit31_16, (1<<GIO_BIT_30)); /* LED_ON */
	}

	return 0;
}
#endif /* CONFIG_ARCH_DSC21 */

static int orinoco_isa_reset(dldwd_priv_t *priv)
{
	/* do a soft reset: I/O=0x3E0, Val=0x80 */
#ifdef CONFIG_ARCH_DSC21
	GIOSetPortBitClear(bit15_0, (1<<GIO_BIT_11));
	udelay(5);		/* 5 uSec */
	outb_p(ORINOCO_RESET_DATA, ORINOCO_INIT_IOADDR);	/* 0x80 */
	udelay(30);
	GIOSetPortBitSet(bit15_0, (1<<GIO_BIT_11));
#else
	outb_p(ORINOCO_RESET_DATA, ORINOCO_INIT_IOADDR);	/* 0x80 */
#endif /* CONFIG_ARCH_DSC21 */
	return 0;
}

static int orinoco_isa_config(struct device *dev)
{
	struct dldwd_priv *priv=orinoco_list; /* redirect to the main struct */
	int retcode;

	funcprintk("orinoco_isa_config\n");
	if (orinoco_isa_register(dev) < 0) {
		return -ENXIO;
	}

	dev->base_addr = ORINOCO_IOADDR;
	dev->mem_start = ORINOCO_IOADDR;
	dev->mem_end = ORINOCO_IOEXTENT;
	dev->if_port = 0;
				/* setup card */
	retcode = dldwd_setup(priv);	/* XXX clean up if <0 */
	if (retcode < 0) {
		orinoco_isa_unregister(dev);
		myprintk(KERN_ERR "orinoco_isa: unable to setup Intersil card[%d]\n", retcode);
		return retcode;
	}
	dev->open = orinoco_isa_open;
	dev->stop = orinoco_isa_close;
	dev->init = dldwd_init;	/* init while register netdevice */
//	priv->card_reset_handler = NULL; /* We have no reset handler */
	priv->card_reset_handler = orinoco_isa_reset;
	myprintk(KERN_DEBUG "Detected Orinoco/Prism2 ISA device at IRQ:[%d], ioaddr:[0x%04x]\n", dev->irq, (unsigned int)dev->base_addr);

				/* ---- init the card ---- */
	/* toggle to IO mode (from memory mode): I/O=0x3E0, Val=0x01 */
#ifdef CONFIG_ARCH_DSC21
/* In our specific H/W, the ARM seems to protect the CIS address: 0400:0000.
 * 0400:0000. We don't have a chance to enter our signal. So we change the CIS
 * address to the same as IO address: 0400:0800. But we need to tell the 
 * system that we are doing INIT, not IO command.
 * We use GIO-11: Normal HIGH. Set GIO-11 LOW, it means that the following IO 
 * commands are for INIT, then set GIO-11 HIGH (back to IO command).
 */
	GIOSetPortBitClear(bit15_0, (1<<GIO_BIT_11));
	udelay(5);		/* 5 uSec */
	outb_p(ORINOCO_INIT_DATA, ORINOCO_INIT_IOADDR);	/* 0x01 */
	GIOSetPortBitSet(bit15_0, (1<<GIO_BIT_11));
#else
	outb_p(ORINOCO_INIT_DATA, ORINOCO_INIT_IOADDR);	/* 0x01 */
#endif /* CONFIG_ARCH_DSC21 */
	dbugprintk(KERN_DEBUG "Init Orinoco/Prism2 ISA device on ioaddr:[0x%04x]\n", ORINOCO_INIT_IOADDR);

				/* ---- starting the driver code ---- */
	hermes_struct_init(&(priv->hw), dev->base_addr);	/* init */
	dev->irq = ORINOCO_IRQ;
	dev->base_addr = ORINOCO_IOADDR;
	dldwd_init(dev);	/* run init NOW!! */

	return 0;
}

#ifdef MODULE
#define init_orinoco_isa init_module
#define exit_orinoco_isa cleanup_module
#endif

#define NAMELEN 	8	/* #of chars for storing dev->name */
static char namelist[NAMELEN] = { 0, };

int init_orinoco_isa(void)
{
	dldwd_priv_t *priv;
	struct device *dev;

	myprintk(KERN_DEBUG "%s\n", isa_version);
	funcprintk("init_orinoco_isa\n");

	/* Initialize device structures */
	priv = kmalloc(sizeof(struct dldwd_priv), GFP_KERNEL);
	if(priv == NULL) {
		return -ENOMEM;
	}
	memset(priv, 0x00, sizeof(struct dldwd_priv));
	orinoco_list = priv;
	dev = &priv->ndev;	/* link both structure together */
	dev->priv = priv;	/* ??? */
	dev->name = namelist;
	dev->irq  = ORINOCO_IRQ;
	dev->base_addr = ORINOCO_IOADDR;
	dev->init = &orinoco_isa_config;

				/* ---- reset the card ---- */
#ifdef CONFIG_ARCH_DSC21
	if (orinoco_isa_hw_reset() != 0) {	/* do a REAL H/W reset */
		kfree(priv);	/* free memory */
		myprintk(KERN_ERR "orinoco_isa: RF-Card init failed\n");
		return -ENODEV;
	}
	GIOSetPortBitClear(bit15_0, (1<<GIO_BIT_11));
	udelay(5);		/* 5 uSec */
	outb_p(ORINOCO_RESET_DATA, ORINOCO_INIT_IOADDR);	/* 0x80 */
	udelay(30);
	GIOSetPortBitSet(bit15_0, (1<<GIO_BIT_11));
#else
/* For our specific H/W, we don't need to do the soft reset. The H/W reset is 
 * enough (real RESET). But in PC system, we can't do a H/W reset. So that's 
 * why we need the soft reset (for x86). */
	/* do a soft reset: I/O=0x3E0, Val=0x80 */
	outb_p(ORINOCO_RESET_DATA, ORINOCO_INIT_IOADDR);	/* 0x80 */
#endif /* CONFIG_ARCH_DSC21 */

	/* Try to create the device */
	if (register_netdev(dev) != 0) {
	      /* DeAllocate everything */
		kfree(priv);	/* free memory */
		myprintk(KERN_ERR "orinoco_isa: register_netdev() failed\n");
		return -ENODEV;
	} else {
	/* SA_SHIRQ is for PCI or PCMCIA, etc. ISA can't share the IRQ: 0 */
//		request_irq(dev->irq, orinoco_isa_interrupt, SA_INTERRUPT, dev->name, dev);
#ifdef CONFIG_ARCH_DSC21	/* uCcs8900.c use IRQ_FLG_STD(=SA_INTERRUPT) */
		request_irq(dev->irq, orinoco_isa_interrupt, IRQ_FLG_STD, dev->name, dev);
#else
		request_irq(dev->irq, orinoco_isa_interrupt, 0, dev->name, dev);
#endif /* CONFIG_ARCH_DSC21 */
		enable_irq(dev->irq);	/* active it */

				/* priv will allocate in config function */
		priv = dev->priv;	/* link both structure together */

		/* If at least one device OK, we do not fail */
		if (dldwd_proc_dev_init(priv) != 0) {
			myprintk(KERN_ERR "%s: Failed to create /proc node\n", dev->name);
			orinoco_isa_unregister(dev);
			unregister_netdev(dev);
			kfree(priv); /* free memory */
			return -EIO;
		}
	}
	/* Note to myself : this replace MOD_INC_USE_COUNT/MOD_DEC_USE_COUNT */
	SET_MODULE_OWNER(dev);
	priv->hw_ready = 1;

	return 0;
}

void exit_orinoco_isa(void)
{
	struct dldwd_priv *priv = orinoco_list;
	struct device *dev;

	funcprintk("exit_orinoco_isa\n");
	dev = &priv->ndev;
	dldwd_proc_dev_cleanup(priv);
	unregister_netdev(dev);
	orinoco_isa_unregister(dev);  /* free H/W */
	kfree(priv); /* free memory */
	orinoco_list = (struct dldwd_priv *) NULL; /* free!! */
}
