/*
 *  linux/arch/arm/kernel/time.c
 *
 *  Copyright (C) 1991, 1992, 1995  Linus Torvalds
 *  Modifications for ARM (C) 1994, 1995, 1996,1997 Russell King
 *
 * This file contains the ARM-specific time handling details:
 * reading the RTC at bootup, etc...
 *
 * 1994-07-02  Alan Modra
 *             fixed set_rtc_mmss, fixed time.year for >= 2000, new mktime
 * 1997-09-10  Updated NTP code according to technical memorandum Jan '96
 *             "A Kernel Model for Precision Timekeeping" by Dave Mills
 */
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/param.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/interrupt.h>

#include <asm/segment.h>
#include <asm/io.h>
#include <asm/irq.h>

#include <linux/timex.h>
#include <asm/hardware.h>

#ifndef BCD_TO_BIN
#define BCD_TO_BIN(val) ((val)=((val)&15) + ((val)>>4)*10)
#endif

#ifndef BIN_TO_BCD
#define BIN_TO_BCD(val) ((val)=(((val)/10)<<4) + (val)%10)
#endif

/* Converts Gregorian date to seconds since 1970-01-01 00:00:00.
 * Assumes input in normal date format, i.e. 1980-12-31 23:59:59
 * => year=1980, mon=12, day=31, hour=23, min=59, sec=59.
 *
 * [For the Julian calendar (which was used in Russia before 1917,
 * Britain & colonies before 1752, anywhere else before 1582,
 * and is still in use by some communities) leave out the
 * -year/100+year/400 terms, and add 10.]
 *
 * This algorithm was first published by Gauss (I think).
 *
 * WARNING: this function will overflow on 2106-02-07 06:28:16 on
 * machines were long is 32-bit! (However, as time_t is signed, we
 * will already get problems at other places on 2038-01-19 03:14:08)
 */
static inline unsigned long mktime(unsigned int year, unsigned int mon,
	unsigned int day, unsigned int hour,
	unsigned int min, unsigned int sec)
{
	if (0 >= (int) (mon -= 2)) {	/* 1..12 -> 11,12,1..10 */
		mon += 12;	/* Puts Feb last since it has leap day */
		year -= 1;
	}
	return (((
		    (unsigned long)(year/4 - year/100 + year/400 + 367*mon/12 + day) +
		      year*365 - 719499
		    )*24 + hour /* now have hours */
		   )*60 + min /* now have minutes */
		  )*60 + sec; /* finally seconds */
}

#include <asm/arch/time.h>

static unsigned long do_gettimeoffset(void)
{
	
	return gettimeoffset ();
}

#if 0
void do_gettimeofday(struct timeval *tv)
{
	unsigned long flags;
	
	save_flags_cli (flags);
	*tv = xtime;
	
	tv->tv_usec += do_gettimeoffset();
	if (tv->tv_usec >= 1000000) {
		tv->tv_usec -= 1000000;
		tv->tv_sec++;
	}
	
	restore_flags(flags);
}
#else //bmb
/* This has to be a little bit fancy in order to deal with the fact
 * that xtime and the offset come from different sources and cannot
 * be read atomically.  As a result, they are not always mutually
 * consistent when they are read.  In order to insure correctness,
 * the values are read in a loop which doesn't exit until it
 * reads the same value for xtime (which is updated by the timer
 * interrupt) twice in a row.  This ensures that the offset did not
 * roll-over and signal a timer interrupt during the initial
 * sequence of reading the two values.
 */
void do_gettimeofday(struct timeval *tv)
{
    unsigned long flags, offset;
    struct timeval tv2;
       	
 // loop until we get a mutually consistent reading of xtime and the offset
    while ( 1 ) {
       	
     // make a copy of xtime and read the offset
        save_flags_cli(flags);
        offset = do_gettimeoffset();
        *tv = xtime; // note: must be read with interrupts disabled to make it atomic
        restore_flags(flags);

     // make a second copy of xtime in order to check for interrupts
        save_flags_cli(flags);
        tv2 = xtime;
        restore_flags(flags);

        if ( tv->tv_usec==tv2.tv_usec && tv->tv_sec==tv2.tv_sec ) break; // we didn't detect an interrupt
          	
    } // while
       	
    tv->tv_usec += offset;
       	
    if (tv->tv_usec >= 1000000) {
        tv->tv_usec -= 1000000;
        tv->tv_sec++;
    }

}
#endif

void do_settimeofday(struct timeval *tv)
{
	cli ();
	/* This is revolting. We need to set the xtime.tv_usec
	 * correctly. However, the value in this location is
	 * is value at the last tick.
	 * Discover what correction gettimeofday
	 * would have done, and then undo it!
	 */
	tv->tv_usec -= do_gettimeoffset();

	if (tv->tv_usec < 0) {
		tv->tv_usec += 1000000;
		tv->tv_sec--;
	}

	xtime = *tv;
	time_adjust = 0;		/* stop active adjtime() */
	time_status |= STA_UNSYNC;
	time_state = TIME_ERROR;	/* p. 24, (a) */
	time_maxerror = NTP_PHASE_LIMIT;
	time_esterror = NTP_PHASE_LIMIT;
	sti ();
}

/*
 * timer_interrupt() needs to keep up the real-time clock,
 * as well as call the "do_timer()" routine every clocktick.
 */
static void timer_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
        if (reset_timer ()) {
		do_timer(regs);
	}

	update_rtc ();
}

static struct irqaction irq_kernel_timer = { timer_interrupt, 0, 0, "system timer", NULL, NULL};

extern int setup_arm_irq(int, struct irqaction *);

#ifdef CONFIG_ARCH_ATMEL
unsigned long setup_timer (void)
{
	volatile struct atmel_timers* tt = (struct atmel_timers*) (TIMER_BASE);
	volatile struct atmel_timer_channel* tc = &tt->chans[KERNEL_TIMER].ch;
	unsigned long v;


	/* No SYNC */
	tt->bcr = 0;  

#if (KERNEL_TIMER==0)
	/* program no external (=internal) clock generation signal on XC0 */
	v = tt->bmr;
	v = (v & ~(3 << 0)) | (1 << 0);
	tt->bmr = v;
#elif (KERNEL_TIMER==1)
	/* program no external (=internal) clock generation signal on XC1 */
	v = tt->bmr;
	v = (v & ~(3 << 2)) | (1 << 2);
	tt->bmr = v;
#elif (KERNEL_TIMER==2)
	/* program no external (=internal) clock generation signal on XC2 */
	v = tt->bmr;
	v = (v & ~(3 << 4)) | (1 << 4);
	tt->bmr = v;
#else
#error Weird -- KERNEL_TIMER does not seem to be defined...
#endif
	tc->ccr = 2;  /* disable the channel */

	/* select ACLK/128 as inupt frequensy for TC1 and enable CPCTRG */
	tc->cmr = 3 | (1 << 14);

	tc->idr = ~0ul;  /* disable all interrupt */
	tc->rc = ((ARM_CLK/128)/HZ - 1);   /* load the count limit into the CR register */
	tc->ier = TC_CPCS;  /* enable CPCS interrupt */

	/* enable the channel */
	tc->ccr = TC_SWTRG|TC_CLKEN;

	return mktime(2000, 1, 1, 0, 0, 0);
}


void time_init(void)
{
#if (KERNEL_TIMER==0)
#   define KERNEL_TIMER_IRQ_NUM	IRQ_TC0
#elif (KERNEL_TIMER==1)
#   define KERNEL_TIMER_IRQ_NUM	IRQ_TC1
#elif (KERNEL_TIMER==2)
#   define KERNEL_TIMER_IRQ_NUM	IRQ_TC2
#else
#error Wierd -- KERNEL_TIMER is not defined or something....
#endif

	setup_arm_irq(KERNEL_TIMER_IRQ_NUM, &irq_kernel_timer);
	xtime.tv_sec = setup_timer();
	xtime.tv_usec = 0;

}
#endif   // ATMEL

void time_init(void)
{
     xtime.tv_sec = 0;
     xtime.tv_usec = 0;
     setup_timer();
     setup_arm_irq(IRQ_TIMER, &irq_kernel_timer);

}
