/*
 * File: api.c
 *
 * GPP-to-DSP communication interface implementation using the api hw bridge.
 *
 * Copyright (C) 2001 RidgeRun, Inc.
 * Author: Clint Medley <cmedley@ridgerun.com> 10/30/01
 *         Based on hpi.c by Gordon McNutt <gmcnutt@ridgerun.com>
 *         Based on dspc.c by Steve Johnson.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS  PROVIDED  ``AS  IS''  AND   ANY  EXPRESS  OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Please report all bugs/problems to the author or <support@dsplinux.net>
 *
 */
#include <linux/kernel.h>
#include <asm/errno.h>
#include <asm/arch/hardware.h>

#include <asm/arch/dspi.h>

#define DEBUG
#ifdef DEBUG
#define dbg(format, arg...) \
        printk(KERN_DEBUG __FUNCTION__ ": " format "\n" , ## arg)
#else
#define dbg(format, arg...) do {} while (0)
#endif
#define err(format, arg...) \
        printk(KERN_ERR __FUNCTION__ ": " format "\n" , ## arg)
#define info(format, arg...) \
        printk(KERN_INFO __FUNCTION__ ": " format "\n" , ## arg)
#define warn(format, arg...) \
        printk(KERN_WARNING __FUNCTION__ ": " format "\n" , ## arg)

#define API_BOOT_OFFSET 0xc000

static int __api_copy_to_dsp(__u8 *to, __u8 *from, int count);
static int __api_copy_from_dsp(__u8 *to, __u8 *from, int count);
static int __api_copy_unaligned_to_dsp(__u8 *to, __u8 *from, int count);
static int __api_copy_unaligned_from_dsp(__u8 *to, __u8 *from, int count);
void apitest(void);

inline static int __api_dsp_aligned_copy(__u16 *to, __u16 *from, int count)
{
   int ret = count - count % 2;
   while (count > 1)
      {
         *to++ = *from++;
         count -= 2;
         dbg("to = 0x%x, from = 0x%x", to[-1], from[-1]);
      }
   return ret;
}

inline static void __api_put_dsp_hibyte(__u8 *to, __u8 val)
{//always put as a 16bit value otherwise the hardware will duplicate the byte
   *((__u16*)&to[-1]) = ((__u16)val << 8) | *to;
}

inline static void __api_put_dsp_lobyte(__u8 *to, __u8 val)
{
   *((__u16*)to) = val | ((__u16)(*to) << 8);
}

inline static __u8 __api_get_dsp_hibyte(__u8 *from)
{
   return from[-1];
}

inline static __u8 __api_get_dsp_lobyte(__u8 *from)
{
   return from[1];
}

inline static void __api_put_dsp_byte(__u8 *to, __u8 val)
{
   if (((__u32)to) % 2)
      __api_put_dsp_hibyte(to, val);
   else
      __api_put_dsp_lobyte(to, val);
}
   
inline static __u8 __api_get_dsp_byte(__u8 *from)
{
   if (((__u32)from) % 2)
      return __api_get_dsp_hibyte(from);
   else
      return __api_get_dsp_lobyte(from);
}
   
static int __api_copy_unaligned_to_dsp(__u8 *to, __u8 *from, int count)
{
   int ret = 0;

   /*take care of any unalinged bytes, assume odd byte is high byte
     which is not quite kosher for byte copy, but makes things work
     out with the hw byte swapping of 16 bit aligned copies.*/
   /*copy byte by byte until both addresses are aligned if this is not possible
     then copy everything byte by byte.*/
   if (count && ((__u32)to) % 2)
      {//get aligned on the dsp side
         __api_put_dsp_hibyte(to++, *from++);
         ret++;
      }
   if (((__u32)from) % 2)
      {//we are aligned on the dsp side, but not the arm so we have to do all
         while (ret < count)
            {//read a byte
               __api_put_dsp_byte(to++, *from++);
               ret++;
            }
      }

   return ret;
}

static int __api_copy_unaligned_from_dsp(__u8 *to, __u8 *from, int count)
{
   int ret = 0;

   /*take care of any unalinged bytes, assume odd byte is high byte
     which is not quite kosher for byte copy, but makes things work
     out with the hw byte swapping of 16 bit aligned copies.*/
   /*copy byte by byte until both addresses are aligned if this is not possible
     then copy everything byte by byte.*/
   if (count && ((__u32)from) % 2)
      {//get aligned on the dsp side
         *to++ = from[-1];
         from++;
         ret++;
      }
   if (((__u32)to) % 2)
      {//we are aligned on the dsp side, but not the arm so we have to do all
         while (ret < count)
            {//Read bytes
               *to++ = __api_get_dsp_byte(from++);
               ret++;
            }
      }

   return ret;
}

static int __api_copy_to_dsp(__u8 *to, __u8 *from, int count)
{
   /*This copy routine takes advantage of the fact that the hardware swaps
     bytes for the write for 16bit values. It will be up to the dsp to swap
     bytes for true byte data (very rare I should think since it is only word
     addressable) or swap words if the data in question is 32bit. This means
     that byte data will appear in the dsp in swapped format. In the unlikely
     event that the dsp really wants byte data in byte order, it would simply
     take the low byte for the first byte and upper byte for the second,
     instead of taking the high byte first and low byte second as woul be
     normal in big endian byte addressable machines. But it is my assertion
     that for a 32 bit value they should be using get and put functions which
     will give the proper value anyway, not a copy. However to give constant
     behavior between the copy and the get/put functions the 1 byte get and
     put will have the same behavior as calling the copy with a lenght of 1.
     CLM 11/09/01
   */
   dbg("to=0x%x,from=0x%x,count=%d",(unsigned)to,(unsigned)from,
       (unsigned)count);

   if (count)
      {//remember we cannot write a byte without writing upper and lower
         int copied = __api_copy_unaligned_to_dsp(to, from, count);

         if (count != copied)
            copied += __api_dsp_aligned_copy((__u16*)&to[copied],
                                             (__u16*)&from[copied],
                                             (count - copied));

         if (count != copied)//catch any trailing bytes
            {/*assume even bytes are low bytes of words, not quite kosher for
               byte copy, but makes things work out with the hw byte swapping of
               16 bit aligned copies.*/
               __api_put_dsp_lobyte(&to[copied], from[copied]);
            }
      }
         
   return 0;
}

static int __api_copy_from_dsp(__u8 *to, __u8 *from, int count)
{
   /*Again this will not be a true byte copy of the memory (see to above), but
     an optimized one for the fact that the dsp memory is only word addressable
     from the dsp side, and the fact that words are byte swapped as they are
     read from the dsp side. This does not give true byte order from the big
     endian dsp memory that is being read, but does give valid word data (most
     commonly used by the dsp) without any byte swapping. When the dsp packs
     strings it puts the first byte in the big endian low byte and the next
     in the high byte (oppisit of normal byte order) of each word. This means
     strings will come out correct on the arm side without further swapping
     when read through the api interface. Real byte data to the arm will need
     to be swapped. As well as 32 bit values read with a copy will have to be
     word swapped as well. But it is my assertion that for a 32 bit value they
     should be using get and put functions which will give the proper value
     anyway, not a copy. However to give constant behavior between the copy and
     the get/put functions the 1 byte get and put will have the same behavior
     as calling the copy with a lenght of 1.
     CLM 11/09/01
   */
   dbg("to=0x%x,from=0x%x,count=%d",(unsigned)to,(unsigned)from,
       (unsigned)count); 
   if (count)
      {
         int copied = __api_copy_unaligned_from_dsp(to, from, count);

         if (count != copied)
            copied += __api_dsp_aligned_copy((__u16*)&to[copied],
                                             (__u16*)&from[copied],
                                             (count - copied));

         if (count != copied)//catch any trailing bytes
            {/*assume even bytes are low bytes of words, not quite kosher for
               byte copy, but makes things work out with the hw byte swapping of
               16 bit aligned copies.*/
               to[copied] = __api_get_dsp_lobyte(&from[copied]);
            }
      }
         
   return 0;
}

int dspi_load_code(void *to, const void *from, int count, int cpu)
{
   /* Revisit -- need a lock?
    */
   __u16 *__to, *__from;
   int i, ret = 0;
   
   /*
    * It would probably cause untold woe if anybody ever tries to copy to
    * an unaligned address in DSP RAM.
    */
   /*
    * Since this function will run on the ARM its arguments need to be in
    * the ARM portion of the shared memory map. Callers can use the
    * __dsp_to_arm() macro provided in the header to map DSP addresses to
    * their corresponding ARM addresses if they need to (the loader is an
    * example of this, since the addresses in the image are DSP 
    * addresses).
    */

   count = (count + 1) / 2;
   if (!dspi_in_dsp(to, cpu) || !dspi_in_dsp(to + count - 1, cpu))
      {//We are not in mapped space, check for boot area
         u32 tmp = (u32)to - API_BOOT_OFFSET;
         if (!dspi_in_dsp(tmp, cpu) || !dspi_in_dsp(tmp + count - 1, cpu))
            {
               warn("0x%x not aligned or in wrong address range", (__u32)to);
               ret = -EINVAL;
            }
         else
            to = (__u16*)tmp;
      }
   dbg("ret %d, to %X", ret, (unsigned)to);
   if (!ret)
      {
         __from = (__u16*)from;
         __to = (__u16*)dspi_dsp_to_arm(to, cpu);

         dbg("__from = 0x%x, __to = 0x%x", (int)__from, (int)__to);
         for (i = 0; i < count; i++)
            {
               *__to = *__from;
               if (*__to++ != *__from++)
                  {
                     ret = -EREMOTEIO;
                     break;
                  }
            }
      }

   return ret;
}

int dspi_dsp_memset(void *to, __u16 pattern, int count, int cpu)
{
   __u16 *__to;
   int i, ret = 0;

   if (!dspi_in_dsp(to, cpu))
      {
         warn("0x%x not aligned or in wrong address range", (__u32)to);
         ret = -EINVAL;
      }
   else
      {
         __to = (__u16*)dspi_dsp_to_arm(to, cpu);
         count = (count + 1) / 2;
         for (i = 0; i < count; i++)
            {
               *__to = pattern;
               if (*__to != pattern)
                  {
                     ret = -EREMOTEIO;
                     break;
                  }
               __to++;
            }

                
      }
   return ret;
}

int dspi_memcpy_from_dsp(void *to, const void *from, int count)
{
   int ret;
   /*
    * 'from' is a remapped dsp address. 'count' is the count in bytes.
    */
   dbg("from to = 0x%x, from = 0x%x, count = %d", (int)to, (int)from, count);
   if (!dspi_in_arm(from))
      {
         err("0x%x not in remapped DSP RAM", (__u32)from);
         ret = -EINVAL;
      }
   else
      ret = __api_copy_from_dsp((__u8*)to, (__u8*)from, count);

   return ret;
}

int dspi_memcpy_to_dsp(void *to, const void *from, int count)
{
   int ret;
   /*
    * 'to' is a remapped dsp address. 'count' is the count in bytes.
    */
   dbg("to to = 0x%x, from = 0x%x, count = %d", (int)to, (int)from, count);
   if (!dspi_in_arm(to))
      {
         err("0x%x not in remapped DSP RAM", (__u32)to);
         ret = -EINVAL;
      }
   else
      ret = __api_copy_to_dsp((__u8*)to, (__u8*)from, count);

   return ret;
}

/* The network code uses this one (see net/ipv4/udo.c:udp_send()). */
#include <asm/checksum.h>
unsigned int dspi_csum_partial_copy_from_dsp(const char *src, char *dst,
                                             int len, int sum)
{
   unsigned int ret;
   
   dbg("");
   /* Ah... I'm an evil bastard. First, pull the data across the bridge.
    */
   if (!dspi_in_arm(src))
      {
         err("0x%x not in remapped DSP RAM", (__u32)src);
         ret = -EINVAL;
      }
   else if ((ret = __api_copy_from_dsp((__u8*)dst, (__u8*)src, len)) >= 0)
      {
         /* Good. Now, use the standard checksum alg, copying the data from the
            buffer to itself. Inefficient? Oh, yes. But you try and figure out
            that fucked-up checksum algorithm for yourself before you criticize
            me. REVISIT. */
         ret = csum_partial_copy_fromuser(dst, dst, len, sum);
      }

   dbg("ret=0x%x", ret);
   return ret;        
}

void dspi_put_user(unsigned long x, void * y, int size)
{
   dbg("x=0x%x,y=0x%x,size=%d",(unsigned)x,(unsigned)y,(unsigned)size);
   
   if ((size < 1) || (size > 4))
      panic(__FUNCTION__ ": bad access length\n");
   else if (!dspi_in_arm(y))
      panic(__FUNCTION__ ": not a dsp arm address");
   else if (1 == size)//quick and easy
      __api_put_dsp_byte((__u8*)y, (x & 0xff));
   else if (!(size % 2) && !(((__u32)y) % 2))
      {/*nice clean word transfer, just remember to swap words for 4 bytes*/
         __u16 tmp;
         
         if (4 == size)
            {
               tmp = (x & 0xffff0000) >> 16;
               y += __api_dsp_aligned_copy((__u16*)y, &tmp, 2);
            }
         tmp = x & 0x0000ffff;
         __api_dsp_aligned_copy((__u16*)y, &tmp, 2);
      }
   else 
      {/*In these cases assume some sort of byte transfer, although
         from the information I have been able to gather you are asking
         for trouble using this functionality for this. This is meant for
         setting byte, word, and 32 bit word values not for byte copies. In
         this case we could be dealing with a byte by byte transfer so we will
         treat it as a byte copy. This should be safe also because the dsp
         cannot access odd addresses at this point so it only makes sense as
         some sort of byte transfer.
       */
         __u8 tmp[4];
         
         switch (size)
            {
            case 2:
               {
                  *((__u16*)tmp) = x & 0xffff;
                  break;
               }
            case 3:
               {
                  int i;
                  for (i=0;i<size;i++)
                     {
                        tmp[i] = x & 0xff;
                        x >>= 8;
                     }
                  break;
               }
            case 4:
               {
                  *((__u32*)tmp) = x;
                  break;
               }
            default:
               panic(__FUNCTION__ ": bad access length\n");
            }
         __api_copy_to_dsp((__u8*)y, tmp, size);
      }
}

unsigned long dspi_get_user(const void *y, int size)
{
   unsigned long ret = 0;
   //apitest();

   dbg("y=0x%x, size=%d", (unsigned)y, size);
   if ((size < 1) || (size > 4))
      panic(__FUNCTION__ ": bad access length\n");
   else if (!dspi_in_arm(y))
      panic(__FUNCTION__ ": not a dsp arm address");
   else if (1 == size)//quick and easy
      ret = __api_get_dsp_byte((__u8 *)y);
   else if (!(size % 2) && !(((__u32)y) % 2))
      {//nice clean word transfer, just remember to swap words for 4 byte
         __u16 tmp;
         
         if (4 == size)
            {
               y += __api_dsp_aligned_copy(&tmp, (__u16*)y, 2);
               ret = (__u32)tmp << 16;
            }
         
         __api_dsp_aligned_copy(&tmp, (__u16*)y, 2);
         ret |= tmp;
      }
   else
      {/*In these cases assume some sort of byte transfer, although
         from the information I have been able to gather you are asking
         for trouble using this functionality for this. This is meant for
         setting byte, word, and 32 bit word values not for byte copies. In
         this case we could be dealing with a byte by byte transfer so we will
         treat it as a byte copy. This should be safe also because the dsp
         cannot access odd addresses at this point so it only makes sense as
         some sort of byte transfer.
       */
         switch (size)
          {
          case 2:
             {
                __u16 tmp;
                __api_copy_from_dsp((__u8*)&tmp, (__u8*)y, size);
                ret = tmp;
                break;
             }
          case 3:
             {
                __u8 tmp[3];
                __api_copy_from_dsp(tmp, (__u8*)y, 3);
                ret = tmp[0] | ((__u32)tmp[1] << 8) | ((__u32)tmp[2] << 16);
                break;
             }
          case 4:
             {
                __api_copy_from_dsp((__u8*)&ret, (__u8*)y, size);
                break;
             }
          default:
             panic(__FUNCTION__ ": bad access length\n");
          }
      }

   dbg("The return value is %X\n", (unsigned)ret);
   return ret;
}

int mystopper = 0;
int istested = 0;
void apitest(void)
{
   char *string = "string";
   __u32 aint = 0xAABBCCDD, i;
   __u16 ashort = 0xFFAA;
   char buf[100];
   void *ptr = (void *)0xffe02f00;
   __u16 *ptr16;
   __u32 *ptr32;
   char *ptr8 = (__u8 *)buf;

   if (istested)
      return;
   else
      istested++;

   *((__u16*)ptr) = 0xBEEF;
   
   dbg("Kernel dsp comm put get test, buff address = %X\n", (unsigned)buf);
   dbg("Doing byte on even test");
   //first lets test the byte stuff using put
   dspi_put_user(string[0], ptr, 1);
   *ptr8 = dspi_get_user(ptr, 1);
   if (string[0] != *ptr8)
      {
         dbg("Failed to read and write a byte on even");
         return;
      }
   ptr += 3;
   ptr8++;
   
   dbg("Doing byte on odd test");
   dspi_put_user(string[0], ptr, 1);
   *ptr8 = dspi_get_user(ptr, 1);
   if (string[0] != *ptr8)
      {
         dbg("Failed to read and write a byte on odd");
         return;
      }
    ptr += 1;
    ptr8++;

    *((__u16*)ptr) = 0xBEEF;
    dbg("Doing word on even test");
    dspi_put_user(ashort, ptr, 2);
    ptr16 = (__u16 *)ptr8;
    *ptr16 = dspi_get_user(ptr, 2);
    if (*ptr16 != ashort)
       {
         dbg("Failed to read and write a word on even %X, %X", *ptr16, ashort);
         return;
      }
    ptr += 3;
    ptr16++;

    dbg("Doing word on odd test");
    dspi_put_user(ashort, ptr, 2);
    *ptr16 = dspi_get_user(ptr, 2);
    if (*ptr16 != ashort)
       {
         dbg("Failed to read and write a word on odd");
         return;
      }
    ptr += 3;
    ptr16++;

    dbg("Doing 3 bytes on even test");
    ptr32 = (__u32*)ptr16;
    dspi_put_user((aint >> 8), ptr, 3);
    *ptr32 = dspi_get_user(ptr, 3);
    if ((aint >> 8) != *ptr32)
       {
          dbg("Failed to read and write 3 bytes on even");
         return;
      }
    ptr += 5;
    ptr32++;

    dbg("Doing 3 bytes on odd test");
    dspi_put_user((aint >> 8), ptr, 3);
    *ptr32 = dspi_get_user(ptr, 3);
    if ((aint >> 8) != *ptr32)
       {
          dbg("Failed to read and write 3 bytes on odd");
         return;
      }
    ptr += 5;
    ptr32++;

    dbg("Doing 32 on even test");
    dspi_put_user(aint, ptr, 4);
    *ptr32 = dspi_get_user(ptr, 4);
    if (aint != *ptr32)
       {
          dbg("Failed to read and write 32 on even");
         return;
      }
    ptr += 5;
    ptr32++;

    dbg("Doing 32 on odd test");
    dspi_put_user(aint, ptr, 4);
    *ptr32 = dspi_get_user(ptr, 4);
    if (aint != *ptr32)
       {
          dbg("Failed to read and write 32 on odd");
         return;
      }
    ptr += 5;
    ptr32++;

    dbg("Doing memcopy on even of even");
    dspi_memcpy_to_dsp(ptr, string, 6);
    ptr8 = (__u8*)ptr32;
    dspi_memcpy_from_dsp(ptr8, ptr, 6);
    for (i = 0;i < 6;i++)
       {
          if (ptr8[i] != string[i])
             {
                dbg("Failed to read and write 32 on odd");
                return;
             }
       }
    ptr8 = &ptr8[i];
    ptr += i;

    dbg("Doing memcopy on odd of even");
    dspi_memcpy_to_dsp(ptr, string, 6);
    ptr8 = (__u8*)ptr32;
    dspi_memcpy_from_dsp(ptr8, ptr, 6);
    for (i = 0;i < 6;i++)
       {
          if (ptr8[i] != string[i])
             {
                dbg("Failed to read and write 32 on odd");
                return;
             }
       }
    ptr8 = &ptr8[i];
    ptr += i;

    dbg("Doing memcopy on even of odd");
    dspi_memcpy_to_dsp(ptr, string, 7);
    ptr8 = (__u8*)ptr32;
    dspi_memcpy_from_dsp(ptr8, ptr, 7);
    for (i = 0;i < 7;i++)
       {
          if (ptr8[i] != string[i])
             {
                dbg("Failed to read and write 32 on odd");
                return;
             }
       }
    ptr8 = &ptr8[i];
    ptr += i+1;

    dbg("Doing memcopy on odd of odd");
    dspi_memcpy_to_dsp(ptr, string, 7);
    ptr8 = (__u8*)ptr32;
    dspi_memcpy_from_dsp(ptr8, ptr, 7);
    for (i = 0;i < 7;i++)
       {
          if (ptr8[i] != string[i])
             {
                dbg("Failed to read and write 32 on odd");
                return;
             }
       }

    while(!mystopper);
    printk("Api comm test passed!\n");
}
