/*
 * File: hpi.c
 *
 * GPP-to-DSP communication for the kernel using hpi.
 *
 * Copyright (C) 2001 RidgeRun, Inc.
 * Author: Gordon McNutt <gmcnutt@ridgerun.com>
 *         Based on dspc.c by Steve Johnson.
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS  PROVIDED  ``AS  IS''  AND   ANY  EXPRESS  OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT,  INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Please report all bugs/problems to the author or <support@dsplinux.net>
 *
 * key: RRGPLCR (do not remove)
 *
 */

#include <linux/sched.h>
#include <asm/errno.h>
#include <asm/arch/hardware.h>

#include <asm/arch/dspi.h>

//#define DEBUG
#ifdef DEBUG
#define dbg(format, arg...) \
        printk(KERN_DEBUG __FUNCTION__ ": " format "\n" , ## arg)
#else
#define dbg(format, arg...) do {} while (0)
#endif
#define err(format, arg...) \
        printk(KERN_ERR __FUNCTION__ ": " format "\n" , ## arg)
#define info(format, arg...) \
        printk(KERN_INFO __FUNCTION__ ": " format "\n" , ## arg)
#define warn(format, arg...) \
        printk(KERN_WARNING __FUNCTION__ ": " format "\n" , ## arg)

#define HPI16       0x0010
#define HCNTL0	     0x1000
#define HCNTL1	     0x2000
#define HPIC        ~(HCNTL0 | HCNTL1)

#define HPI8_WRITE(x) {*((volatile __u16 *)(DSPRAM_BASE)) = x;}
#define HPI8_READ(x) {x = *((volatile __u16 *)(DSPRAM_BASE));}

typedef struct {
     __u16 hpib_ctrl;	
     __u16 hpib_stat;	
     __u16 int_clkset;	
     __u16 prog_stat;	
} dspc_t;

volatile static dspc_t *dspc = (dspc_t *) DSPC_REGISTER_BASE;

int dspi_load_code(void *to, const void *from, int count, int cpu)
{
        /* Revisit -- need a lock?
         */

        __u16 *__to, *__from;
        int i, ret = 0;

        /*
         * It would probably cause untold woe if anybody ever tries to copy to
         * an unaligned address in DSP RAM.
         */
        /*
         * Since this function will run on the ARM its arguments need to be in
         * the ARM portion of the shared memory map. Callers can use the
         * __dsp_to_arm() macro provided in the header to map DSP addresses to
         * their corresponding ARM addresses if they need to (the loader is an
         * example of this, since the addresses in the image are DSP 
         * addresses).
         */
        
        if (!dspi_in_dsp(to, cpu)) {
                warn("0x%x not aligned or in wrong address range", (__u32)to); 
                ret = -EINVAL;
                goto exit;
        }

        dspc->hpib_ctrl |= HPI16;

        __from = (__u16*)from;
        __to = (__u16*)dspi_dsp_to_arm(to, cpu);

//        dbg("__from = 0x%x, __to = 0x%x", (int)__from, (int)__to);

        for (i = 0; i < (count + 1) / 2; i++) {
                *__to = *__from;
                if (*__to++ != *__from++) {
                        ret = -EREMOTEIO;
                        goto exit_hpi16;
                }
        }

 exit_hpi16:
        dspc->hpib_ctrl &= ~HPI16;
 exit:
        return ret;
}

static void __hpi_aligned_copy_from_dsp(__u8 *to, int count)
{        
        while(count) {
                __u16 tmp;
                HPI8_READ(tmp);                
                *to = tmp & 0xffff;
                dbg("tmp = 0x%x, *to = 0x%x", tmp, *to);
                to++;
                count--;
                if (count) {
                        *to = (tmp >> 8) & 0xff;
                        dbg("tmp = 0x%x, *to = 0x%x", tmp, *to);
                        to++;
                        count--;
                }
        }
}
static void __hpi_aligned_copy_to_dsp(__u8 *from, int count)
{
        /*
         * Well, here we are facing the address alignment mismatch problem 
         * again. Assume 16bit data, and pack two ARM bytes per DSP word.
         * Then cross your fingers and hope the DSP app isn't reading into
         * a byte array.
         */
        while(count > 1) {
                __u16 tmp;
                tmp = *from & 0xff;
                from++;
                tmp |= (*from & 0xff) << 8;
                from++;
                dbg("tmp = 0x%x", tmp);
                HPI8_WRITE(tmp);
                count -= 2;
        }
        /*
         * Put the last byte by itself.
         */
        if (count) {
                dbg("*from = 0x%x", *from);
                HPI8_WRITE(*from & 0xff);
        }
}

int dspi_memcpy_from_dsp(void *to, const void *from, int count)
{
        int ret = 0;
        __u16 tmp;
        __u16 __from;

        /*
         * 'from' is a remapped dsp address. 'count' is the count in bytes.
         */

        dbg("to = 0x%x, from = 0x%x, count = %d", (int)to, (int)from, count);
        
        /* Make sure 'from' is in the part of DSP RAM the ARM can see. */
        if (!dspi_in_arm(from)) {
                err("0x%x not in remapped DSP RAM", (__u32)from);
                ret = -EINVAL;
                goto exit;
        }

        __from = dspi_arm_to_dsp((__u32)from);

        tmp = dspc->hpib_ctrl & HPIC;
        dspc->hpib_ctrl = tmp | HCNTL1;
        HPI8_WRITE(__from);
        dspc->hpib_ctrl = tmp | HCNTL0;

        /* If 'from' is not on a word boundary then handle the first byte
           specially. */
        if ((__u32)from % 2) {
                __u16 tmp;

                /* Copy the high byte only. */
                HPI8_READ(tmp);                
                *((__u8*)&to) = (tmp >> 8);
                dbg("tmp = 0x%x, *to = 0x%x", tmp, *((__u8*)&to));
                to++;
                count--;
        }

        __hpi_aligned_copy_from_dsp((__u8*)to, count);

 exit:
        return ret;
        
}
int dspi_memcpy_to_dsp(void *to, const void *from, int count)
{
        int ret = 0;
        __u16 tmp;
        __u16 __to;

        /*
         * 'to' is a remapped dsp address. 'count' is the count in bytes.
         */

        dbg("to = 0x%x, from = 0x%x, count = %d", (int)to, (int)from, count);
        
        /* Make sure 'to' is in the part of DSP RAM the ARM can see. */
        if (!dspi_in_arm(to)) {
                err("0x%x not in remapped DSP RAM", (__u32)to);
                ret = -EINVAL;
                goto exit;
        }

        __to = dspi_arm_to_dsp((__u32)to);

        tmp = dspc->hpib_ctrl & HPIC;
        dspc->hpib_ctrl = tmp | HCNTL1;
        HPI8_WRITE(__to - 1); /* Don't forget -- HPI8 preincrements on write */
        dspc->hpib_ctrl = tmp | HCNTL0;

        if ((__u32)to % 2) {
                __u16 tmp;

                /* Write to the high byte only. */
                tmp = *((__u8*)&from) << 8;
                HPI8_WRITE(tmp);
                dbg("tmp = 0x%x", tmp);
                from++;
                count--;
        }

        __hpi_aligned_copy_to_dsp((__u8*)from, count);

 exit:
     return ret;
}

/* The network code uses this one (see net/ipv4/udo.c:udp_send()). */
#include <asm/checksum.h>
unsigned int dspi_csum_partial_copy_from_dsp(const char *src /* fake */, 
                                           char *dst, int len, int sum)
{
        unsigned int ret;

        dbg("");

        /* Ah... I'm an evil bastard. First, pull the data across the bridge.
         */
        if ((ret = dspi_memcpy_from_dsp(dst, src, len)) < 0)
                goto exit;

        /* Good. Now, use the standard checksum alg, copying the data from the
           buffer to itself. Inefficient? Oh, yes. But you try and figure out
           that fucked-up checksum algorithm for yourself before you criticize
           me. REVISIT. */
        ret = csum_partial_copy_fromuser(dst, dst, len, sum);        

 exit:
        dbg("ret=0x%x", ret);
        return ret;        
}

void dspi_put_user(unsigned long x, void * y /* fake */, int size)
{
        __u16 __y;
        __u16 tmp;

        __y = dspi_arm_to_dsp((__u32)y);

        dbg("x=0x%x,y=0x%x,size=%d",x,y,size);

        if ((__u32)y % 2) {

                /* Oh yes, this is the shits. But we have to support unaligned
                   puts. Somebody (anybody!) feel free to improve upon this.
                   Oh, by the way, this has never been tested ;).
                 */
                
                if ((size < 1) || (size > 4))
                        panic(__FUNCTION__ ": bad access length\n");

                /* Write byte 0 to the high byte of the first word address. */
                tmp = dspc->hpib_ctrl & HPIC;
                dspc->hpib_ctrl = tmp | HCNTL1;
                HPI8_WRITE(__y);
                dspc->hpib_ctrl = tmp | HCNTL0;

                HPI8_READ(tmp);
                tmp &= 0xff;
                tmp |= x << 8;

                /* That read just autoincremented the address on the DSP to
                   address 2. Sigh. Put it back to address 1. */
                tmp = dspc->hpib_ctrl & HPIC;
                dspc->hpib_ctrl = tmp | HCNTL1;
                HPI8_WRITE(__y - 1);
                dspc->hpib_ctrl = tmp | HCNTL0;
                HPI8_WRITE(tmp);

                if (size < 2)
                        return;

                x >>= 8;
                __y++; /* inc to address 2 */

                if (size < 3) {
                        
                        /* Write byte 1 to the low byte of the second word 
                           address. The last write should have autoincremented
                           to the next word address. */
                        goto write_low_byte;
                        return;
                }

                printk("**********************************\n");
                printk(__FUNCTION__ " putting more than 2 bytes unaligned\n");
                printk("is probably not implemented correctly!\n");
                printk("**********************************\n");


                /* Write bytes 1 and 2 to word address 2. The last write should
                   have autoincremented to address 2. */
                HPI8_WRITE((__u16)(x & 0xffff));

                if (size != 4)
                        return;

                x >>= 8;
                __y++; /* inc to address 3 */

                /* Write byte 3 to the low byte of address 3. The last write
                   should have autoincremented to address 3. */
        write_low_byte:
                HPI8_READ(tmp);
                tmp &= 0xff00;
                tmp |= x >> 8;
                
                /* That read just autoincremented the address on the DSP. Put 
                   it back. */
                tmp = dspc->hpib_ctrl & HPIC;
                dspc->hpib_ctrl = tmp | HCNTL1;
                HPI8_WRITE(__y - 1);
                dspc->hpib_ctrl = tmp | HCNTL0;
                HPI8_WRITE(tmp);                
                return;

        } else
        {
                tmp = dspc->hpib_ctrl & HPIC;
                dspc->hpib_ctrl = tmp | HCNTL1;
                HPI8_WRITE(__y - 1);
                dspc->hpib_ctrl = tmp | HCNTL0;

                switch (size) {
                case 1:
                        HPI8_WRITE((__u8)(x & 0xff));
                        break;
                case 2:
                        HPI8_WRITE((__u16)(x & 0xffff));
                        break;
                case 3:
                        HPI8_WRITE((__u16)(x & 0xffff));
                        HPI8_WRITE((__u8)((x >> 16) & 0xff));
                        break;
                case 4: /* swap words */
                        HPI8_WRITE((__u16)((x >> 16) & 0xffff));
                        HPI8_WRITE((__u16)(x & 0xffff));
                        break;
                default:
                   panic(__FUNCTION__ ": bad user access length %d\n", size);
                }
        }
}

unsigned long dspi_get_user(const void *y /* fake */, int size)
{
        __u16 __y;
        __u16 tmp;
        unsigned long ret;

         __y = dspi_arm_to_dsp((__u32)y);

        dbg("y=0x%x, __y=0x%x, size=%d", y, __y, size);

        tmp = dspc->hpib_ctrl & HPIC;
        dspc->hpib_ctrl = tmp | HCNTL1;
        HPI8_WRITE(__y);
        dspc->hpib_ctrl = tmp | HCNTL0;

        if ((__u32)y % 2) {
                /* Unaligned reads. Not as bad as unaligned puts... */
                switch (size) {
                case 1:
                        HPI8_READ(ret);
                        ret = ret >> 8;
                        break;
                case 2:
                        HPI8_READ(ret);
                        ret = ret >> 8;
                        HPI8_READ(tmp);
                        ret |= (tmp & 0xff) << 8;
                        break;
                case 3:
                        HPI8_READ(tmp);
                        ret = ret >> 8;
                        HPI8_READ(ret);
                        ret |= ((tmp & 0xff) << 8);
                        break;
                case 4: /* swap the words */
                        HPI8_READ(tmp);
                        HPI8_READ(ret);
                        ret |= ((tmp & 0xffff) << 16);
                        break;
                default:
                        panic(__FUNCTION__ ": bad_user_access_length %d\n", 
                              size);
        }
        } else
           switch (size) {
           case 1:
              HPI8_READ(ret);
              ret &= 0xff;                
              break;
           case 2:
              HPI8_READ(ret);
              ret &= 0xffff;
              break;
           case 3:
              HPI8_READ(tmp);
              HPI8_READ(ret);
              ret |= ((tmp & 0xff) << 16);
              break;
           case 4: /* swap the words */
              HPI8_READ(tmp);
              //                printk(__FUNCTION__ ":lo=0x%x\n", ret);
              HPI8_READ(ret);
              //                printk(__FUNCTION__ ":hi=0x%x\n", tmp);
              ret |= ((tmp & 0xffff) << 16);
              break;
           default:
              panic(__FUNCTION__ ": bad_user_access_length %d\n", size);
           }

        dbg("ret=0x%x", ret);
       
        return ret;
}

int dspi_dsp_memset(void *to, __u16 pattern, int count, int cpu)
{
        __u16 *__to;
        int i, ret = 0;
        
        dspc->hpib_ctrl |= HPI16;

        __to = (__u16*)to;

        for (i = 0; i < (count + 1) / 2; i++) {
                *__to = pattern;
                if (*__to != pattern) {
                        ret = -EREMOTEIO;
                        goto exit;
                }
                __to++;
        }

 exit:
        dspc->hpib_ctrl &= ~HPI16;
        return ret;        
}

