/*
 * irqs-dsc21.c:
 *         dsc21 IRQ mask, unmask & ack routines.
 * copyright:
 *         (C) 2001 RidgeRun, Inc. (http://www.ridgerun.com)
 * author: Steve Johnson <stevej@ridgerun.com>
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 * THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 * WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 * USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * You should have received a copy of the  GNU General Public License along
 * with this program; if not, write  to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/autoconf.h>

#include <asm/io.h>
#include <asm/system.h>

#include <asm/hardware.h>
#include <asm/irq.h>
#include <asm/arch/irq.h>

extern void do_IRQ(int irq, struct pt_regs * regs);

static inline void irq_ack(unsigned int irq)
{
    u32 reg;
    reg = (irq > 15) ? DSC21_IRQ0_STATUS : DSC21_IRQ1_STATUS;
    *((u16*)reg) = irq_to_mask[irq];
}

asmlinkage int probe_IRQ_interrupt(int irq, struct pt_regs * regs)
{
    printk("probe interrupt %d received!\n", irq);
    mask_irq(irq);
    irq_ack(irq);
    return 0;
}

asmlinkage int bad_IRQ_interrupt(int irq, struct pt_regs * regs)
{
    printk("bad interrupt %d received!\n", irq);
    irq_ack(irq);
    return 0;
}

asmlinkage int IRQ_interrupt(int irq, struct pt_regs * regs)
{
    register unsigned long flags;
    register unsigned long saved_count;

    saved_count = intr_count;
    intr_count = saved_count + 1;

    save_flags(flags);
    irq_ack(irq);
    do_IRQ(irq, regs);
    restore_flags(flags);
    intr_count = saved_count;
    return 0;
}


asmlinkage int timer_IRQ_interrupt(int irq, struct pt_regs * regs)
{
    register unsigned long flags;
    register unsigned long saved_count;

#ifdef IRQ_LATENCY_TIMING
    static int timing_max = 0;
    static int timing_min = 0xffff;
    register int timing_start = inw(0x7000);
    static int timing_average = 0;
    static int timing_list_cnt = 0;
    static int timing_list[50];
    static int timing_latency = 0;
    static int timing_latency_max = 0;
    static int timing_latency_min = 0xffff;
    register int timing_current = inw(DSPLINUX_TIMER0_CNT);
    static int timing_latency_average = 0;

    if (timing_start > timing_max) {
         timing_max = timing_start;
    }
    if (timing_start < timing_min) {
         timing_min = timing_start;
    }
    timing_average = (timing_average + timing_start) / 2;
    if ((timing_start > 0x30) &&
        (timing_list_cnt < 50)) {
         timing_list[timing_list_cnt] = timing_start;
         timing_list_cnt++;
    }
    if (timing_start > timing_current) {   // check for wrap
         timing_latency = timing_current - timing_start + 27000;
    }
    else {
         timing_latency = timing_current - timing_start;
    }
    if (timing_latency > timing_latency_max) {
         timing_latency_max = timing_latency;
    }
    if (timing_latency < timing_latency_min) {
         timing_latency_min = timing_latency;
    }
    timing_latency_average = (timing_latency_average + timing_latency) / 2;
#endif    

    saved_count = intr_count;
    intr_count = saved_count + 1;

    save_flags(flags);
    do_IRQ(irq, regs);
    restore_flags(flags);
    intr_count = saved_count;
    irq_ack(irq);
    return 0;
}


asmlinkage int fast_IRQ_interrupt(int irq, struct pt_regs * regs)
{
    register unsigned long saved_count;

    saved_count = intr_count;
    intr_count = saved_count + 1;

    do_IRQ(irq, regs);
    cli();
    intr_count = saved_count;
    *((u16*)(DSC21_FIQ_STATUS)) |= (0x1 << irq);
    return 1;
}

void init_irq(void)
{

	/* disable all */
	outw(0, DSC21_FIQ_ENABLE);
	outw(0, DSC21_IRQ0_ENABLE);
	outw(0, DSC21_IRQ1_ENABLE);
    
	/* clear all status */
	outw(0xffff, DSC21_FIQ_STATUS);
	outw(0xffff, DSC21_IRQ0_STATUS);
	outw(0xffff, DSC21_IRQ1_STATUS);
    
}
