/*
 * irqs-orion.c:
 *         Orion (C5472) IRQ mask, unmask & ack routines.
 *
 * copyright:
 *         (C) 2001 RidgeRun, Inc. (http://www.ridgerun.com)
 *
 * author: Steve Kranz <skranz@ridgerun.com>
 *
 * leverage: This file leveraged from the irqs-dsc21.c file
 *           originally written by Steve Johnson <stevej@ridgerun.com>
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 * THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 * WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 * USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * You should have received a copy of the  GNU General Public License along
 * with this program; if not, write  to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/autoconf.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/arch/irq.h>

extern void do_IRQ(int irq, struct pt_regs * regs);

#define SRC_IRQ_REG  ((volatile unsigned long *)0xffff2d08)
#define SRC_FIQ_REG  ((volatile unsigned long *)0xffff2d0C)
#define INT_CTRL_REG ((volatile unsigned long *)0xffff2d18)

/******************************
 Routine:
 Description:
 ******************************/
static inline void irq_ack(unsigned int irq)
{
    unsigned long reg;
    reg = *SRC_IRQ_REG; // Insure appropriate IT_REG bit clears.
    *INT_CTRL_REG |= 0x00000001; // write the NEW_IRQ_AGR bit.
}

/******************************
 Routine:
 Description:
 ******************************/
static inline void fiq_ack(unsigned int irq)
{
    unsigned long reg;
    reg = *SRC_FIQ_REG; // Insure appropriate IT_REG bit clears.
    *INT_CTRL_REG |= 0x00000002; // write the NEW_FIQ_AGR bit.
}

/******************************
 Routine:
 Description:
 ******************************/
asmlinkage int probe_IRQ_interrupt(int irq, struct pt_regs * regs)
{
    printk("probe interrupt %d received!\n", irq);
    mask_irq(irq);
    irq_ack(irq);
    return 0;
}

/******************************
 Routine:
 Description:
 ******************************/
asmlinkage int bad_IRQ_interrupt(int irq, struct pt_regs * regs)
{
    printk("bad interrupt %d received!\n", irq);
    irq_ack(irq);
    return 0;
}

/******************************
 Routine:
 Description:
 ******************************/
asmlinkage int IRQ_interrupt(int irq, struct pt_regs * regs)
{
    register unsigned long flags;
    register unsigned long saved_count;
    saved_count = intr_count;
    intr_count = saved_count + 1;
    save_flags(flags);
    irq_ack(irq);
    do_IRQ(irq, regs);
    restore_flags(flags);
    intr_count = saved_count;
    return 0;
}

/******************************
 Routine:
 Description:
 ******************************/
asmlinkage int timer_IRQ_interrupt(int irq, struct pt_regs * regs)
{
    register unsigned long flags;
    register unsigned long saved_count;

#ifdef IRQ_LATENCY_TIMING
    static int timing_max = 0;
    static int timing_min = 0xffff;
    register int timing_start = inw(0x7000);
    static int timing_average = 0;
    static int timing_list_cnt = 0;
    static int timing_list[50];
    static int timing_latency = 0;
    static int timing_latency_max = 0;
    static int timing_latency_min = 0xffff;
    register int timing_current = inw(DSPLINUX_TIMER2_CNT);
    static int timing_latency_average = 0;

    if (timing_start > timing_max) {
         timing_max = timing_start;
    }
    if (timing_start < timing_min) {
         timing_min = timing_start;
    }
    timing_average = (timing_average + timing_start) / 2;
    if ((timing_start > 0x30) &&
        (timing_list_cnt < 50)) {
         timing_list[timing_list_cnt] = timing_start;
         timing_list_cnt++;
    }
    if (timing_start > timing_current) {   // check for wrap
         timing_latency = timing_current - timing_start + 27000;
    }
    else {
         timing_latency = timing_current - timing_start;
    }
    if (timing_latency > timing_latency_max) {
         timing_latency_max = timing_latency;
    }
    if (timing_latency < timing_latency_min) {
         timing_latency_min = timing_latency;
    }
    timing_latency_average = (timing_latency_average + timing_latency) / 2;
#endif

    saved_count = intr_count;
    intr_count = saved_count + 1;
    save_flags(flags);
    irq_ack(irq);
    do_IRQ(irq, regs);
    restore_flags(flags);
    intr_count = saved_count;
    return 0;
}


/******************************
 Routine:
 Description:
 ******************************/
asmlinkage int fast_IRQ_interrupt(int irq, struct pt_regs * regs)
{
    register unsigned long saved_count;
    saved_count = intr_count;
    intr_count = saved_count + 1;
    do_IRQ(irq, regs);
    cli();
    intr_count = saved_count;
    fiq_ack(irq);
    return 1;
}

/******************************
 Routine:
 Description:
 ******************************/
void init_irq(void)
{
    #define INT_FIRST_IO  0xffff2d00
    #define INT_IO_RANGE  0x5C
    #define ILR_IRQ0_REG  ((volatile unsigned long *)0xffff2d1C) /* 0  - Timer 0            */
    #define ILR_IRQ1_REG  ((volatile unsigned long *)0xffff2d20) /* 1  - Timer 1            */
    #define ILR_IRQ2_REG  ((volatile unsigned long *)0xffff2d24) /* 2  - Timer 2            */
    #define ILR_IRQ3_REG  ((volatile unsigned long *)0xffff2d28) /* 3  - GPIO0              */
    #define ILR_IRQ4_REG  ((volatile unsigned long *)0xffff2d2c) /* 4  - Ethernet           */
    #define ILR_IRQ5_REG  ((volatile unsigned long *)0xffff2d30) /* 5  - KBGPIO[7:0]        */
    #define ILR_IRQ6_REG  ((volatile unsigned long *)0xffff2d34) /* 6  - Uart (serial port) */
    #define ILR_IRQ7_REG  ((volatile unsigned long *)0xffff2d38) /* 7  - uart (IRDA)        */
    #define ILR_IRQ8_REG  ((volatile unsigned long *)0xffff2d3c) /* 8  - KBGPIO[15:8]       */
    #define ILR_IRQ9_REG  ((volatile unsigned long *)0xffff2d40) /* 9  - GPIO3              */
    #define ILR_IRQ10_REG ((volatile unsigned long *)0xffff2d44) /* 10 - GPIO2              */
    #define ILR_IRQ11_REG ((volatile unsigned long *)0xffff2d48) /* 11 - I2C                */
    #define ILR_IRQ12_REG ((volatile unsigned long *)0xffff2d4c) /* 12 - GPIO1              */
    #define ILR_IRQ13_REG ((volatile unsigned long *)0xffff2d50) /* 13 - SPI                */
    #define ILR_IRQ14_REG ((volatile unsigned long *)0xffff2d54) /* 14 - GPIO[19:4]         */
    #define ILR_IRQ15_REG ((volatile unsigned long *)0xffff2d58) /* 15 - API                */
    #define EdgeSensitive 0x00000020
    #define Priority      0x0000001E
  
    *((volatile unsigned long *)MASK_IT_REG) = 0x0000ffff; // Disable all interrupts.
    irq_ack(0); // clear pending.
    fiq_ack(0); // clear pending.
    *((volatile unsigned long *)IT_REG) = 0x00000000; // clear pending.
    *ILR_IRQ2_REG = EdgeSensitive | Priority;  // override the h/w default.
    *ILR_IRQ4_REG = EdgeSensitive | Priority;  // override the h/w default.
    *ILR_IRQ6_REG = Priority;                  // override the h/w default.
    *ILR_IRQ15_REG = EdgeSensitive | Priority; // override the h/w default.

    // Next, Be a good citizen and do what it takes to insure
    // the user gets accurate info if they type "cat /proc/ioports"
    request_region(INT_FIRST_IO,INT_IO_RANGE,"interrupt controller");
}
