/****************************************************************************/

/*
 *	hifn7901.c   -- Driver for Hi/fn 7901 encryption engine.
 *
 *	(C) Copyright 2000, Greg Ungerer (gerg@lineo.com)
 *
 *	This program is a part of the Linux kernel, and may be freely
 *	copied under the terms of the GNU General Public License (GPL),
 *	version 2, or at your option any later version.
 */

/****************************************************************************/

#include <linux/config.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/string.h>

#include <asm/system.h>
#include <asm/delay.h>
#include <asm/irq.h>
#include <asm/bitops.h>

#include "hifn7901.h"

/****************************************************************************/

/*
 *	Hifn 7901 hardware support.
 */
volatile unsigned int	hifn7901_memaddr = 0x30c00000;
int			hifn7901_present = 0;


/*
 *	Define maximum buffer size we will be able to decode at once.
 *	We need enough room in Hifn memory for source and destination
 *	buffers, and enough for command overhead...
 */
#define	HIFN7901_MAXBUF	0xf00


/*
 *	Global pointers for current destination and results. This is just
 *	plain laziness on my part, these should be read back from the
 *	Hifn chip.
 */
void		*hifn7901_resp;
void		*hifn7901_dstp;


/*
 *	Standard DES type definitions. Probably should use a des.h
 *	from somewhere for this...
 */
typedef unsigned char	des_cblock[8];
typedef struct des_ks_struct { des_cblock _; }	des_key_schedule[16];


/*
 *	Macro to calculate a relative pointer based on Hifn memory address.
 */ 
#define	HIFN7901_PTR(x)		(((unsigned long) (x)) - hifn7901_memaddr - \
					 HIFN7901_MEMBASE)
#define	HIFN7901_DSOFF(x)	(((x) % 8) * sizeof(struct hifn7901desc))


/*
 *	If testing or debugging agaisnt a real libdes then use these
 *	redefines to rename functions so both can co-exist.
 */

/*#define	HIFN7901_TEST	1*/

#ifdef HIFN7901_TEST
#define	lib_des_set_key			des_set_key
#define	lib_des_cbc_encrypt		des_cbc_encrypt
#define	lib_des_ede3_cbc_encrypt	des_ede3_cbc_encrypt
#else
#define	hifn7901_des_set_key		des_set_key
#define	hifn7901_des_cbc_encrypt	des_cbc_encrypt
#define	hifn7901_des_ede3_cbc_encrypt	des_ede3_cbc_encrypt
#endif

/****************************************************************************/

/*
 *	Fast 32bit based memory set for Hifn. It is actually better
 *	(and smaller) to inline this.
 */

void __inline__ *hifn7901_memset(void *s, int c, int len)
{
	register volatile unsigned long	*sp;

	sp = (volatile unsigned long *) s;
	len >>= 2;

	for (; len; len--)
		*sp++ = c;
	return(s);
}

/****************************************************************************/

/*
 *	Fast 32bit based memory copy for Hifn. It is actually better
 *	(and smaller) to inline this.
 */

void __inline__ *hifn7901_memcpyto(void *dst, void *src, int len)
{
	register volatile unsigned long	*dp;
	register unsigned long		*sp;

	dp = (volatile unsigned long *) dst;
	sp = (unsigned long *) src;
	len >>= 2;

	for (; len; len--)
		*dp++ = *sp++;
	return(dst);
}

/****************************************************************************/

/*
 *	Fast 32bit based memory copy for Hifn. It is actually better
 *	(and smaller) to inline this.
 */

void __inline__ *hifn7901_memcpyfrom(void *dst, void *src, int len)
{
	register volatile unsigned long	*sp;
	register unsigned long		*dp;

	dp = (unsigned long *) dst;
	sp = (volatile unsigned long *) src;
	len >>= 2;

	for (; len; len--)
		*dp++ = *sp++;
	return(dst);
}

/****************************************************************************/

/*
 *	An efficent key copy function. We know keys are 8 bytes,
 *	so I will just do 2 int copies (better then memcpy for this).
 *	This optimizes very nicely...
 */

void __inline__ hifn7901_keycpy(void *dst, void *src)
{
	unsigned int	*dp = (unsigned int *) dst;
	unsigned int	*sp = (unsigned int *) src;

	dp[0] = sp[0];
	dp[1] = sp[1];
}

/*
 *	Copying ivecs is the same as copying keys for DES.
 */
#define	hifn7901_iveccpy	hifn7901_keycpy

/****************************************************************************/

/*
 *	Process a Hifn 7901 encryption command. The mechanism is essentially
 *	the same, no matter what algorithm or options are specified. Cmd
 *	specifies what to actually do (DES, Triple DES, etc).
 */

int hifn7901_sendcmd(void *input, long inlength, long outlength,
	unsigned long cmd, unsigned long cmdenc, unsigned long cmdcomp,
	void *key, int keysize, des_cblock *ivec)
{
	struct hifn7901desc	hifnds;
	struct hifn7901cmd	hifncmd;
	struct hifn7901comp	hifncomp;
	struct hifn7901enc	hifnenc;
	volatile unsigned char	*cmdp, *srcp, *dstp, *resp;
	unsigned long		stat;
	unsigned char		*mp;
	int			cmdlen, reslen, i;
	int			cmdindex, srcindex, dstindex, resindex;

#if 0
	printk("hifn7901_sendcmd(input=%x,inlength=%d,outlength=%d,cmd=%x,"
		"cmsenc=%x,cmdcomp=%x,key=%x,keysize=%d,ivec=%x)\n",
		(int) input, (int) inlength, (int) outlength, (int) cmd,
		(int) cmdenc, (int) cmdcomp, (int) key, keysize, (int) ivec);
#endif

	mp = (unsigned char *) (hifn7901_memaddr + HIFN7901_PKTDATA);

	/*
	 *	Setup command structure in Hifn memory. We must ONLY do
	 *	32bit reads and writes to the hifn, so fill out structures
	 *	locally then use hifn_memcpy to put them in place.
	 */
	/* Base command structure. */
	hifncmd.cmd = cmd;
	hifncmd.scount = HIFN7901_SWAPB(inlength);
	hifncmd.dcount = HIFN7901_SWAPB(outlength);

	hifn7901_memcpyto(mp, &hifncmd, sizeof(hifncmd));
	cmdp = (volatile unsigned char *) mp;
	mp += sizeof(hifncmd);

	/* Compression command structure. */
	if (cmd & HIFN7901_CMD_COMP) {
		hifncomp.cmd = cmdcomp;
		hifncomp.hdrcnt = HIFN7901_SWAPB(0);
		hifncomp.srccnt = HIFN7901_SWAPB(inlength);
		hifncomp.reserved = 0;

		hifn7901_memcpyto(mp, &hifncomp, sizeof(hifncomp));
		mp += sizeof(hifncomp);
	}

	/* Encryption command structure. */
	if (cmd & HIFN7901_CMD_ENCRYPT) {
		hifnenc.cmd = cmdenc | HIFN7901_ENC_CLEAR;
		if (key)
			hifnenc.cmd |= HIFN7901_ENC_NEWKEY;
		if (ivec)
			hifnenc.cmd |= HIFN7901_ENC_NEWIV;
		hifnenc.hdrcnt = HIFN7901_SWAPB(0);
		hifnenc.srccnt = HIFN7901_SWAPB(inlength);
		hifnenc.reserved = 0;

		hifn7901_memcpyto(mp, &hifnenc, sizeof(hifnenc));
		mp += sizeof(hifnenc);

		/* Encryption keys and initialization vector, if new. */
		if (key) {
			hifn7901_memcpyto(mp, key, keysize);
			mp += keysize;
		}
		if (ivec) {
			hifn7901_memcpyto(mp, ivec, 8);
			mp += 8;
		}
	}

	cmdlen = mp - cmdp;	/* Command length including keys and ivec. */

	/* Setup source data for engine - round to nearest word boundary. */
	i = (inlength + 3) & 0xfffffffc;
	hifn7901_memcpyto(mp, input, i);
	srcp = mp;
	mp += i;

	/* Setup memory space for the results structure. */
	resp = mp;
	reslen = sizeof(struct hifn7901res);
	if (cmd & HIFN7901_CMD_COMP)
		reslen += sizeof(struct hifn7901compres);
	if (cmd & HIFN7901_CMD_ENCRYPT)
		reslen += sizeof(struct hifn7901encres);
	mp += reslen;

	/* Destination buffer space. */
	dstp = mp;
	mp += outlength;

	/*
	 *	Figure out what index number we can ue. Query the device
	 *	index registers (in Status register) to find out.
	 */
	stat = *((volatile unsigned long *) (hifn7901_memaddr + HIFN7901_PKTS));

	/* Command index and descriptors */
	cmdindex = HIFN7901_PKTS_CINDEX(stat);
	hifnds.cmd = (cmdlen & HIFN7901_DESC_LENMASK) | HIFN7901_DESC_LAST;
	hifnds.ptr = HIFN7901_PTR(cmdp);
	hifn7901_memcpyto((void *) (hifn7901_memaddr + HIFN7901_CMDD +
		HIFN7901_DSOFF(cmdindex)), &hifnds, sizeof(hifnds));

	/* Source descriptors for keys/ivec and data */
	srcindex = HIFN7901_PKTS_SINDEX(stat);
	hifnds.cmd = (inlength & HIFN7901_DESC_LENMASK) | HIFN7901_DESC_LAST;
	hifnds.ptr = HIFN7901_PTR(srcp);
	hifn7901_memcpyto((void *) (hifn7901_memaddr + HIFN7901_SRCD +
		HIFN7901_DSOFF(srcindex)), &hifnds, sizeof(hifnds));

	/* Results descriptor info */
	resindex = HIFN7901_PKTS_RINDEX(stat);
	hifnds.cmd = (reslen & HIFN7901_DESC_LENMASK);
	hifnds.ptr = HIFN7901_PTR(resp);
	hifn7901_memcpyto((void *) (hifn7901_memaddr + HIFN7901_RESD +
		HIFN7901_DSOFF(resindex)), &hifnds, sizeof(hifnds));

	/* Setup destination index and descriptors */
	dstindex = HIFN7901_PKTS_DINDEX(stat);
	hifnds.cmd = (outlength & HIFN7901_DESC_LENMASK);
	hifnds.ptr = HIFN7901_PTR(dstp);
	hifn7901_memcpyto((void *) (hifn7901_memaddr + HIFN7901_DSTD +
		HIFN7901_DSOFF(dstindex)), &hifnds, sizeof(hifnds));

	/* Dude, lazy... */
	hifn7901_resp = (void *) resp;
	hifn7901_dstp = (void *) dstp;

	/*
	 *	The Hifn 7901 doco is very misleading on what the host index
	 *	values should be. The trick is that the device will use its
	 *	device index for the next command, you need only make the
	 *	host index different, so that it knows to start processing.
	 */
	cmdindex = (cmdindex + 1) & 0xf;
	srcindex = (srcindex + 1) & 0xf;
	resindex = (resindex + 1) & 0xf;
	dstindex = (dstindex + 1) & 0xf;

	*((volatile unsigned long *)(hifn7901_memaddr+HIFN7901_HRI)) = resindex;
	*((volatile unsigned long *)(hifn7901_memaddr+HIFN7901_HDI)) = dstindex;
	*((volatile unsigned long *)(hifn7901_memaddr+HIFN7901_HSI)) = srcindex;
	*((volatile unsigned long *)(hifn7901_memaddr+HIFN7901_HCI)) = cmdindex;

	return(0);
}

/****************************************************************************/

/*
 *	Ugh, poll the command completeion bit from the Hifn. We should
 *	really switch this to use interrupts _real_ soon.
 */

int hifn7901_pollcmd(void)
{
	volatile unsigned long	*statp;
	int			i;

	statp = (volatile unsigned long *) (hifn7901_memaddr + HIFN7901_PKTS);

#if 0
	printk("hifn7901_pollcmd(): status=%x\n", (int) *statp);
#endif

	for (i = 0; (i < 0x20000); i++) {
		if ((*statp & HIFN7901_PKTS_ALLDONE) == HIFN7901_PKTS_ALLDONE)
			return(0);
	}

#if 1
	printk("%s(%d): HIFN7901 POLL -FAILED- with i=%d, STATUS=%x\n",
		__FILE__, __LINE__, i, (int) *statp);
#endif
	return(0);
}

/****************************************************************************/

/*
 *	Complete a Hifn command. Copy the resultant data back into the
 *	output buffer. Also update ivec if it is not null.
 */

int hifn7901_getcmd(void *src, long inlength, void *dst, long outlength,
	des_cblock *ivec, int enc)
{
	struct hifn7901res		res;
	/*struct hifn7901compres	rescomp;*/
	/*struct hifn7901encres		resenc;*/
	volatile unsigned long		*statp;
	unsigned long			status;
	int				size;

	statp = (volatile unsigned long *) (hifn7901_memaddr + HIFN7901_PKTS);

#if 0
	printk("hifn7901_getcmd(src=%x,inlength=%d,dst=%x,outlength=%d,"
		"ivec=%x,enc=%d)\n", (int) src, (int) inlength, (int) dst,
		(int) outlength, (int) ivec, enc);
#endif

	/* Read Packet Status bits */
	status = *statp;
 
	if ((enc == 0) && ivec)
		hifn7901_memcpyfrom(ivec, (src + inlength - 8), 8);
		
	/* Get results structure... */
	hifn7901_memcpyfrom(&res, hifn7901_resp, sizeof(res));
	res.srccnt = HIFN7901_SWAPB(res.srccnt);
	res.dstcnt = HIFN7901_SWAPB(res.dstcnt);

	size = (int) (outlength - res.dstcnt);
	if (size < 0) 
		size = 0;
	if (size > outlength)
		size = outlength;

	/* Get resulting data from hifn into output buffer */
	/* This rounding is bad!  Could be past end of buffer :-( */
	hifn7901_memcpyfrom(dst, hifn7901_dstp, ((size + 3) & 0xfffffffc));

	if (enc && ivec)
		hifn7901_memcpyfrom(ivec, (hifn7901_dstp + outlength - 8), 8);

	/* Clear done status bits from Packet Status -- doco wrong */
	*statp = 0xffffffff;

	return(size);
}

/****************************************************************************/

/*
 *	As per standard DES library functions.
 */

int hifn7901_des_set_key(des_cblock *key, des_key_schedule schedule)
{
#if 0
	unsigned int	*kp = (unsigned int *) key;
	printk("des_set_key(key=%x,schedule=%x) -> key=%08x%08x\n",
		(int) key, (int) schedule, kp[0], kp[1]);
#endif

	/* FIXME: need to check for non-odd parity of key, return -1 */
	/* FIXME: need to check for illegal week key, return -2 */

	memcpy(schedule, key, sizeof(*key));
	return(0);
}

/****************************************************************************/

/*
 *	As per standard DES library functions.
 */

int hifn7901_des_cbc_encrypt(des_cblock *input, des_cblock *output, long length,
	des_key_schedule schedule, des_cblock *ivec, int enc)
{
	des_cblock	myvec;
	long		size;
	int		cmd;

#if 0
	printk("des_cbc_encrypt(input=%x,output=%x,length=%d,schedule=%x,"
		"ivec=%x,encrypt=%d)\n", (int) input, (int) output,
		(int) length, (int) schedule, (int) ivec, enc);
#endif

	cmd = HIFN7901_CMD_ENCRYPT |
		(enc ? HIFN7901_CMD_ENCODE : HIFN7901_CMD_DECODE);
	hifn7901_iveccpy(&myvec, ivec);

	while (length > 0) {
		size = (length > HIFN7901_MAXBUF) ? HIFN7901_MAXBUF : length;

		hifn7901_sendcmd(input, size, size, cmd,
			(HIFN7901_ENC_DES | HIFN7901_ENC_CBC), 0,
			schedule, 8, &myvec);
		hifn7901_pollcmd();
		hifn7901_getcmd(input, size, output, size, &myvec, enc);

		input = (des_cblock *) (((void *) input) + size);
		output = (des_cblock *) (((void *) output) + size);
		length -= size;
	}

	return(0);
}

/****************************************************************************/

/*
 *	As per standard DES library functions.
 */

int hifn7901_des_ede3_cbc_encrypt(des_cblock *input, des_cblock *output,
	long length, des_key_schedule ks1, des_key_schedule ks2,
	des_key_schedule ks3, des_cblock *ivec, int enc)
{
	unsigned char		key[24];
	long			size;
	int			cmd;

#if 0
	printk("des_ede3_cbc_encrypt(input=%x,output=%x,length=%d,"
		"ks1=%x,ks2=%x,ks3=%x,ivec=%x,encrypt=%d)\n",
		(int) input, (int) output, (int) length,
		(int) ks1, (int) ks2, (int) ks3, (int) ivec, enc);
#endif

	cmd = HIFN7901_CMD_ENCRYPT |
		(enc ? HIFN7901_CMD_ENCODE : HIFN7901_CMD_DECODE);

	hifn7901_keycpy(&key[0], ks1);
	hifn7901_keycpy(&key[8], ks2);
	hifn7901_keycpy(&key[16], ks3);

	while (length > 0) {
		size = (length > HIFN7901_MAXBUF) ? HIFN7901_MAXBUF : length;

		hifn7901_sendcmd(input, size, size, cmd,
			(HIFN7901_ENC_3DES | HIFN7901_ENC_CBC), 0,
			&key[0], 24, ivec);
		hifn7901_pollcmd();
		hifn7901_getcmd(input, size, output, size, ivec, enc);

		input = (des_cblock *) (((void *) input) + size);
		output = (des_cblock *) (((void *) output) + size);
		length -= size;
	}

	return(0);
}

/****************************************************************************/

/*
 *	LZS compression support. I am not at all sure how to deal with
 *	data buffers that are larger than what the HIFN can compress/
 *	decompress in one go. The end marker system of LZS will ignore
 *	any extra data...
 */

int hifn7901_lzs_compress(unsigned char *input, int inlength,
	unsigned char *output, int outlength, int comp)
{
	unsigned int	cmd;

#if 0
	printk("hifn7901_lzs_comrpess(input=%x,inlength=%d,output=%x,"
		"outlength=%d,comp=%d)\n", (int) input, inlength,
		(int) output, (int) outlength, (int) comp);
#endif

	if ((inlength < 0) || (inlength > HIFN7901_MAXBUF))
		return(-1);

	cmd = HIFN7901_CMD_COMP |
                (comp ? HIFN7901_CMD_ENCODE : HIFN7901_CMD_DECODE);

	hifn7901_sendcmd(input, inlength, outlength, cmd, 0,
		(HIFN7901_COMP_LZS | HIFN7901_COMP_CLEARHIST),
		NULL, 0, NULL);
	hifn7901_pollcmd();
	outlength = hifn7901_getcmd(input, inlength, output, outlength,
		NULL, 0);

	return(outlength);
}

/****************************************************************************/
#ifdef HIFN7901_TEST
/****************************************************************************/

#define	HIFN7901_MAXDUMP	64

void hifn7901_dumpdata(char *str, des_cblock *k1p, des_cblock *k2p,
	des_cblock *k3p, des_cblock *ivec, unsigned char *dp, int len)
{
	unsigned char	*sp;
	int		i, orglen;

	orglen = len;
	printk("\n");

	if (k1p) {
		sp = (unsigned char *) k1p;
		printk("%s KEY[1]      ->  %02x %02x %02x %02x %02x %02x "
			"%02x %02x\n", str, sp[0], sp[1], sp[2], sp[3],
			sp[4], sp[5], sp[6], sp[7]);
	}
	if (k2p) {
		sp = (unsigned char *) k2p;
		printk("%s KEY[2]      ->  %02x %02x %02x %02x %02x %02x "
			"%02x %02x\n", str, sp[0], sp[1], sp[2], sp[3],
			sp[4], sp[5], sp[6], sp[7]);
	}
	if (k3p) {
		sp = (unsigned char *) k3p;
		printk("%s KEY[3]      ->  %02x %02x %02x %02x %02x %02x "
			"%02x %02x\n", str, sp[0], sp[1], sp[2], sp[3],
			sp[4], sp[5], sp[6], sp[7]);
	}

	if (ivec) {
		sp = (unsigned char *) ivec;
		printk("%s IVEC        ->  %02x %02x %02x %02x %02x %02x "
			"%02x %02x\n", str, sp[0], sp[1], sp[2], sp[3],
			sp[4], sp[5], sp[6], sp[7]);
	}

	if (len > HIFN7901_MAXDUMP) {
		dp += (len - HIFN7901_MAXDUMP);
		len = HIFN7901_MAXDUMP;
	}

	printk("%s DATA (%04d) ->  ", str, orglen);
	for (i = 0; (i < len); ) {
		printk("%02x ", dp[i++]);
		if ((i % 16) == 0)
			printk("\n                     ");
	}

	printk("\n");
	if (i % 16)
		printk("\n");
}

/****************************************************************************/

/*
 *	Some Hifn7901 test code...
 */

#define	TESTSIZE	8

unsigned char		ibuf[TESTSIZE];
unsigned char		obuf1[TESTSIZE], obuf2[TESTSIZE];
des_key_schedule	sched;
des_cblock		key  = { 0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17 };
des_cblock		ivec = { 0x08,0x09,0x0a,0x0b,0x0c,0x0d,0x0e,0x0f };

void hifn7901_test(void)
{
	int	i;

	for (i = 0; (i < TESTSIZE); i++)
		ibuf[i] = (unsigned char) i;

	hifn7901_dumpdata("DES ", &key, NULL, NULL, &ivec, &ibuf[0], TESTSIZE);

	printk("-----------------------------------------------------------\n");
	i = lib_des_set_key(&key, sched);
	printk("%s(%d): des_set_key() returned %d\n", __FILE__, __LINE__, i);
	lib_des_cbc_encrypt((des_cblock *) &ibuf[0],
		(des_cblock *) &obuf1[0], TESTSIZE, sched, &ivec, 1);

	hifn7901_dumpdata("LIB ", &key, NULL, NULL, &ivec, &obuf1[0], TESTSIZE);

	printk("-----------------------------------------------------------\n");

	i = hifn7901_des_set_key(&key, sched);
	printk("%s(%d): des_set_key() returned %d\n", __FILE__, __LINE__, i);
	hifn7901_des_cbc_encrypt((des_cblock *) &ibuf[0],
		(des_cblock *) &obuf2[0], TESTSIZE, sched, &ivec, 1);

	hifn7901_dumpdata("HIFN", &key, NULL, NULL, &ivec, &obuf2[0], TESTSIZE);
	printk("-----------------------------------------------------------\n");
#if 1
	hifn7901_des_cbc_encrypt((des_cblock *) &obuf2[0],
		(des_cblock *) &obuf1[0], TESTSIZE, sched, &ivec, 0);

	hifn7901_dumpdata("HIFN", &key, NULL, NULL, &ivec, &obuf1[0], TESTSIZE);
	printk("-----------------------------------------------------------\n");
#endif
}

/****************************************************************************/

/*
 *	Some Hifn7901 3DES test code...
 */

#define	TESTSIZE3	32

unsigned char		i3buf[TESTSIZE3];
unsigned char		o3buf1[TESTSIZE3], o3buf2[TESTSIZE3];
des_key_schedule	sched1, sched2, sched3;
des_cblock		key1  = { 0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17 };
des_cblock		key2  = { 0x18,0x19,0x1a,0x1b,0x1c,0x1d,0x1e,0x1f };
des_cblock		key3  = { 0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27 };
des_cblock		ivec1 = { 0x08,0x09,0x0a,0x0b,0x0c,0x0d,0x0e,0x0f };

void hifn7901_test3des(void)
{
	des_cblock	iv;
	int		i, j, k;

	for (i = 0; (i < TESTSIZE3); i++)
		i3buf[i] = (unsigned char) i;

	hifn7901_dumpdata("3DES", &key1, &key2, &key3, &ivec1, &i3buf[0],
		TESTSIZE3);

	printk("-----------------------------------------------------------\n");
	i = lib_des_set_key(&key1, sched1);
	j = lib_des_set_key(&key2, sched2);
	k = lib_des_set_key(&key3, sched3);
	printk("%s(%d): des_set_key() returned %d,%d,%d\n",
		__FILE__, __LINE__, i, j, k);

	memcpy(&iv, &ivec1, sizeof(iv));
	lib_des_ede3_cbc_encrypt((des_cblock *) &i3buf[0],
		(des_cblock *) &o3buf1[0], TESTSIZE3,
		sched1, sched2, sched3, &iv, 1);

	hifn7901_dumpdata("LIB ", &key1, &key2, &key3, &iv, &o3buf1[0],
		 TESTSIZE3);

	printk("-----------------------------------------------------------\n");

	i = hifn7901_des_set_key(&key1, sched1);
	j = hifn7901_des_set_key(&key2, sched2);
	k = hifn7901_des_set_key(&key3, sched3);
	printk("%s(%d): des_set_key() returned %d,%d,%d\n",
		__FILE__, __LINE__, i, j, k);

	memcpy(&iv, &ivec1, sizeof(iv));
	hifn7901_des_ede3_cbc_encrypt((des_cblock *) &i3buf[0],
		(des_cblock *) &o3buf2[0], TESTSIZE3,
		sched1, sched2, sched3, &iv, 1);

	hifn7901_dumpdata("HIFN", &key1, &key2, &key3, &iv, &o3buf2[0],
		TESTSIZE3);
	printk("-----------------------------------------------------------\n");
#if 1
	memcpy(&iv, &ivec1, sizeof(iv));
	hifn7901_des_ede3_cbc_encrypt((des_cblock *) &o3buf2[0],
		(des_cblock *) &o3buf1[0], TESTSIZE3,
		sched1, sched2, sched3, &iv, 0);

	hifn7901_dumpdata("HIFN", &key1, &key2, &key3, &iv, &o3buf1[0],
		 TESTSIZE3);
	printk("-----------------------------------------------------------\n");
#endif
}

/****************************************************************************/

/*
 *	Some Hifn7901 compression test code...
 */

#define	CTESTSIZE	512

unsigned char		cibuf[CTESTSIZE];
unsigned char		cobuf[CTESTSIZE];


void hifn7901_compresstest(void)
{
	int	i, j, size;

	for (i = j = 0; (i < CTESTSIZE); i += 4, j++)
		*((unsigned int *) &cibuf[i]) = (j & 0xf);

	hifn7901_dumpdata("STRT", NULL, NULL, NULL, NULL, &cibuf[0], CTESTSIZE);

	printk("-----------------------------------------------------------\n");

	size = hifn7901_lzs_compress(&cibuf[0], CTESTSIZE, &cobuf[0], CTESTSIZE, 1);

	hifn7901_dumpdata("LZS ", NULL, NULL, NULL, NULL, &cobuf[0], size);
	printk("-----------------------------------------------------------\n");

#if 1
	memset(&cibuf[0], 0, CTESTSIZE);
	size = hifn7901_lzs_compress(&cobuf[0], size, &cibuf[0], CTESTSIZE, 0);
	hifn7901_dumpdata("BACK", NULL, NULL, NULL, NULL, &cibuf[0], size);
	printk("-----------------------------------------------------------\n");
#endif
}

/****************************************************************************/

/*
 *	Some Hifn7901 3DES benchmark code.
 */

#define	BENCHSIZE	1024

unsigned char		benchbuf[BENCHSIZE];


void hifn7901_bench3des(void)
{
	des_cblock	iv;
	unsigned long	startjiffies, endjiffies, total;
	int		i, j, k, iterations;

	iterations = 20000;

	printk("-----------------------------------------------------------\n");

	for (i = 0; (i < BENCHSIZE); i++)
		benchbuf[i] = (unsigned char) i;

	printk("%s(%d): BENCHMARK SOFTWARE 3DES...\n", __FILE__, __LINE__);

	i = lib_des_set_key(&key1, sched1);
	j = lib_des_set_key(&key2, sched2);
	k = lib_des_set_key(&key3, sched3);
	memcpy(&iv, &ivec1, sizeof(iv));

	startjiffies = jiffies;

	for (i = 0; (i < iterations); i++) {
		lib_des_ede3_cbc_encrypt((des_cblock *) &benchbuf[0],
			(des_cblock *) &benchbuf[0], BENCHSIZE,
			sched1, sched2, sched3, &iv, 1);
	}

	endjiffies = jiffies;
	total = endjiffies - startjiffies;
	if (total == 0) total = 1;	/* HACK, just in case 0 */

	printk("%s(%d): TIME  start=%d  end=%d  total=%d jiffies\n",
		__FILE__, __LINE__, (int) startjiffies, (int) endjiffies,
		(int) total);
	printk("%s(%d): RATE  iterations=%d  time=%d s  rate=%d bytes/s\n",
		__FILE__, __LINE__, iterations, (int) (total / 100),
		(int) (((BENCHSIZE * iterations) / total) * 100));

	printk("%s(%d): DATA  ", __FILE__, __LINE__);
	for (i = 16; (i > 0); i--)
		printk("%02x ", (int) benchbuf[BENCHSIZE - i]);
	printk("\n");

	printk("-----------------------------------------------------------\n");

	for (i = 0; (i < BENCHSIZE); i++)
		benchbuf[i] = (unsigned char) i;

	printk("%s(%d): BENCHMARK HIFN 3DES...\n", __FILE__, __LINE__);

	i = hifn7901_des_set_key(&key1, sched1);
	j = hifn7901_des_set_key(&key2, sched2);
	k = hifn7901_des_set_key(&key3, sched3);
	memcpy(&iv, &ivec1, sizeof(iv));

	startjiffies = jiffies;

	for (i = 0; (i < iterations); i++) {
		hifn7901_des_ede3_cbc_encrypt((des_cblock *) &benchbuf[0],
			(des_cblock *) &benchbuf[0], BENCHSIZE,
			sched1, sched2, sched3, &iv, 1);
	}

	endjiffies = jiffies;
	total = endjiffies - startjiffies;
	if (total == 0) total = 1;	/* HACK, just in case 0 */

	printk("%s(%d): TIME  start=%d  end=%d  total=%d jiffies\n",
		__FILE__, __LINE__, (int) startjiffies, (int) endjiffies,
		(int) total);
	printk("%s(%d): RATE  iterations=%d  time=%d s  rate=%d bytes/s\n",
		__FILE__, __LINE__, iterations, (int) (total / 100),
		(int) (((BENCHSIZE * iterations) / total) * 100));

	printk("%s(%d): DATA  ", __FILE__, __LINE__);
	for (i = 16; (i > 0); i--)
		printk("%02x ", (int) benchbuf[BENCHSIZE - i]);
	printk("\n");

	printk("-----------------------------------------------------------\n");
}

/****************************************************************************/

#if 0
/*
 *	Perform Hifn's own diagnostic tests.
 */

void hifn7901_diags(void)
{
	int	rc;

	printk("%s(%d): Reset test     ->  ", __FILE__, __LINE__);
	rc = diag7901_resetTest(hifn7901_memaddr);
	if (rc)
		printk("FAILED, rc=%d\n", rc);
	else
		printk("PASSED\n");

	printk("%s(%d): Register Test  ->  ", __FILE__, __LINE__);
	rc = diag7901_registerTest(hifn7901_memaddr);
	if (rc)
		printk("FAILED, rc=%d\n", rc);
	else
		printk("PASSED\n");

	printk("%s(%d): IPM Test       ->  ", __FILE__, __LINE__);
	rc = diag7901_ipmTest(hifn7901_memaddr);
	if (rc)
		printk("FAILED, rc=%d\n", rc);
	else
		printk("PASSED\n");

	printk("%s(%d): SRAM Test      ->  ", __FILE__, __LINE__);
	rc = diag7901_emTest(hifn7901_memaddr);
	if (rc)
		printk("FAILED, rc=%d\n", rc);
	else
		printk("PASSED\n");

	printk("%s(%d): RNG Test       ->  ", __FILE__, __LINE__);
	rc = diag7901_rngOscillators(hifn7901_memaddr);
	if (rc)
		printk("FAILED, rc=%d\n", rc);
	else
		printk("PASSED\n");
}
#endif

/****************************************************************************/
#endif	/* HIFN7901_TEST */
/****************************************************************************/

void hifn7901_init(void)
{
	volatile unsigned long	*gcrp;
	unsigned long		id;
	int			i;

#if 0
	hifn7901_diags();
#endif

	id = *((volatile unsigned long *) (hifn7901_memaddr + HIFN7901_ID));
	if (id == HIFN7901_ID_PRODREV)
		hifn7901_present = 1;

	printk("HIFN7901: cryptographic accelerator ");
	if (hifn7901_present)
		printk("PRESENT, ID=%08x\n", (int) id);
	else
		printk("NOT present\n");
	printk("HIFN7901: Copyright (C) 2000, Lineo (www.lineo.com)\n");

	if (hifn7901_present) {
		/* Reset the hifn chip */
		gcrp = (volatile unsigned long *)
			(hifn7901_memaddr + HIFN7901_GC);
		*gcrp = HIFN7901_GC_RESET;

		for (i = 0; (i < 0x100000); i++) {
			if ((*gcrp & HIFN7901_GC_RESET) == 0)
				break;
		}

		/* Clear out memory, good for debugging */
		hifn7901_memset((void *) (hifn7901_memaddr + HIFN7901_MEMBASE),
			0, HIFN7901_MEMSIZE);
	}

#if 0
	hifn7901_test();
#endif
#if 0
	hifn7901_test3des();
#endif
#if 0
	hifn7901_compresstest();
#endif
#if 0
	hifn7901_bench3des();
#endif
}

/****************************************************************************/
