/*
 *  Linux/drivers/char/serial.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 *
 *  Extensively rewritten by Theodore Ts'o, 8/16/92 -- 9/14/92.  Now
 *  much more extensible to support other serial cards based on the
 *  16450/16550A UART's.  Added support for the AST FourPort and the
 *  Accent Async board.  
 *
 *  set_serial_info fixed to set the flags, custom divisor, and uart
 * 	type fields.  Fix suggested by Michael K. Johnson 12/12/92.
 *
 *  11/95: TIOCMIWAIT, TIOCGICOUNT by Angelo Haritsis <ah@doc.ic.ac.uk>
 *
 *  03/96: Modularised by Angelo Haritsis <ah@doc.ic.ac.uk>
 *
 *  rs_set_termios fixed to look also for changes of the input
 *      flags INPCK, BRKINT, PARMRK, IGNPAR and IGNBRK.
 *                                            Bernd Anhupl 05/17/96.
 * 
 * Added Support for PCI serial boards which contain 16x50 Chips
 * 31.10.1998 Henning P. Schmiedehausen <hps@tanstaafl.de>
 * 
 * This module exports the following rs232 io functions:
 *
 *	int rs_init(void);
 * 	int rs_open(struct tty_struct * tty, struct file * filp)
 */

#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/serial.h>
//#include <linux/serial_reg.h>
#include <linux/config.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/mm.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/bitops.h>
#include <asm/serial.h>

#define SERIAL_DSC21_INLINE inline
//#define SERIAL_DSC21_INLINE

static char *serial_name = "Serial driver";
static char *serial_version = "4.13p1";

DECLARE_TASK_QUEUE(tq_serial);

struct tty_driver serial_driver, callout_driver;
static int serial_refcount;

/* serial subtype definitions */
#define SERIAL_TYPE_NORMAL	1
#define SERIAL_TYPE_CALLOUT	2

/* number of characters left in xmit buffer before we ask for more */
#define WAKEUP_CHARS 256

#ifdef CONFIG_COLDFIRE
#undef	TTY_MAJOR
#define	TTY_MAJOR	24
#undef	TTYAUX_MAJOR
#define	TTYAUX_MAJOR	25
#endif

/*
 * Serial driver configuration section.  Here are the various options:
 *
 * CONFIG_HUB6
 *		Enables support for the venerable Bell Technologies
 *		HUB6 card.
 *
 * SERIAL_PARANOIA_CHECK
 * 		Check the magic number for the dsc21_async_structure where
 * 		ever possible.
 */

#define SERIAL_PARANOIA_CHECK
#define CONFIG_SERIAL_NOPAUSE_IO
#define SERIAL_DO_RESTART

#undef SERIAL_DEBUG_INTR
#undef SERIAL_DEBUG_OPEN
#undef SERIAL_DEBUG_FLOW

#ifdef CONFIG_SERIAL_PCI
# undef SERIAL_DEBUG_PCI
#endif

#define RS_ISR_PASS_LIMIT 256

#define IRQ_T(info) ((info->flags & ASYNC_SHARE_IRQ) ? SA_SHIRQ : SA_INTERRUPT)

#if defined(MODULE) && defined(SERIAL_DEBUG_MCOUNT)
#define DBG_CNT(s) printk("(%s): [%x] refc=%d, serc=%d, ttyc=%d -> %s\n", \
 kdevname(tty->device), (info->flags), serial_refcount,info->count,tty->count,s)
#else
#define DBG_CNT(s)
#endif

static void autoconfig(struct dsc21_async_struct * info);
static void change_speed(struct dsc21_async_struct *info);
	
/*
 * Well, it is not a 24,756 MHz clock but it is at least a start. 
 * This PCI board here has a 14,7456 MHz crystal oscillator which is 
 * eight times as fast as the standard serial clock...
 */

#define PCI_BAUD  ( 14745600 / 16 )

/* Standard COM flags (except for COM4, because of the 8514 problem) */
#define STD_COM4_FLAGS ASYNC_BOOT_AUTOCONF

#define FOURPORT_FLAGS ASYNC_FOURPORT
#define ACCENT_FLAGS 0
#define BOCA_FLAGS 0
#define HUB6_FLAGS 0
	
#ifdef CONFIG_SERIAL_PCI

#define PCI_FLAGS (ASYNC_PCI|ASYNC_BOOT_AUTOCONF)

#ifndef PCI_DEVICE_ID_PLX_SPCOM200
#define PCI_DEVICE_ID_PLX_SPCOM200 0x1103
#endif

/*
 * The chips we know about
 */

#define PCISER_PLX9050  0    /* PLX 9050 local bus bridge as serial card */
#define PCISER_PCCOM4   1    /* "PC COM PCI Bus 4 port serial Adapter" -- from Alvin Sim <alvin@alloycp.com.au> */

struct pci_serial_boards pci_serial_tbl[] = {
  { PCI_VENDOR_ID_PLX, PCI_DEVICE_ID_PLX_SPCOM200, "SPCom 200", PCISER_PLX9050, pci_space_0|pci_space_1, 1, 0, 128, PCI_BAUD },
  { PCI_VENDOR_ID_DCI, PCI_DEVICE_ID_DCI_PCCOM4,   "PC COM 4",  PCISER_PCCOM4,  pci_space_0,             4, 8, 128, BASE_BAUD },
  { 0x10b5, 0x9050, "RAStel", PCISER_PLX9050,  pci_space_0, 4, 8, 128, BASE_BAUD },
  { 0x15aa, 0x2000, "RAStel", PCISER_PLX9050,  pci_space_0|pci_space_1, 1, 0, 128, BASE_BAUD },
  {	0, 0, 0, 0, 0, 0, 0, 0, 0 }
};

#endif
	
/*
 * The following define the access methods for the HUB6 card. All
 * access is through two ports for all 24 possible chips. The card is
 * selected through the high 2 bits, the port on that card with the
 * "middle" 3 bits, and the register on that port with the bottom
 * 3 bits.
 *
 * While the access port and interrupt is configurable, the default
 * port locations are 0x302 for the port control register, and 0x303
 * for the data read/write register. Normally, the interrupt is at irq3
 * but can be anything from 3 to 7 inclusive. Note that using 3 will
 * require disabling com2.
 */

#define C_P(card,port) (((card)<<6|(port)<<3) + 1)

struct dsc21_async_struct rs_table[] = {
	STD_SERIAL_PORT_DEFNS
};

#define NR_PORTS	(sizeof(rs_table)/sizeof(struct dsc21_async_struct))

#ifdef CONFIG_SERIAL_PCI

/*
 * currently you can have up to four PCI serial boards in your
 * system. Increase the size of this structure to have more
 */

struct pci_struct pci_rs_chips[] = {                 
  {0, 0,},
  {0, 0,},
  {0, 0,},
  {0, 0,},
};

#define PCI_NR_BOARDS	(sizeof(pci_rs_chips)/sizeof(struct pci_struct))
#define PCI_NR_PORTS	8

#define PCI_PORT_START (NR_PORTS - PCI_NR_PORTS)

#endif

static struct tty_struct *serial_table[NR_PORTS];
static struct termios *serial_termios[NR_PORTS];
static struct termios *serial_termios_locked[NR_PORTS];

#ifndef MIN
#define MIN(a,b)	((a) < (b) ? (a) : (b))
#endif

/*
 * tmp_buf is used as a temporary buffer by serial_write.  We need to
 * lock it in case the memcpy_fromfs blocks while swapping in a page,
 * and some other program tries to do a serial write at the same time.
 * Since the lock will only come under contention when the system is
 * swapping and available memory is low, it makes sense to share one
 * buffer across all the serial ports, since it significantly saves
 * memory if large numbers of serial ports are open.
 */
static unsigned char *tmp_buf = 0;
static struct semaphore tmp_buf_sem = MUTEX;

static inline int serial_paranoia_check(struct dsc21_async_struct *info,
					kdev_t device, const char *routine)
{
#ifdef SERIAL_PARANOIA_CHECK
	static const char * const badmagic =
		"Warning: bad magic number for serial struct (%s) in %s\n";
	static const char * const badinfo =
		"Warning: null dsc21_async_struct for (%s) in %s\n";

	if (!info) {
		printk(badinfo, kdevname(device), routine);
		return 1;
	}
	if (info->magic != SERIAL_MAGIC) {
		printk(badmagic, kdevname(device), routine);
		return 1;
	}
#endif
	return 0;
}

/*
 * This is used to figure out the divisor speeds and the timeouts
 */
static int baud_table[] = {
	0, 50, 75, 110, 134, 150, 200, 300, 600, 1200, 1800, 2400, 4800,
	9600, 19200, 38400, 57600, 115200, 0 };

static SERIAL_DSC21_INLINE unsigned int serial_in(struct dsc21_async_struct *info, 
					 int offset)
{
	return inw((u32)(info->iomem_base + (offset<<info->iomem_reg_shift)));
}

static SERIAL_DSC21_INLINE void serial_out(struct dsc21_async_struct *info, int offset,
				  int value)
{
	outw(value, (unsigned long) info->iomem_base +
	     (offset<<info->iomem_reg_shift));
}
static SERIAL_DSC21_INLINE void serial_dsc21_disable_tx_int(struct dsc21_async_struct *info)
{
	info->MSR &= ~UART_MSR_TFTIE;
	serial_out(info, UART_MSR, info->MSR);
}

static SERIAL_DSC21_INLINE void serial_dsc21_disable_rx_int(struct dsc21_async_struct *info)
{
	info->MSR &= ~UART_MSR_RFTIE;
	serial_out(info, UART_MSR, info->MSR);
}

static SERIAL_DSC21_INLINE void serial_dsc21_enable_tx_int(struct dsc21_async_struct *info)
{
	info->MSR |= UART_MSR_TFTIE;
	serial_out(info, UART_MSR, info->MSR);
}

static SERIAL_DSC21_INLINE void serial_dsc21_enable_rx_int(struct dsc21_async_struct *info)
{
	info->MSR |= UART_MSR_RFTIE;
	serial_out(info, UART_MSR, info->MSR);
}

static SERIAL_DSC21_INLINE void serial_dsc21_disable_ints(struct dsc21_async_struct *info,
					     u16 *msr)
{
	*msr = info->MSR & UART_MSR_INTS;
	info->MSR &= ~UART_MSR_INTS;
	serial_out(info, UART_MSR, info->MSR);
}
static SERIAL_DSC21_INLINE void serial_dsc21_clear_fifos(struct dsc21_async_struct *info)
{
	serial_out(info, UART_RFCR, 0x8000);
	serial_out(info, UART_TFCR, 0x8000);
}
static SERIAL_DSC21_INLINE void serial_dsc21_reset_rx_block(struct dsc21_async_struct *info)
{
	serial_out(info, UART_RFCR, 0x4000);
}
static SERIAL_DSC21_INLINE void serial_dsc21_disable_breaks(struct dsc21_async_struct *info)
{
	info->LCR &= ~UART_LCR_BOC;
	serial_out(info, UART_LCR, info->LCR);
}
static SERIAL_DSC21_INLINE void serial_dsc21_enable_breaks(struct dsc21_async_struct *info)
{
	info->LCR |= UART_LCR_BOC;
	serial_out(info, UART_LCR, info->LCR);
}
static SERIAL_DSC21_INLINE void 
serial_dsc21_set_rate(struct dsc21_async_struct *info, u16 rate)
{
	serial_out(info, UART_BRSR, rate);
}
static SERIAL_DSC21_INLINE void serial_dsc21_set_mode(struct dsc21_async_struct *info,
					 u16 mode)
{
	info->MSR &= ~UART_MSR_MODE_BITS;
	info->MSR |= mode;
	serial_out(info, UART_MSR, info->MSR);
}
static SERIAL_DSC21_INLINE void serial_dsc21_set_rx_trigger(struct dsc21_async_struct *info,
					       u16 val)
{
	serial_out(info, UART_RFCR, val);
}
static SERIAL_DSC21_INLINE void serial_dsc21_set_tx_trigger(struct dsc21_async_struct *info,
					       u16 val)
{
	serial_out(info, UART_TFCR, val);
}
static SERIAL_DSC21_INLINE void serial_dsc21_out(
	struct dsc21_async_struct *info, u8 val)
{
	serial_out(info, UART_DTRR, val);
}
static SERIAL_DSC21_INLINE u8 serial_dsc21_in(struct dsc21_async_struct *info)
{
	return serial_in(info, UART_DTRR);
}
static SERIAL_DSC21_INLINE u8 serial_dsc21_tx_fifo_empty(struct dsc21_async_struct *info)
{
	return ((serial_in(info, UART_SR) & UART_SR_TREF) == 0);
}
static SERIAL_DSC21_INLINE u8 serial_dsc21_room_in_tx_fifo(
	struct dsc21_async_struct *info)
{
	return ((serial_in(info, UART_SR) & UART_SR_TFTI) != 0);
}
static SERIAL_DSC21_INLINE void serial_dsc21_save_registers(struct dsc21_async_struct *info)
{
	info->MSR = serial_in(info, UART_MSR);
	info->LCR = serial_in(info, UART_LCR);
}

#define serial_in_p(i,o) serial_in(i, o)
#define serial_out_p(i,o,v) serial_out(i, o, v)

/******************************************************************************
			    C o n s o l e
 ******************************************************************************/
#ifdef CONFIG_SERIAL_DSPLINUX_CONSOLE
static struct dsc21_async_struct *dsc21_sercons = &rs_table[0];
extern void register_console(void (*proc)(const char *));
static SERIAL_DSC21_INLINE void wait_for_xmitr(struct dsc21_async_struct *info)
{
	int tmp = 1000;
	while(!serial_dsc21_room_in_tx_fifo(info)) if (!--tmp) break;
}
static SERIAL_DSC21_INLINE void serial_dsc21_restore_ints(struct dsc21_async_struct *info,
					     u16 msr)
{
	info->MSR |= msr & UART_MSR_INTS;
	serial_out(info, UART_MSR, info->MSR);
}
void serial_dsc21_console_print(const char * b)
{
	struct dsc21_async_struct *info = dsc21_sercons;
	u16 ier;


	serial_dsc21_disable_ints(info, &ier);

	// HACK!
	// Somebody seems to be disrupting the LCR register causing
	// our bytes not to go out. Try this to see if it fixes it.
	// --gmcnutt
	serial_dsc21_disable_breaks(info);

	while(*b != 0) {
		wait_for_xmitr(info);
		serial_dsc21_out(info, *b);
		if (*b == 10) { // LF? add CR
			wait_for_xmitr(info);
			serial_dsc21_out(info, 13);
		}
		b++;
	}

	wait_for_xmitr(info);
	serial_dsc21_restore_ints(info, ier);
}
void serial_dsplinux_console_init(void)
{
	u16 junk;
	struct dsc21_async_struct *info = dsc21_sercons;

	serial_dsc21_disable_ints(info, &junk);
	serial_dsc21_clear_fifos(info);
	serial_dsc21_disable_breaks(info);
	serial_dsc21_set_rate(info, 0x13); // 115200
//	serial_dsc21_set_rate(info, 0xf0); // 9600
	serial_dsc21_set_mode(info, 0x00); // 8N1

	register_console(serial_dsc21_console_print);	
}
#endif /* CONFIG_SERIAL_DSPLINUX_CONSOLE ******************************************/

/*
 * ------------------------------------------------------------
 * rs_stop() and rs_start()
 *
 * This routines are called before setting or resetting tty->stopped.
 * They enable or disable transmitter interrupts, as necessary.
 * ------------------------------------------------------------
 */
static void rs_stop(struct tty_struct *tty)
{
	struct dsc21_async_struct *info = (struct dsc21_async_struct *)tty->driver_data;
	unsigned long flags;

	if (serial_paranoia_check(info, tty->device, "rs_stop"))
		return;
	
	save_flags(flags); cli();
	if (info->MSR & UART_MSR_TFTIE) {
		serial_dsc21_disable_tx_int(info);
	}
	restore_flags(flags);
}

static void rs_start(struct tty_struct *tty)
{
	struct dsc21_async_struct *info = (struct dsc21_async_struct *)tty->driver_data;
	unsigned long flags;
	
	if (serial_paranoia_check(info, tty->device, "rs_start"))
		return;
	
	save_flags(flags); cli();
	if (info->xmit_cnt && info->xmit_buf && 
	    !(info->MSR & UART_MSR_TFTIE)) {
		serial_dsc21_enable_tx_int(info);
	}
	restore_flags(flags);
}

/*
 * ----------------------------------------------------------------------
 *
 * Here starts the interrupt handling routines.  All of the following
 * subroutines are declared as inline and are folded into
 * rs_interrupt().  They were separated out for readability's sake.
 *
 * Note: rs_interrupt() is a "fast" interrupt, which means that it
 * runs with interrupts turned off.  People who may want to modify
 * rs_interrupt() should try to keep the interrupt handler as fast as
 * possible.  After you are done making modifications, it is not a bad
 * idea to do:
 * 
 * gcc -S -DKERNEL -Wall -Wstrict-prototypes -O6 -fomit-frame-pointer serial.c
 *
 * and look at the resulting assemble code in serial.s.
 *
 * 				- Ted Ts'o (tytso@mit.edu), 7-Mar-93
 * -----------------------------------------------------------------------
 */


/*
 * This routine is used by the interrupt handler to schedule
 * processing in the software interrupt portion of the driver.
 */
static SERIAL_DSC21_INLINE void rs_sched_event(struct dsc21_async_struct *info,
				  int event)
{
	info->event |= 1 << event;
	queue_task_irq_off(&info->tqueue, &tq_serial);
	mark_bh(SERIAL_BH);
}

static SERIAL_DSC21_INLINE void discard_bad_chars(struct dsc21_async_struct *info)
{
        int err = 0;
        
        while (serial_in(info, UART_SR) & UART_SR_RFER) {
                serial_in(info, UART_DTRR);
                err++;
        }

        printk(__FUNCTION__ ": %d error bytes\n", err);
}

static SERIAL_DSC21_INLINE void receive_chars(struct dsc21_async_struct *info, u16 *status)
{
	struct tty_struct *tty = info->tty;
	unsigned char ch;
	u16 dtrr;

	do {
		ch = dtrr = serial_in(info, UART_DTRR);

		if (tty->flip.count >= TTY_FLIPBUF_SIZE)
			break;
		tty->flip.count++;			
		if(*status & UART_SR_RFER)
		{
			if(dtrr & UART_DTRR_PEF)
				*tty->flip.flag_buf_ptr++ = TTY_PARITY;
			else if (dtrr & UART_DTRR_FEX)
				*tty->flip.flag_buf_ptr++ = TTY_FRAME;
			else if (dtrr & UART_DTRR_FEX)
				*tty->flip.flag_buf_ptr++ = TTY_OVERRUN;
		}
		else
			*tty->flip.flag_buf_ptr++ = 0;

		*tty->flip.char_buf_ptr++ = ch;
		*status = serial_in(info, UART_SR);
	} while ((*status & UART_SR_RFNEF) || (serial_in(info, UART_RFCR) & UART_RFCR_RWC_BITS));

#ifdef CONFIG_SERIAL_DSPLINUX_CONSOLE
	if(info == dsc21_sercons)
		/*
		 * We are the console. Tell tty_io.c that we received a
		 * character.
		 */
		wake_up(&keypress_wait);
#endif

	queue_task_irq_off(&tty->flip.tqueue, &tq_timer);
}

static SERIAL_DSC21_INLINE void transmit_chars(struct dsc21_async_struct *info)
{
	int count;
	
	if (info->x_char) {
		serial_dsc21_out(info, info->x_char);
		info->x_char = 0;
		return;
	}
	if ((info->xmit_cnt <= 0) || info->tty->stopped ||
	    info->tty->hw_stopped) {
		serial_dsc21_disable_tx_int(info);
		return;
	}
	

	// send while we still have data & room in the fifo
	count = 32 - (serial_in(info, UART_TFCR) & 0x3f);
	do {
			serial_dsc21_out(info, 
					 info->xmit_buf[info->xmit_tail]);
			info->xmit_tail = (info->xmit_tail + 1) & 
				(SERIAL_XMIT_SIZE-1);
			if (--info->xmit_cnt <= 0)
				goto done;
	} while(--count > 0);
	
 done:
	
	if (info->xmit_cnt < WAKEUP_CHARS)
		rs_sched_event(info, RS_EVENT_WRITE_WAKEUP);

	if (info->xmit_cnt <= 0)
		serial_dsc21_disable_tx_int(info);
}


static int unsafe_to_exit(struct dsc21_async_struct *info, u16 *sr, u16 *msr)
{
        /*
         * Assumption: hardware will not post another interrupt until all
         * relevant MSR & SR bits logically AND to zero. This routine 
         * gaurantees that we do not exit the ISR until we are certain we have
         * met this condition at least once during ISR execution.
         */
        /*
         * Note: I don't know if this routine is really necessary or not. I 
         *      added it to try and fix the problems we were seeing with
         *      gdbserver, wherein we would stop getting UART interrupts.
         *      However, I cannot reproduce the symptoms now even without
         *      this routine -- probably because the version of gdbserver
         *      I'm using now is compiled without printf debugging. This
         *      routine is probably healthy for the UART ISR but may be
         *      damaging to overall system performance if it causes us
         *      to spend too much time in the UART ISR. printk's revealed
         *      that while using '(gdb) x /128 0x8e00000' style commands
         *      we would routinely go through this code twice per ISR. Twice
         *      is probably not too bad. Five or six would not be good...
         */
       // u16 msr, sr;

        *msr = serial_in(info, UART_MSR);
        *sr = serial_in(info, UART_SR);

        /* Check the rx fifo trigger. */
        if ((*msr & UART_MSR_RFTIE) && (*sr & UART_SR_RFTI))
                return 1;

        /* Check the tx fifo trigger. */
        if ((*msr & UART_MSR_TFTIE) && (*sr & UART_SR_TFTI))
                return 1;

        /* Check the rx error. */
        if ((*msr & UART_MSR_RDEIE) && (*sr & UART_SR_RFER))
                return 1;

        /* Check the timeout. */
        if ((*msr & UART_MSR_TOIC) && (*sr & UART_SR_TOIF))
                return 1;

        /* Ok -- should be safe to exit. */
        return 0;
}
/*
 * This is the serial driver's interrupt routine for a single port
 */
static void rs_interrupt_single(int irq, void *dev_id, struct pt_regs * regs)
{
	u16 sr = 0, msr = 0 ;
	int count = 0;
	struct dsc21_async_struct * info;

	info = (struct dsc21_async_struct *)dev_id;

	if (!info || !info->tty) {
        printk(__FUNCTION__ ": bad port or tty state\n");
		return;
	}

	do {
		if (count++ > 100)
			panic("Stuck in " __FUNCTION__ "\n");

		sr = serial_in(info, UART_SR);

		//if (sr & UART_SR_RFER) discard_bad_chars(info);
		if (sr & UART_SR_RFTI) receive_chars(info, &sr);
		if (sr & UART_SR_TFTI) transmit_chars(info);

	} while(unsafe_to_exit(info, &sr, &msr));
}
/*
 * -------------------------------------------------------------------
 * Here ends the serial interrupt routines.
 * -------------------------------------------------------------------
 */

/*
 * This routine is used to handle the "bottom half" processing for the
 * serial driver, known also the "software interrupt" processing.
 * This processing is done at the kernel interrupt level, after the
 * rs_interrupt() has returned, BUT WITH INTERRUPTS TURNED ON.  This
 * is where time-consuming activities which can not be done in the
 * interrupt driver proper are done; the interrupt driver schedules
 * them using rs_sched_event(), and they get done here.
 */
static void do_serial_bh(void)
{
	run_task_queue(&tq_serial);
}

static void do_softint(void *private_)
{
	struct dsc21_async_struct	*info = (struct dsc21_async_struct *) private_;
	struct tty_struct	*tty;
	
	tty = info->tty;
	if (!tty)
		return;

	if (clear_bit(RS_EVENT_WRITE_WAKEUP, &info->event)) {
		if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
		    tty->ldisc.write_wakeup)
			(tty->ldisc.write_wakeup)(tty);
		wake_up_interruptible(&tty->write_wait);
	}
}

/*
 * This routine is called from the scheduler tqueue when the interrupt
 * routine has signalled that a hangup has occurred.  The path of
 * hangup processing is:
 *
 * 	serial interrupt routine -> (scheduler tqueue) ->
 * 	do_serial_hangup() -> tty->hangup() -> rs_hangup()
 * 
 */
static void do_serial_hangup(void *private_)
{
	struct dsc21_async_struct	*info = (struct dsc21_async_struct *) private_;
	struct tty_struct	*tty;
	
	tty = info->tty;
	if (!tty)
		return;

	tty_hangup(tty);
}

static int startup(struct dsc21_async_struct * info)
{

	unsigned long flags;
	int	retval;

	unsigned long page;
	u16 junk;

	page = get_free_page(GFP_KERNEL);
	if (!page)
		return -ENOMEM;

	
	save_flags(flags); cli();

	if (info->flags & ASYNC_INITIALIZED) {
		free_page(page);
		restore_flags(flags);
		return 0;
	}

	if (info->xmit_buf)
		free_page(page);
	else
		info->xmit_buf = (unsigned char *) page;

#ifdef SERIAL_DEBUG_OPEN
	printk("starting up ttys%d (irq %d)...", info->line, info->irq);
#endif

	serial_dsc21_disable_ints(info, &junk);
	serial_dsc21_clear_fifos(info);
	serial_dsc21_reset_rx_block(info);
	serial_dsc21_disable_breaks(info);	

	retval = request_irq(info->irq, rs_interrupt_single, 0, "serial", info);
	if (retval) {
		restore_flags(flags);
		if (suser()) {
			if (info->tty)
				set_bit(TTY_IO_ERROR,
					&info->tty->flags);
			return 0;
		} else
			return retval;
	}
	
	/*
	 * Finally, enable interrupts
	 */
	serial_dsc21_enable_rx_int(info);

	if (info->tty)
		clear_bit(TTY_IO_ERROR, &info->tty->flags);
	info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;


	/*
	 * and set the speed of the serial port
	 */
	change_speed(info);

	info->flags |= ASYNC_INITIALIZED;
	restore_flags(flags);
	return 0;
}

/*
 * This routine will shutdown a serial port; interrupts are disabled, and
 * DTR is dropped if the hangup on close termio flag is on.
 */
static void shutdown(struct dsc21_async_struct * info)
{
	unsigned long	flags;

	u16             msr;

	if (!(info->flags & ASYNC_INITIALIZED))
		return;

#ifdef SERIAL_DEBUG_OPEN
	printk("Shutting down serial port %d (irq %d)....", info->line,
	       info->irq);
#endif
	
	save_flags(flags); cli(); /* Disable interrupts */

	/*
	 * clear delta_msr_wait queue to avoid mem leaks: we may free the irq
	 * here so the queue might never be waken up
	 */
	wake_up_interruptible(&info->delta_msr_wait);
	
	free_irq(info->irq, info);

	if (info->xmit_buf) {
		free_page((unsigned long) info->xmit_buf);
		info->xmit_buf = 0;
	}

	serial_dsc21_disable_ints(info, &msr);

	if (info->tty)
		set_bit(TTY_IO_ERROR, &info->tty->flags);
	
	info->flags &= ~ASYNC_INITIALIZED;
	restore_flags(flags);
}

/*
 * This routine is called to set the UART divisor registers to match
 * the specified baud rate for a serial port.
 */
static int tty_get_baud_rate(struct tty_struct *tty)
{
	struct dsc21_async_struct * info = 
		(struct dsc21_async_struct *)tty->driver_data;
	unsigned int cflag, i;

	cflag = tty->termios->c_cflag;

	i = cflag & CBAUD;
	if (i & CBAUDEX) {
		i &= ~CBAUDEX;
		if (i < 1 || i > 2) 
			tty->termios->c_cflag &= ~CBAUDEX;
		else
			i += 15;
	}
	if (i == 15) {
		if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI)
			i += 1;
		if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI)
			i += 2;
	}
	return baud_table[i];
}

static void change_speed(struct dsc21_async_struct *info)
{
	int	quot = 0, baud_base, baud;
	unsigned cflag, cval = 0;
	int	bits;
	unsigned long	flags;

	if (!info->tty || !info->tty->termios)
		return;
	cflag = info->tty->termios->c_cflag;


	/* byte size and parity */
	switch (cflag & CSIZE) {
	case CS7: cval = 0x01; bits = 9; break;
	case CS8: cval = 0x00; bits = 10; break;
	default:  cval = 0x00; bits = 7; break;
	}

	if (cflag & CSTOPB) {
		cval |= 0x04;
		bits++;
	}
	if (cflag & PARENB) {
		cval |= UART_MSR_PEB;
		bits++;
	}
	if (cflag & PARODD)
		cval |= UART_MSR_PSB;

	/* Determine divisor based on baud rate */
	baud = tty_get_baud_rate(info->tty);
	if (!baud)
		baud = 9600;	/* B0 transition handled in rs_set_termios */
	baud_base = info->baud_base;
	if (baud == 38400 &&
	    ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_CUST))
		quot = info->custom_divisor;
	else {
		if (baud == 134)
			/* Special case since 134 is really 134.5 */
			quot = (2*baud_base / 269);
		else if (baud)
			quot = (baud_base / baud) - 1;
	}
	/* As a last resort, if the quotient is zero, default to 9600 bps */
	if (!quot)
		quot = (baud_base / 9600) - 1;
	
	info->timeout = ((info->xmit_fifo_size*HZ*bits*quot) / baud_base);
	info->timeout += HZ/50;		/* Add .02 seconds of slop */

	/*
	 * Set up parity check flag
	 */
#define RELEVANT_IFLAG(iflag) (iflag & (IGNBRK|BRKINT|IGNPAR|PARMRK|INPCK))


    save_flags(flags); cli();

	serial_dsc21_set_rate(info, quot);
	serial_dsc21_set_mode(info, cval);

	/*
	 * We should always set the receive FIFO trigger to the lowest value
	 * because we don't poll.
	 * --gmcnutt
	 */
	serial_dsc21_set_rx_trigger(info, UART_FCR_FTL1);

	restore_flags(flags);

}

static void rs_put_char(struct tty_struct *tty, unsigned char ch)
{
	struct dsc21_async_struct *info = (struct dsc21_async_struct *)tty->driver_data;
	unsigned long flags;

	if (serial_paranoia_check(info, tty->device, "rs_put_char"))
		return;

	if (!tty || !info->xmit_buf)
		return;

	save_flags(flags); cli();
	if (info->xmit_cnt >= SERIAL_XMIT_SIZE - 1) {
		restore_flags(flags);
		return;
	}

	info->xmit_buf[info->xmit_head++] = ch;
	info->xmit_head &= SERIAL_XMIT_SIZE-1;
	info->xmit_cnt++;
	restore_flags(flags);
}

static void rs_flush_chars(struct tty_struct *tty)
{
	struct dsc21_async_struct *info = (struct dsc21_async_struct *)tty->driver_data;
	unsigned long flags;
				
	if (serial_paranoia_check(info, tty->device, "rs_flush_chars"))
		return;

	if (info->xmit_cnt <= 0 || tty->stopped || tty->hw_stopped ||
	    !info->xmit_buf)
		return;

	save_flags(flags); cli();
	serial_dsc21_enable_tx_int(info);
	restore_flags(flags);
}

static int rs_write(struct tty_struct * tty, int from_user,
		    const unsigned char *buf, int count)
{
	int	c, total = 0;
	struct dsc21_async_struct *info = (struct dsc21_async_struct *)tty->driver_data;
	unsigned long flags;
				
	if (serial_paranoia_check(info, tty->device, "rs_write"))
		return 0;

	if (!tty || !info->xmit_buf || !tmp_buf)
		return 0;
	    
	if (from_user)
		down(&tmp_buf_sem);
	save_flags(flags);
	while (1) {
		cli();		
		c = MIN(count, MIN(SERIAL_XMIT_SIZE - info->xmit_cnt - 1,
				   SERIAL_XMIT_SIZE - info->xmit_head));
		if (c <= 0)
			break;

		if (from_user) {
			memcpy_fromfs(tmp_buf, buf, c);
			c = MIN(c, MIN(SERIAL_XMIT_SIZE - info->xmit_cnt - 1,
				       SERIAL_XMIT_SIZE - info->xmit_head));
			memcpy(info->xmit_buf + info->xmit_head, tmp_buf, c);
		} else
			memcpy(info->xmit_buf + info->xmit_head, buf, c);
		info->xmit_head = (info->xmit_head + c) & (SERIAL_XMIT_SIZE-1);
		info->xmit_cnt += c;
		restore_flags(flags);
		buf += c;
		count -= c;
		total += c;
	}
	if (from_user)
		up(&tmp_buf_sem);
	if (info->xmit_cnt && 
	    !tty->stopped &&
	    !tty->hw_stopped) {
		if (serial_dsc21_room_in_tx_fifo(info)) {
			transmit_chars(info);
		}
		serial_dsc21_enable_tx_int(info);
	}

	restore_flags(flags);
	return total;
}

static int rs_write_room(struct tty_struct *tty)
{
	struct dsc21_async_struct *info = (struct dsc21_async_struct *)tty->driver_data;
	int	ret;
				
	if (serial_paranoia_check(info, tty->device, "rs_write_room"))
		return 0;
	ret = SERIAL_XMIT_SIZE - info->xmit_cnt - 1;
	if (ret < 0)
		ret = 0;
	return ret;
}

static int rs_chars_in_buffer(struct tty_struct *tty)
{
	struct dsc21_async_struct *info = (struct dsc21_async_struct *)tty->driver_data;
				
	if (serial_paranoia_check(info, tty->device, "rs_chars_in_buffer"))
		return 0;
	return info->xmit_cnt;
}

static void rs_flush_buffer(struct tty_struct *tty)
{
	struct dsc21_async_struct *info = (struct dsc21_async_struct *)tty->driver_data;
				
	if (serial_paranoia_check(info, tty->device, "rs_flush_buffer"))
		return;
	{
		unsigned long flags;
		save_flags_cli(flags);
		info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
		restore_flags(flags);
	}
	wake_up_interruptible(&tty->write_wait);
	if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
	    tty->ldisc.write_wakeup)
		(tty->ldisc.write_wakeup)(tty);
}

/*
 * ------------------------------------------------------------
 * rs_throttle()
 * 
 * This routine is called by the upper-layer tty layer to signal that
 * incoming characters should be throttled.
 * ------------------------------------------------------------
 */
static void rs_throttle(struct tty_struct * tty)
{
	struct dsc21_async_struct *info = (struct dsc21_async_struct *)tty->driver_data;
#ifdef SERIAL_DEBUG_THROTTLE
	char	buf[64];
	
	printk("throttle %s: %d....\n", _tty_name(tty, buf),
	       tty->ldisc.chars_in_buffer(tty));
#endif

	if (serial_paranoia_check(info, tty->device, "rs_throttle"))
		return;
	
	if (I_IXOFF(tty))
		info->x_char = STOP_CHAR(tty);

}

static void rs_unthrottle(struct tty_struct * tty)
{
	struct dsc21_async_struct *info = (struct dsc21_async_struct *)tty->driver_data;
#ifdef SERIAL_DEBUG_THROTTLE
	char	buf[64];
	
	printk("unthrottle %s: %d....\n", _tty_name(tty, buf),
	       tty->ldisc.chars_in_buffer(tty));
#endif

	if (serial_paranoia_check(info, tty->device, "rs_unthrottle"))
		return;
	
	if (I_IXOFF(tty)) {
		if (info->x_char)
			info->x_char = 0;
		else
			info->x_char = START_CHAR(tty);
	}
}

/*
 * ------------------------------------------------------------
 * rs_ioctl() and friends
 * ------------------------------------------------------------
 */

static int get_serial_info(struct dsc21_async_struct * info,
			   struct serial_struct * retinfo)
{
	struct serial_struct tmp;
  
	if (!retinfo)
		return -EFAULT;
	memset(&tmp, 0, sizeof(tmp));
	tmp.line = info->line;
	tmp.port = info->port;
	tmp.irq = info->irq;
	tmp.flags = info->flags;
	tmp.baud_base = info->baud_base;
	tmp.close_delay = info->close_delay;
	tmp.closing_wait = info->closing_wait;
	tmp.custom_divisor = info->custom_divisor;
	tmp.hub6 = info->hub6;
	memcpy_tofs(retinfo,&tmp,sizeof(*retinfo));
	return 0;
}

static int set_serial_info(struct dsc21_async_struct * info,
			   struct serial_struct * new_info)
{
	struct serial_struct new_serial;
	struct dsc21_async_struct old_info;
	unsigned int		i,change_irq,change_port;
	int 			retval = 0;

	if (!new_info)
		return -EFAULT;
	memcpy_fromfs(&new_serial,new_info,sizeof(new_serial));
	old_info = *info;

	change_irq = new_serial.irq != info->irq;
	change_port = (new_serial.port != info->port) || (new_serial.hub6 != info->hub6);

	if (!suser()) {
		if (change_irq || change_port ||
		    (new_serial.baud_base != info->baud_base) ||
		    (new_serial.close_delay != info->close_delay) ||
		    ((new_serial.flags & ~ASYNC_USR_MASK) !=
		     (info->flags & ~ASYNC_USR_MASK)))
			return -EPERM;
		info->flags = ((info->flags & ~ASYNC_USR_MASK) |
			       (new_serial.flags & ASYNC_USR_MASK));
		info->custom_divisor = new_serial.custom_divisor;
		goto check_and_exit;
	}

#ifndef CONFIG_COLDFIRE
	if (new_serial.irq == 2)
		new_serial.irq = 9;

	if ((new_serial.irq > 15) || (new_serial.port > 0xffff))
		return -EINVAL;
#endif

	/* Make sure address is not already in use */
	for (i = 0 ; i < NR_PORTS; i++)
		if ((info != &rs_table[i]) &&
		    (rs_table[i].port == new_serial.port))
			return -EADDRINUSE;

	if ((change_port || change_irq) && (info->count > 1))
		return -EBUSY;

	/*
	 * OK, past this point, all the error checking has been done.
	 * At this point, we start making changes.....
	 */

	info->baud_base = new_serial.baud_base;
	info->flags = ((info->flags & ~ASYNC_FLAGS) |
			(new_serial.flags & ASYNC_FLAGS));
	info->custom_divisor = new_serial.custom_divisor;
	info->close_delay = new_serial.close_delay * HZ/100;
	info->closing_wait = new_serial.closing_wait * HZ/100;

	if (change_port || change_irq) {
		/*
		 * We need to shutdown the serial port at the old
		 * port/irq combination.
		 */
		shutdown(info);
		info->irq = new_serial.irq;
		info->port = new_serial.port;
		info->hub6 = new_serial.hub6;
	}

	
check_and_exit:
	if (info->flags & ASYNC_INITIALIZED) {
		if (((old_info.flags & ASYNC_SPD_MASK) !=
		     (info->flags & ASYNC_SPD_MASK)) ||
		    (old_info.custom_divisor != info->custom_divisor))
			change_speed(info);
	} else
		retval = startup(info);
	return retval;
}


static int do_autoconfig(struct dsc21_async_struct * info)
{
	int			retval;
	
	if (!suser())
		return -EPERM;
	
	if (info->count > 1)
		return -EBUSY;
	
	shutdown(info);

	{
		unsigned long flags;
		save_flags_cli(flags);
		autoconfig(info);
		restore_flags(flags);
	}

	retval = startup(info);
	if (retval)
		return retval;
	return 0;
}

static void send_break(	struct dsc21_async_struct * info, int duration)
{
	current->state = TASK_INTERRUPTIBLE;
	current->timeout = jiffies + duration;
	{
		unsigned long flags;
		save_flags_cli(flags);
		serial_dsc21_enable_breaks(info);
		schedule();
		serial_dsc21_disable_breaks(info);
		restore_flags(flags);
	}
}

static void rs_break(struct dsc21_async_struct * info, int break_state)
{
        unsigned long flags;
        
        save_flags_cli(flags);
        if (break_state == -1)
		serial_dsc21_enable_breaks(info);
        else
		serial_dsc21_disable_breaks(info);
        restore_flags(flags);
}

static int rs_ioctl(struct tty_struct *tty, struct file * file,
		    unsigned int cmd, unsigned long arg)
{
	int error;
	struct dsc21_async_struct * info = (struct dsc21_async_struct *)tty->driver_data;
	int retval;
	struct async_icount cnow;	/* kernel counter temps */
	struct serial_icounter_struct *p_cuser;	/* user space */

	if (serial_paranoia_check(info, tty->device, "rs_ioctl"))
		return -ENODEV;

	if ((cmd != TIOCGSERIAL) && (cmd != TIOCSSERIAL) &&
	    (cmd != TIOCSERCONFIG) && (cmd != TIOCSERGWILD)  &&
	    (cmd != TIOCSERSWILD) && (cmd != TIOCSERGSTRUCT) &&
	    (cmd != TIOCMIWAIT) && (cmd != TIOCGICOUNT)) {
		if (tty->flags & (1 << TTY_IO_ERROR))
		    return -EIO;
	}
	
	switch (cmd) {
                case TIOCSBRK:  /* Turn break on, unconditionally */
			retval = tty_check_change(tty);
			if (retval)
				return retval;
			tty_wait_until_sent(tty, 0);
			rs_break(info,-1);
			return 0;
                case TIOCCBRK:  /* Turn break off, unconditionally */
			retval = tty_check_change(tty);
			if (retval)
				return retval;
			tty_wait_until_sent(tty, 0);
			rs_break(info,0);
			return 0;
		case TCSBRK:	/* SVID version: non-zero arg --> no break */
			retval = tty_check_change(tty);
			if (retval)
				return retval;
			tty_wait_until_sent(tty, 0);
			if (!arg)
				send_break(info, HZ/4);	/* 1/4 second */
			return 0;
		case TCSBRKP:	/* support for POSIX tcsendbreak() */
			retval = tty_check_change(tty);
			if (retval)
				return retval;
			tty_wait_until_sent(tty, 0);
			send_break(info, arg ? arg*(HZ/10) : HZ/4);
			return 0;
		case TIOCGSOFTCAR:
			error = verify_area(VERIFY_WRITE, (void *) arg,sizeof(long));
			if (error)
				return error;
			put_fs_long(C_CLOCAL(tty) ? 1 : 0,
				    (unsigned long *) arg);
			return 0;
		case TIOCSSOFTCAR:
			error = verify_area(VERIFY_READ, (void *) arg,sizeof(long));
			if (error)
				return error;
			arg = get_fs_long((unsigned long *) arg);
			tty->termios->c_cflag =
				((tty->termios->c_cflag & ~CLOCAL) |
				 (arg ? CLOCAL : 0));
			return 0;
		case TIOCGSERIAL:
			error = verify_area(VERIFY_WRITE, (void *) arg,
						sizeof(struct serial_struct));
			if (error)
				return error;
			return get_serial_info(info,
					       (struct serial_struct *) arg);
		case TIOCSSERIAL:
			error = verify_area(VERIFY_READ, (void *) arg,
						sizeof(struct serial_struct));
			if (error)
				return error;
			return set_serial_info(info,
					       (struct serial_struct *) arg);
		case TIOCSERCONFIG:
			return do_autoconfig(info);

		case TIOCSERGSTRUCT:
			error = verify_area(VERIFY_WRITE, (void *) arg,
						sizeof(struct dsc21_async_struct));
			if (error)
				return error;
			memcpy_tofs((struct dsc21_async_struct *) arg,
				    info, sizeof(struct dsc21_async_struct));
			return 0;
			
		/* 
		 * Get counter of input serial line interrupts (DCD,RI,DSR,CTS)
		 * Return: write counters to the user passed counter struct
		 * NB: both 1->0 and 0->1 transitions are counted except for
		 *     RI where only 0->1 is counted.
		 */
		case TIOCGICOUNT:
			error = verify_area(VERIFY_WRITE, (void *) arg,
				sizeof(struct serial_icounter_struct));
			if (error)
				return error;
			{
				unsigned long flags;
				save_flags_cli(flags);
				cnow = info->icount;
				restore_flags(flags);
			}
			p_cuser = (struct serial_icounter_struct *) arg;
			put_user(cnow.cts, &p_cuser->cts);
			put_user(cnow.dsr, &p_cuser->dsr);
			put_user(cnow.rng, &p_cuser->rng);
			put_user(cnow.dcd, &p_cuser->dcd);
			return 0;

		default:
			return -ENOIOCTLCMD;
		}
	return 0;
}

static void rs_set_termios(struct tty_struct *tty, struct termios *old_termios)
{
	struct dsc21_async_struct *info = (struct dsc21_async_struct *)tty->driver_data;

	if (   (tty->termios->c_cflag == old_termios->c_cflag)
	    && (   RELEVANT_IFLAG(tty->termios->c_iflag) 
		== RELEVANT_IFLAG(old_termios->c_iflag)))
	  return;

	change_speed(info);

	if ((old_termios->c_cflag & CRTSCTS) &&
	    !(tty->termios->c_cflag & CRTSCTS)) {
		tty->hw_stopped = 0;
		rs_start(tty);
	}

}

/*
 * ------------------------------------------------------------
 * rs_close()
 * 
 * This routine is called when the serial port gets closed.  First, we
 * wait for the last remaining data to be sent.  Then, we unlink its
 * async structure from the interrupt chain if necessary, and we free
 * that IRQ if nothing is left in the chain.
 * ------------------------------------------------------------
 */
static void rs_close(struct tty_struct *tty, struct file * filp)
{
	struct dsc21_async_struct * info = (struct dsc21_async_struct *)tty->driver_data;
	unsigned long flags;
	unsigned long timeout;

	if (!info || serial_paranoia_check(info, tty->device, "rs_close"))
		return;
	
	save_flags_cli(flags);
	
	if (tty_hung_up_p(filp)) {
		DBG_CNT("before DEC-hung");
		MOD_DEC_USE_COUNT;
		restore_flags(flags);
		return;
	}
	
#ifdef SERIAL_DEBUG_OPEN
	printk("rs_close ttys%d, count = %d\n", info->line, info->count);
#endif
	if ((tty->count == 1) && (info->count != 1)) {
		/*
		 * Uh, oh.  tty->count is 1, which means that the tty
		 * structure will be freed.  Info->count should always
		 * be one in these conditions.  If it's greater than
		 * one, we've got real problems, since it means the
		 * serial port won't be shutdown.
		 */
		printk("rs_close: bad serial port count; tty->count is 1, "
		       "info->count is %d\n", info->count);
		info->count = 1;
	}
	if (--info->count < 0) {
		printk("rs_close: bad serial port count for ttys%d: %d\n",
		       info->line, info->count);
		info->count = 0;
	}
	if (info->count) {
		DBG_CNT("before DEC-2");
		MOD_DEC_USE_COUNT;
		restore_flags(flags);
		return;
	}
	info->flags |= ASYNC_CLOSING;
	/*
	 * Save the termios structure, since this port may have
	 * separate termios for callout and dialin.
	 */
	if (info->flags & ASYNC_NORMAL_ACTIVE)
		info->normal_termios = *tty->termios;
	if (info->flags & ASYNC_CALLOUT_ACTIVE)
		info->callout_termios = *tty->termios;
	/*
	 * Now we wait for the transmit buffer to clear; and we notify 
	 * the line discipline to only process XON/XOFF characters.
	 */
	tty->closing = 1;
	if (info->closing_wait != ASYNC_CLOSING_WAIT_NONE)
		tty_wait_until_sent(tty, info->closing_wait);
	/*
	 * At this point we stop accepting input.  To do this, we
	 * disable the receive line status interrupts, and tell the
	 * interrupt driver to stop checking the data ready bit in the
	 * line status register.
	 */
	serial_dsc21_disable_rx_int(info);	

	if (info->flags & ASYNC_INITIALIZED) {
		serial_dsc21_enable_tx_int(info);
		/*
		 * Before we drop DTR, make sure the UART transmitter
		 * has completely drained; this is especially
		 * important if there is a transmit FIFO!
		 */
		timeout = jiffies+HZ;
		while (!serial_dsc21_tx_fifo_empty(info)) {
			current->state = TASK_INTERRUPTIBLE;
			current->timeout = jiffies + info->timeout;
			schedule();
			if (jiffies > timeout)
				break;
		}
	}
	shutdown(info);
	if (tty->driver.flush_buffer)
		tty->driver.flush_buffer(tty);
	if (tty->ldisc.flush_buffer)
		tty->ldisc.flush_buffer(tty);
	tty->closing = 0;
	info->event = 0;
	info->tty = 0;
	if (info->blocked_open) {
		if (info->close_delay) {
			current->state = TASK_INTERRUPTIBLE;
			current->timeout = jiffies + info->close_delay;
			schedule();
		}
		wake_up_interruptible(&info->open_wait);
	}
	info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE|
			 ASYNC_CLOSING);
	wake_up_interruptible(&info->close_wait);
	MOD_DEC_USE_COUNT;
	restore_flags(flags);
}

/*
 * rs_hangup() --- called by tty_hangup() when a hangup is signaled.
 */
void rs_hangup(struct tty_struct *tty)
{
	struct dsc21_async_struct * info = (struct dsc21_async_struct *)tty->driver_data;
	
	if (serial_paranoia_check(info, tty->device, "rs_hangup"))
		return;
	
	rs_flush_buffer(tty);
	shutdown(info);
	info->event = 0;
	info->count = 0;
	info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE);
	info->tty = 0;
	wake_up_interruptible(&info->open_wait);
}


static int block_til_ready(struct tty_struct *tty, struct file * filp,
			   struct dsc21_async_struct *info)
{
	struct wait_queue wait = { current, NULL };
	int		retval;
	int		do_clocal = 0;

	/*
	 * If the device is in the middle of being closed, then block
	 * until it's done, and then try again.
	 */
	if (tty_hung_up_p(filp) ||
	    (info->flags & ASYNC_CLOSING)) {
		if (info->flags & ASYNC_CLOSING)
			interruptible_sleep_on(&info->close_wait);
#ifdef SERIAL_DO_RESTART
		if (info->flags & ASYNC_HUP_NOTIFY)
			return -EAGAIN;
		else
			return -ERESTARTSYS;
#else
		return -EAGAIN;
#endif
	}

	/*
	 * If this is a callout device, then just make sure the normal
	 * device isn't being used.
	 */
	if (tty->driver.subtype == SERIAL_TYPE_CALLOUT) {
		if (info->flags & ASYNC_NORMAL_ACTIVE)
			return -EBUSY;
		if ((info->flags & ASYNC_CALLOUT_ACTIVE) &&
		    (info->flags & ASYNC_SESSION_LOCKOUT) &&
		    (info->session != current->session))
		    return -EBUSY;
		if ((info->flags & ASYNC_CALLOUT_ACTIVE) &&
		    (info->flags & ASYNC_PGRP_LOCKOUT) &&
		    (info->pgrp != current->pgrp))
		    return -EBUSY;
		info->flags |= ASYNC_CALLOUT_ACTIVE;
		return 0;
	}
	
	/*
	 * If non-blocking mode is set, or the port is not enabled,
	 * then make the check up front and then exit.
	 */
	if ((filp->f_flags & O_NONBLOCK) ||
	    (tty->flags & (1 << TTY_IO_ERROR))) {
		if (info->flags & ASYNC_CALLOUT_ACTIVE)
			return -EBUSY;
		info->flags |= ASYNC_NORMAL_ACTIVE;
		return 0;
	}

	if (info->flags & ASYNC_CALLOUT_ACTIVE) {
		if (info->normal_termios.c_cflag & CLOCAL)
			do_clocal = 1;
	} else {
		if (tty->termios->c_cflag & CLOCAL)
			do_clocal = 1;
	}
	
	/*
	 * Block waiting for the carrier detect and the line to become
	 * free (i.e., not in use by the callout).  While we are in
	 * this loop, info->count is dropped by one, so that
	 * rs_close() knows when to free things.  We restore it upon
	 * exit, either normal or abnormal.
	 */
	retval = 0;
	add_wait_queue(&info->open_wait, &wait);
#ifdef SERIAL_DEBUG_OPEN
	printk("block_til_ready before block: ttys%d, count = %d\n",
	       info->line, info->count);
#endif
	{
		unsigned long flags;
		save_flags_cli(flags);
		if (!tty_hung_up_p(filp)) 
			info->count--;
		restore_flags(flags);
	}
	info->blocked_open++;
	while (1) {
		current->state = TASK_INTERRUPTIBLE;
		if (tty_hung_up_p(filp) ||
		    !(info->flags & ASYNC_INITIALIZED)) {
#ifdef SERIAL_DO_RESTART
			if (info->flags & ASYNC_HUP_NOTIFY)
				retval = -EAGAIN;
			else
				retval = -ERESTARTSYS;	
#else
			retval = -EAGAIN;
#endif
			break;
		}
		if (!(info->flags & ASYNC_CALLOUT_ACTIVE) &&
		    !(info->flags & ASYNC_CLOSING))
			break;
		if (current->signal & ~current->blocked) {
			retval = -ERESTARTSYS;
			break;
		}
#ifdef SERIAL_DEBUG_OPEN
		printk("block_til_ready blocking: ttys%d, count = %d\n",
		       info->line, info->count);
#endif
		schedule();
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&info->open_wait, &wait);
	if (!tty_hung_up_p(filp))
		info->count++;
	info->blocked_open--;
#ifdef SERIAL_DEBUG_OPEN
	printk("block_til_ready after blocking: ttys%d, count = %d\n",
	       info->line, info->count);
#endif
	if (retval)
		return retval;
	info->flags |= ASYNC_NORMAL_ACTIVE;
	return 0;
}	

/*
 * This routine is called whenever a serial port is opened.  It
 * enables interrupts for a serial port, linking in its async structure into
 * the IRQ chain.   It also performs the serial-specific
 * initialization for the tty structure.
 */
int rs_open(struct tty_struct *tty, struct file * filp)
{
	struct dsc21_async_struct	*info;
	int 			retval, line;
	unsigned long		page;

	line = MINOR(tty->device) - tty->driver.minor_start;
	if ((line < 0) || (line >= NR_PORTS))
		return -ENODEV;
	info = rs_table + line;
	if (serial_paranoia_check(info, tty->device, "rs_open"))
		return -ENODEV;

	info->count++;
	tty->driver_data = info;
	info->tty = tty;

	if (!tmp_buf) {
		page = get_free_page(GFP_KERNEL);
		if (!page)
			return -ENOMEM;
		if (tmp_buf)
			free_page(page);
		else
			tmp_buf = (unsigned char *) page;
	}
	
	/*
	 * Start up serial port
	 */
	retval = startup(info);
	if (retval)
		return retval;

	MOD_INC_USE_COUNT;
	retval = block_til_ready(tty, filp, info);
	if (retval) {
#ifdef SERIAL_DEBUG_OPEN
		printk("rs_open returning after block_til_ready with %d\n",
		       retval);
#endif
		return retval;
	}

	if ((info->count == 1) && (info->flags & ASYNC_SPLIT_TERMIOS)) {
		if (tty->driver.subtype == SERIAL_TYPE_NORMAL)
			*tty->termios = info->normal_termios;
		else 
			*tty->termios = info->callout_termios;
		change_speed(info);
	}

	info->session = current->session;
	info->pgrp = current->pgrp;

#ifdef SERIAL_DEBUG_OPEN
	printk("rs_open ttys%d successful...", info->line);
#endif
	return 0;
}

/*
 * ---------------------------------------------------------------------
 * rs_init() and friends
 *
 * rs_init() is called at boot-time to initialize the serial driver.
 * ---------------------------------------------------------------------
 */

/*
 * This routine prints out the appropriate serial driver version
 * number, and identifies which options were configured into this
 * driver.
 */
static void show_serial_version(void)
{
 	printk(KERN_INFO "%s version %s with", serial_name, serial_version);
#ifdef CONFIG_SERIAL_PCI
  printk(" PCI");
#define SERIAL_OPT
#endif
#ifdef CONFIG_HUB6
	printk(" HUB-6");
#define SERIAL_OPT
#endif
#ifdef SERIAL_OPT
	printk(" enabled\n");
#else
	printk(" no serial options enabled\n");
#endif
#undef SERIAL_OPT
}

/*
 * This routine is called by rs_init() to initialize a specific serial
 * port.  It determines what type of UART chip this serial port is
 * using: 8250, 16450, 16550, 16550A.  The important question is
 * whether or not this UART is a 16550A or not, since this will
 * determine whether or not we can use its FIFO features or not.
 */
static void autoconfig(struct dsc21_async_struct * info)
{

	unsigned long flags;
	u16 mrs;

	save_flags_cli(flags);
	
	serial_dsc21_disable_ints(info, &mrs);
	serial_dsc21_clear_fifos(info);
	serial_dsc21_disable_breaks(info);
	serial_dsc21_set_tx_trigger(info, UART_FCR_FTL1);
	serial_dsc21_set_rx_trigger(info, UART_FCR_FTL1);

	restore_flags(flags);
}

void init_port(struct dsc21_async_struct *info, int num)
{
  info->magic = SERIAL_MAGIC;
  info->line = num;
  info->tty = 0;

  info->custom_divisor = 0;
  info->close_delay = 5*HZ/10;
  info->closing_wait = 30*HZ;
  info->x_char = 0;
  info->event = 0;
  info->count = 0;
  info->blocked_open = 0;
  info->tqueue.routine = do_softint;
  info->tqueue.data = info;
  info->tqueue_hangup.routine = do_serial_hangup;
  info->tqueue_hangup.data = info;
  info->callout_termios =callout_driver.init_termios;
  info->normal_termios = serial_driver.init_termios;
  info->open_wait = 0;
  info->close_wait = 0;
  info->delta_msr_wait = 0;
  info->icount.cts = info->icount.dsr = 
	info->icount.rng = info->icount.dcd = 0;
  info->next_port = 0;
  info->prev_port = 0;
  
  /*
   * Initialize all the shadow registers to reflect current (bootup) h/w state.
   * This way we won't clobber the settings established for console printk's
   * during boot.
   */
  serial_dsc21_save_registers(info);


  autoconfig(info);

  printk(KERN_INFO "ttyS%02d%s%s at 0x%04x (irq = %d) is a DSC21 UART\n", 
	 info->line, 
	 (info->flags & ASYNC_FOURPORT) ? " FourPort" : "",
	 (info->flags & ASYNC_PCI)      ? " PCI" : "",
	 info->port, info->irq);
  
}

#if 0
int register_serial(struct serial_struct *req);
void unregister_serial(int line);

static struct symbol_table serial_syms = {
#include <linux/symtab_begin.h>
	X(register_serial),
	X(unregister_serial),
#include <linux/symtab_end.h>
};
#endif
/*
 * The serial driver boot-time initialization code!
 */

int serial_dsplinux_init(void)
{
	int i;
	struct dsc21_async_struct * info;
	
	init_bh(SERIAL_BH, do_serial_bh);
	timer_table[RS_TIMER].expires = 0;

	show_serial_version();

	/* Initialize the tty_driver structure */
	
	memset(&serial_driver, 0, sizeof(struct tty_driver));
	serial_driver.magic = TTY_DRIVER_MAGIC;
	serial_driver.name = "ttyS";
	serial_driver.major = TTY_MAJOR;
	serial_driver.minor_start = 64;
	serial_driver.num = NR_PORTS;
	serial_driver.type = TTY_DRIVER_TYPE_SERIAL;
	serial_driver.subtype = SERIAL_TYPE_NORMAL;
	serial_driver.init_termios = tty_std_termios;
	serial_driver.init_termios.c_cflag =
		B115200 | CS8 | CREAD | HUPCL | CLOCAL;
	serial_driver.flags = TTY_DRIVER_REAL_RAW;
	serial_driver.refcount = &serial_refcount;
	serial_driver.table = serial_table;
	serial_driver.termios = serial_termios;
	serial_driver.termios_locked = serial_termios_locked;

	serial_driver.open = rs_open;
	serial_driver.close = rs_close;
	serial_driver.write = rs_write;
	serial_driver.put_char = rs_put_char;
	serial_driver.flush_chars = rs_flush_chars;
	serial_driver.write_room = rs_write_room;
	serial_driver.chars_in_buffer = rs_chars_in_buffer;
	serial_driver.flush_buffer = rs_flush_buffer;
	serial_driver.ioctl = rs_ioctl;
	serial_driver.throttle = rs_throttle;
	serial_driver.unthrottle = rs_unthrottle;
	serial_driver.set_termios = rs_set_termios;
	serial_driver.stop = rs_stop;
	serial_driver.start = rs_start;
	serial_driver.hangup = rs_hangup;

	/*
	 * The callout device is just like normal device except for
	 * major number and the subtype code.
	 */
	callout_driver = serial_driver;
	callout_driver.name = "cua";
	callout_driver.major = TTYAUX_MAJOR;
	callout_driver.subtype = SERIAL_TYPE_CALLOUT;

	if (tty_register_driver(&serial_driver))
		panic("Couldn't register serial driver\n");
	if (tty_register_driver(&callout_driver))
		panic("Couldn't register callout driver\n");
	
	for (i = 0, info = rs_table; i < NR_PORTS; i++,info++) {
				init_port(info, i);
		};

	return 0;
}




#ifdef MODULE
int init_module(void)
{
  // *revisit-gmcnutt* Should the following call really be
  // to serial_dsplinux_console_init() instead? This question
  // added by skranz on Gordon's behalf. May 22, 2001
	return rs_init();
}

void cleanup_module(void) 
{
	unsigned long flags;
	int e1, e2;
	int i;

	/* printk("Unloading %s: version %s\n", serial_name, serial_version); */
	save_flags_cli(flags);
	timer_active &= ~(1 << RS_TIMER);
	timer_table[RS_TIMER].fn = NULL;
	timer_table[RS_TIMER].expires = 0;
	if ((e1 = tty_unregister_driver(&serial_driver)))
		printk("SERIAL: failed to unregister serial driver (%d)\n",
		       e1);
	if ((e2 = tty_unregister_driver(&callout_driver)))
		printk("SERIAL: failed to unregister callout driver (%d)\n", 
		       e2);
	restore_flags(flags);

	for (i = 0; i < NR_PORTS; i++) {
		release_region(rs_table[i].port, 8);
	}


	if (tmp_buf) {
		free_page((unsigned long) tmp_buf);
		tmp_buf = NULL;
	}
}
#endif /* MODULE */
