/*
 *	Industrial Computer Source WDT500/501 driver for Linux 2.1.x
 *
 *      (c) Copyright 2001 RidgeRun, Inc <www.ridgerun.com>
 *	(c) Copyright 1996-1997 Alan Cox <alan@redhat.com>, All Rights Reserved.
 *				http://www.redhat.com
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *	
 *	Neither Alan Cox nor CymruNet Ltd. admit liability nor provide 
 *	warranty for any of this software. This material is provided 
 *	"AS-IS" and at no charge.	
 *
 *	(c) Copyright 1995    Alan Cox <alan@lxorguk.ukuu.org.uk>
 *
 *	Release 0.08.
 *
 *	Fixes
 *		Dave Gregorich	:	Modularisation and minor bugs
 *		Alan Cox	:	Added the watchdog ioctl() stuff
 *		Alan Cox	:	Fixed the reboot problem (as noted by
 *					Matt Crocker).
 *		Alan Cox	:	Added wdt= boot option
 *		Alan Cox	:	Cleaned up copy/user stuff
 *		Tim Hockin	:	Added insmod parameters, comment cleanup
 *					Parameterized timeout
 *		Tigran Aivazian	:	Restructured wdt_init() to handle failures
 *              Gordon McNutt   :       Leveraged for TI watchdog timers
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <asm/arch/sysdep.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/malloc.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <asm/io.h>
#include <asm/system.h>
#include <linux/notifier.h>
#ifndef LINUX_20
#include <asm/uaccess.h>
#include <linux/reboot.h>
#include <linux/init.h>
#endif

//#include "../timer/timer.h"

#ifdef DEBUGVM
#include "vm_debug_stub.h"
#endif
 
#define WATCHDOG_MAJOR 240  // First one out of the 240-254 Local/Exp. block

#ifdef CONFIG_ARCH_DSC2X
#define WDT_TIMER 4
#define MAX_WDT_USEC 0x25e368
#elif CONFIG_ARCH_ORION
#define WDT_TIMER 0
#define MAX_WDT_USEC 353200
#endif
#define RESET_MODE 0
#define INTERRUPT_MODE 1

extern void timer_unregister_cb(int timer_val, void (*handler)(int, void *, struct pt_regs *));
extern int timer_register_cb(int timer_val, const char *name_string, void (*handler)(int, void *, struct pt_regs *));
extern int timer_set_watchdog(int timer_val, u32 usec_value);
extern int timer_set_usec(int timer_val, u32 usec_value);

static int wdt_is_open;
//static int timer;

#define WDT_DEBUG 

#ifdef WDT_DEBUG
#define DBG(format, arg...) printk(KERN_DEBUG __FUNCTION__ ": " format "\n" , ##arg)
#define static
#else
#define DBG(x...)
#endif

#ifndef MODULE

/**
 *	wdt_setup:
 *	@str: command line string
 *
 *	Setup options. N/A.
 */
 
static int __init wdt_setup(char *str)
{
	return 1;
}

__setup("wdt=", wdt_setup);

#endif /* !MODULE */

/**
 *	wdt_interrupt:
 *	@timer:		Timer number
 *	@dev_id:	Unused as we don't allow multiple devices.
 *
 *	Handle an interrupt from the timer. This only happens when the timer
 *      is configured in 'interrupt' instead of 'reset' mode (which is never
 *      by this driver). I include this here as an example of how we could
 *      do something else besides h/w reset (Dan mentioned some realtime
 *      applications that might want to shutdown equipment or something 
 *      instead of just blindly resetting the MPU).
 */
 
void wdt_cb(int timer, void *dev_id, struct pt_regs *regs)
{
	printk(KERN_CRIT "WDT expired!\n");
	
#ifdef SOFTWARE_REBOOT
#ifdef ONLY_TESTING
		printk(KERN_CRIT "Would Reboot.\n");
#else		
		printk(KERN_CRIT "Initiating system reboot.\n");
		machine_restart(NULL);
#endif		
#else
		printk(KERN_CRIT "But since the WDT is configued in interrupt "\
                       "mode -- without software reboot -- nothing will happen\n");
#endif		
}


#ifdef LINUX_20
static int wdt_llseek(struct inode *inode, struct file *file, off_t offset, int org)
#else
static long long wdt_llseek(struct file *file, long long offset, int org)
#endif
{
	return -ESPIPE;
}

/*
 *	wdt_ping:
 *
 *	Reset the counter. 
 */

static void wdt_ping(void)
{
    DBG("wdt_ping");
    DBG("timer_set_usec addr is %p", &timer_set_usec);

    timer_set_usec(WDT_TIMER, MAX_WDT_USEC); /* Max usec seconds possible */
}


/**
 *	wdt_write:
 *	@file: file handle to the watchdog
 *	@buf: buffer to write (unused as data does not matter here 
 *	@count: count of bytes
 *	@ppos: pointer to the position to write. No seeks allowed
 *
 *	A write to a watchdog device is defined as a keepalive signal. Any
 *	write of data will do, as we we don't define content meaning.
 */
 
#ifdef LINUX_20
static int wdt_write(struct inode *inode, struct file *file, const char *buf, int count)
{
#else
static ssize_t wdt_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	/*  Can't seek (pwrite) on this device  */
	if (ppos != &file->f_pos)
		return -ESPIPE;
#endif

    DBG("Write call");
	if(count)
	{
		wdt_ping();
		return 1;
	}
	return 0;
}

/**
 *	wdt_ioctl:
 *	@inode: inode of the device
 *	@file: file handle to the device
 *	@cmd: watchdog command
 *	@arg: argument pointer
 *
 *	The watchdog API defines a common set of functions for all watchdogs
 *	according to their available features. We only actually usefully support
 *	querying capabilities and current status. 
 */
 
static int wdt_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	unsigned long arg)
{

	static struct watchdog_info ident=
	{
        options: 0,
		firmware_version: 0,
		identity: "C5471 WDT"
	};
	
    DBG("ioctl Call: inode=%p file=%p cmd=%x arg=%lx",inode,file,cmd,arg);

    /* Range check the IOCTL cmd */
    if ((_IOC_TYPE(cmd) != WATCHDOG_IOCTL_BASE) || (_IOC_NR(cmd) > WDIOC_IOC_MAXNR)) 
    {
        DBG("Bad ioctl value");
        return -EINVAL;
    }

	switch(cmd)
	{
        default:
                DBG("Error Default");
                return -ENOIOCTLCMD;
        case WDIOC_GETSUPPORT:
                DBG("WDIOC_GETSUPPORT");
                return copy_to_user((struct watchdog_info *)arg, &ident, 
                                    sizeof(ident))?-EFAULT:0;
                
        case WDIOC_GETSTATUS:
                DBG("WDIOC_GETSTATUS");
                return PUT_USER(0, (int *)arg);
        case WDIOC_GETBOOTSTATUS:
                DBG("WDIOC_GETBOOTSTATUS");
                return PUT_USER(0, (int *)arg);
        case WDIOC_KEEPALIVE:
                DBG("WDIOC_KEEPALIVE");
                wdt_ping();
                return 0;
        }
}

/**
 *	wdt_open:
 *	@inode: inode of device
 *	@file: file handle to device
 *
 *	Our misc device has been opened. It's single open and on open we load
 *      the counter and start the countdown.
 */
 
static int wdt_open(struct inode *inode, struct file *file)
{
    int ret;

    DBG("Open Module");

    if(wdt_is_open)
        return -EBUSY;
    /*
     *	Activate 
     */

    wdt_is_open=1;
    /* This will automatically load the timer with its max
       count and start it running.
     */
    if ((ret = timer_set_watchdog(WDT_TIMER, RESET_MODE)))
    {
        wdt_is_open = 0;
        return ret;
    }
    return 0;
}

/**
 *	wdt_close:
 *	@inode: inode to board
 *	@file: file handle to board
 *
 *	The watchdog has a configurable API. There is a religious dispute 
 *	between people who want their watchdog to be able to shut down and 
 *	those who want to be sure if the watchdog manager dies the machine
 *	reboots. In the former case we disable the counters, in the latter
 *	case you have to open it again very soon.
 */
 
#ifdef LINUX_20
static void wdt_release(struct inode *inode, struct file *file)
#else
static int wdt_release(struct inode *inode, struct file *file)
#endif
{
    DBG("Release Module");

//  lock_kernel();
#ifndef CONFIG_WATCHDOG_NOWAYOUT	
    timer_set_watchdog(WDT_TIMER, INTERRUPT_MODE); /* reset off */
#endif		
    wdt_is_open=0;
//  unlock_kernel();
#ifndef LINUX_20
	return 0;
#endif
}

/*
 *	Kernel Interfaces
 */
 
 
static struct file_operations wdt_fops = {
	TAG_LLSEEK:	wdt_llseek,
	write:		wdt_write,
	ioctl:		wdt_ioctl,
	open:		wdt_open,
	release:	wdt_release,
};


/**
 *	cleanup_module:
 *
 *	Unload the watchdog. You cannot do this with any file handles open.
 *	If your watchdog is set to continue ticking on close and you unload
 *	it, well it keeps ticking. We won't get the interrupt but the board
 *	will not touch PC memory so all is fine. You just have to load a new
 *	module in 19 seconds or reboot.
 */
 
#ifdef LINUX_20
void cleanup_module(void)
#else
void __exit cleanup_module(void)
#endif
{
    DBG("Cleanup Module");

	unregister_chrdev(WATCHDOG_MAJOR, "watchdog");
	timer_unregister_cb(WDT_TIMER, &wdt_cb);
}

/**
 * 	wdt_init:
 *
 *	Set up the WDT watchdog board. All we have to do is grab the
 *	resources we require and bitch if anyone beat us to them.
 *	The open() function will actually kick the board off.
 *
 */
 
#ifdef LINUX_20
int init_module(void)
#else
int __init init_module(void)
#endif
{
	int ret;

    DBG("Init Module");

	ret = register_chrdev(WATCHDOG_MAJOR, "watchdog", &wdt_fops);
	if (ret) {
		printk(KERN_ERR "wdt: can't register on major=%d ret=%d\n", 
                       WATCHDOG_MAJOR,ret);
		goto out;
	}
	ret = timer_register_cb(WDT_TIMER, "watchdog", &wdt_cb);
	if(ret) {
		printk(KERN_ERR "wdt: watchdog timer (%d) is not free.\n", WDT_TIMER);
		goto outmisc;
	}

	ret = 0;
	printk(KERN_INFO "DSPLinux Watchdog Driver (c) 2001\n");
out:
	return ret;
outmisc:
	unregister_chrdev(WATCHDOG_MAJOR, "watchdog");
	goto out;
}

//module_init(wdt_init);
//module_exit(wdt_exit);

