/*
 * File:   uclinux/include/asm-armnommu/arch-dsplinux/dma.h
 *
 * Author: Steve Kranz <skranz@ridgerun.com>
 *
 * Date:   May 2001
 *
 * Description:
 *   This file was generated as part of getting the CompactFlash
 *   slot of the TI TMS320DSC21 platform working. This file was necessary 
 *   to provide the DMA compabilities used to transfer bytes between
 *   the system's SDRAM and the CARD slot (for both the CompactFlash
 *   slot or the SmartMedia slot). This implemention is board specific
 *   and implements many of the function prototypes declared and made
 *   available to clients in the general header:
 *   uclinux/include/asm-armnommu/dma.h
 */

#ifndef __ASM_ARCH_DMA_H
#define __ASM_ARCH_DMA_H

#define MAX_DMA_CHANNELS 1

typedef enum {
  DMA_MODE_CARD_TO_SDRAM,
  DMA_MODE_SDRAM_TO_CARD,
  DMA_MODE_FLASH_TO_SDRAM,
  DMA_MODE_SDRAM_TO_FLASH
} dmamode_t;

#ifdef CONFIG_ARCH_DSC21

#include <asm-armnommu/io.h>

#define MAX_DMA_ADDRESS 0x9fffffff /* SDRAM range: 0x0800.0000-0x9fff.ffff */

// #define DMA_VIRTUAL_FLOPPY 0

// #define DMA_FLOPPY DMA_VIRTUAL_FLOPPY

// DSC21 Memory Map Specifics
#define SDRAM_BASE_ADDR 0x08000000
#define CARD_BASE_ADDR  0x04000000
#define FLASH_BASE_ADDR 0x02000000
#define DMA_BASE_REG    0x00030B80
#define DMA_CTL_REG     DMA_BASE_REG+0x00
#define DMA_MODE_REG    DMA_BASE_REG+0x02
#define DMA_TRSSZ_REG   DMA_BASE_REG+0x04
#define DMA_EXB_AD0_REG DMA_BASE_REG+0x06
#define DMA_EXB_AD1_REG DMA_BASE_REG+0x08
#define DMA_STAD0_REG   DMA_BASE_REG+0x0a
#define DMA_STAD1_REG   DMA_BASE_REG+0x0c
#define DMA_STATUS_REG  DMA_BASE_REG+0x0e

/* enable/disable a specific DMA channel */
static __inline__ void enable_dma(unsigned int channel)
{
  outw(0x0001,DMA_CTL_REG);
}

static __inline__ void disable_dma(unsigned int dmanr)
{
  outw(0x0002,DMA_CTL_REG);
}

/* set mode for a specific DMA channel */
static __inline__ void set_dma_mode(unsigned int dmanr, dmamode_t mode)
{
  switch (mode) {
  case DMA_MODE_CARD_TO_SDRAM:
    outw(0x0000,DMA_MODE_REG);
    break;
  case DMA_MODE_SDRAM_TO_CARD:
    outw(0x0001,DMA_MODE_REG);
    break;
  case DMA_MODE_FLASH_TO_SDRAM:
    outw(0x0002,DMA_MODE_REG);
    break;
  case DMA_MODE_SDRAM_TO_FLASH:
    outw(0x0003,DMA_MODE_REG);
    break;
  }
}

/* Set SDRAM transfer address for specific DMA channel.
 */
static __inline__ void set_dma_addr(unsigned int dmanr, unsigned long a)
{
  unsigned int offset;
  unsigned short lo;
  unsigned short hi;

  // Note: The DSC21 DMA chipset doesn't allow us to simply program
  // in the physical byte address of the SDRAM -- instead we need to
  // program in the "offset from start of SDRAM".
  if (SDRAM_BASE_ADDR > a) {
    printk("ERROR; set_dma_addr: invalid address specified\n");
  }
  offset = (a-SDRAM_BASE_ADDR); // compute offset.
  
  // Now, Program the chipset.
  lo = (offset&0xFFFF);       // low 16 bits.
  hi = ((offset>>16)&0x07FF); // hi 11 bits.
  outw(lo,DMA_STAD0_REG);
  outw(hi,DMA_STAD1_REG);
}

/* Set Device transfer address for specific DMA channel.
   Note1: This address is incremented during the DMA process
          only if the transfer is between SDRAM and FLASH.
          If between SDRAM and CARD then the device address
          remains fixed during the xfer.
   Note2: the mode parameter was required here to make it possible
          to transform the incoming physical bytes address "a" into
          an internally required address offset suitable for the
          DSC21 chipset.
*/
static __inline__ void set_dma_device_addr(unsigned int dmanr, unsigned long a, dmamode_t mode)
{
  unsigned int offset;
  unsigned short hi,lo;
  unsigned long BaseAddr;

  // First, compute offset from incoming physical byte address "a".
  switch (mode) {
  case DMA_MODE_CARD_TO_SDRAM:
    BaseAddr = CARD_BASE_ADDR;
    break;
  case DMA_MODE_SDRAM_TO_CARD:
    BaseAddr = CARD_BASE_ADDR;
    break;
  case DMA_MODE_FLASH_TO_SDRAM:
    BaseAddr = FLASH_BASE_ADDR;
    break;
  case DMA_MODE_SDRAM_TO_FLASH:
    BaseAddr = FLASH_BASE_ADDR;
    break;
  }
  if (BaseAddr > a) {
    printk("ERROR; set_dma_device_addr: invalid address specified\n");
  }
  offset = (a-BaseAddr);

  // Now, Program the chipset.
  lo = (offset&0xFFFF);       // low 16 bits.
  hi = ((offset>>16)&0x007F); // hi 7 bits.
  outw(lo,DMA_EXB_AD0_REG);
  outw(hi,DMA_EXB_AD1_REG);
}

/* Set transfer size (bytes) for a specific DMA channel.
 * valid values 1-0x7ff.
 */
static __inline__ void set_dma_count(unsigned int dmanr, unsigned long count)
{
  outw(count,DMA_TRSSZ_REG);
}

/* Get DMA residue count. After a DMA transfer, this
 * should return zero. Reading this while a DMA transfer is
 * still in progress will return unpredictable results.
 * If called before the channel has been used, it may return 1.
 * Otherwise, it returns the number of _bytes_ left to transfer.
 */
static __inline__ int get_dma_residue(unsigned int dmanr)
{
  return inw(DMA_STATUS_REG)&0x07ff;
}

/*
 * Returns to caller when current DMA operation completes.
 * Returns immediately if no current operation in-progress.
*/
static __inline__ void wait_for_dma_completion(unsigned int dmanr)
{
  while (inw(DMA_STATUS_REG)&0x8000) {}
}

#endif /* CONFIG_ARCH_DSC21 */

#endif /* __ASM_ARCH_DMA_H */
