#include <linux/string.h>	/* memset, memcpy */
#include <asm/system.h>		/* cli, sti       */
#include <linux/buffer.h>	/* private stuff  */

/*****************************************************************/
/*                                                               */
/*****************************************************************/
BUF freeImgBuf;     // Free frame buffer
BUF rawImgBuf;      // All ram image from CMOS lens will be queued here.

// Re-entry protection: MaskIntr() and UnmaskIntr() are not required
// if only interrupts call these services.
/* I want to make the code simpler, so I remark them. (The caller needs to
 * store the FLAG if I use the following declarations.)
 */
#if 0
#define MaskIntr(flags)	do { save_flags(flags);	cli(); } while(0)
#define UnmaskIntr(unsigned long flags)	do { restore_flags(flags); } while(0)
#endif /* 0 */           

static PFRAME GetFrame(PBUF pBuf);
static void   PutFrame(PBUF pBuf, PFRAME pFrame);

#define _FIX_REENTRY_
#ifdef _FIX_REENTRY_
static int FreeBufCount=0;	/* vincent: re-entry */
static int RawBufCount=0;	/* vincent: re-entry */
#endif /* _FIX_REENTRY_ */

/*****************************************************************/
/*                                                               */
/*****************************************************************/
/*
 *  +---------+  +---------+  +---------+
 *  | 1       |->| 2       |->| 3       |  (FreeBuffer/RawBuffer)
 *  +---------+  +---------+  +---------+
 *  ^            ^            ^
 *  start (next--+)           |
 *               (next--------+)
 *                            (next-->NULL)
 *                            end
 * Flow:
 * 1. CMOS module:
 *    if (GetFreeFrame() != NULL) { (could get new free)
 *       PutRawFrame();
 *    } else {
 *       Use Current FreeFrameBuffer!
 *       Do NOT "PutRawFrame"
 *    }
 *    MAX Keep "FreeFrame" = 3 (all), but MAX Put "RawFrame" = 2.
 *    Must keep one working buffer!!
 * 2. DSP modules:
 *    if ((GetRawFrame() != NULL)&&(GetDSPBuffer() != NULL)) {
 *       Deal with the Input Video (RAW->Mpeg4)
 *       ....WEB Get Data??
 *       PutFreeFrame();
 *    } else {
 *       (Keep the RawFrame unitl GetDSPBuffer) || 
 *       (Keep the DSPBuffer until GetRawFrame)
 *    }
 *    Only Keep 1 RawFrame at a time!!
 */
asmlinkage int sys_InitBuf(void)
{
	int i;
	unsigned long flags;

  /*---------------------------------------------*/
	save_flags(flags);     // Mask Interrupt
	cli();

  /*---------------------------------------------*/
	// Init. frame buffer
	memset(&freeImgBuf, '\0', sizeof(freeImgBuf));
	memset(&rawImgBuf,  '\0', sizeof(rawImgBuf));

	i = IMG_FRAME_NUM;
	freeImgBuf.pStart = freeImgBuf.pEnd = (PFRAME)IMG_BUF_START;
	freeImgBuf.nTotal = 1;	/* 1st buffer HERE! */

	while (i > 1) {		/* N buffers, do (N-1) shift (not N) */
		freeImgBuf.pEnd->pNext = (PFRAME)((unsigned long)freeImgBuf.pEnd + (unsigned long)IMG_FRAME_SIZE);
		freeImgBuf.pEnd = freeImgBuf.pEnd->pNext;
		freeImgBuf.nTotal++;
		i--;
	}
	freeImgBuf.pEnd->pNext= NULL; /* last buffer */

  /*---------------------------------------------*/
	// Init. raw image buffer
	rawImgBuf.pStart = rawImgBuf.pEnd = NULL;
	rawImgBuf.nTotal = 0;

#ifdef _FIX_REENTRY_
	FreeBufCount = RawBufCount = 0;
#endif /* _FIX_REENTRY_ */

  /*---------------------------------------------*/
	restore_flags(flags);     // UnMask Interrupt

  /*---------------------------------------------*/
	return 0;
}

/*****************************************************************/
/*                                                               */
/*****************************************************************/
asmlinkage PFRAME sys_GetFreeFrame(void)
{
#ifdef _FIX_REENTRY_
	unsigned long flags;
	PFRAME pFreeImgBuf;

	if (FreeBufCount != 0) {
		printk("GetFreeFrame[%d]\n", FreeBufCount);
	}
  /*---------------------------------------------*/
	save_flags(flags);     // Mask Interrupt
	cli();

	FreeBufCount++;
	pFreeImgBuf = GetFrame(&freeImgBuf);
	FreeBufCount--;

  /*---------------------------------------------*/
	restore_flags(flags);     // UnMask Interrupt

	return pFreeImgBuf;
#else
	return(GetFrame(&freeImgBuf));
#endif /* _FIX_REENTRY_ */
}

/*****************************************************************/
/*                                                               */
/*****************************************************************/
asmlinkage int sys_PutFreeFrame(PFRAME pFrame)
{
#ifdef _FIX_REENTRY_
	unsigned long flags;

	if (FreeBufCount != 0) {
		printk("PutFreeFrame[%d]-0x%08lX\n", FreeBufCount, pFrame);
	}
  /*---------------------------------------------*/
	save_flags(flags);     // Mask Interrupt
	cli();

	FreeBufCount++;
	PutFrame(&freeImgBuf, pFrame);
	FreeBufCount--;

  /*---------------------------------------------*/
	restore_flags(flags);     // UnMask Interrupt

	return 0;
#else
	PutFrame(&freeImgBuf, pFrame);
	return 0;
#endif /* _FIX_REENTRY_ */
}

/*****************************************************************/
/*                                                               */
/*****************************************************************/
asmlinkage PFRAME sys_GetRawFrame(void)
{
#ifdef _FIX_REENTRY_
	unsigned long flags;
	PFRAME pRawImgBuf;

	if (RawBufCount != 0) {
		printk("GetRawFrame[%d]\n", RawBufCount);
	}
  /*---------------------------------------------*/
	save_flags(flags);     // Mask Interrupt
	cli();

	RawBufCount++;
	pRawImgBuf = GetFrame(&rawImgBuf);
	RawBufCount--;

  /*---------------------------------------------*/
	restore_flags(flags);     // UnMask Interrupt

	return pRawImgBuf;
#else
	return(GetFrame(&rawImgBuf));	
#endif /* _FIX_REENTRY_ */
}

/*****************************************************************/
/*                                                               */
/*****************************************************************/
asmlinkage int sys_PutRawFrame(PFRAME pFrame)
{
#ifdef _FIX_REENTRY_
	unsigned long flags;

	if (RawBufCount != 0) {
		printk("PutRawFrame[%d]-0x%08lX\n", RawBufCount, pFrame);
	}
  /*---------------------------------------------*/
	save_flags(flags);     // Mask Interrupt
	cli();

	RawBufCount++;		/* vincent: debug */
	PutFrame(&rawImgBuf, pFrame);
	RawBufCount--;

  /*---------------------------------------------*/
	restore_flags(flags);     // UnMask Interrupt

	return 0;
#else
	PutFrame(&rawImgBuf, pFrame);
	return 0;
#endif /* _FIX_REENTRY_ */
}

/*****************************************************************/
/*                                                               */
/*****************************************************************/
static asmlinkage PFRAME GetFrame(PBUF pBuf)
{
	PFRAME pTmp;
////	unsigned long flags;

  /*---------------------------------------------*/
	if (pBuf->nTotal == 0) {
		return NULL;
	}

  /*---------------------------------------------*/
////	save_flags(flags);     // Mask Interrupt
////	cli();

  /*---------------------------------------------*/
	pTmp = pBuf->pStart;
	pBuf->nTotal--;			/* decrease 1 */
	if (pBuf->nTotal == 0) {	/* no more buffer */
		pBuf->pStart = NULL;
		pBuf->pEnd = NULL;
		/* NULL->pNext = !@#$%^&, So Give Up pNext! */
	} else {			/* change to NEXT */
		pBuf->pStart = pBuf->pStart->pNext;
	}

  /*---------------------------------------------*/
////	restore_flags(flags);     // UnMask Interrupt

  /*---------------------------------------------*/
	return pTmp;
}

/*****************************************************************/
/*                                                               */
/*****************************************************************/
static asmlinkage void PutFrame(PBUF pBuf, PFRAME pFrame)
{
////	unsigned long flags;

  /*---------------------------------------------*/
////	save_flags(flags);     // Mask Interrupt
////	cli();

  /*---------------------------------------------*/
	if (pBuf->nTotal == 0) {	/* chain current buffer */
		pBuf->pStart = pBuf->pEnd = pFrame;
		pBuf->pStart->pNext = NULL;
		pBuf->pEnd->pNext   = NULL;
	} else {
		pBuf->pEnd->pNext= pFrame;
		pBuf->pEnd= pFrame;
		pBuf->pEnd->pNext= NULL;   // pFrame->pNext= NULL;
	}
	pBuf->nTotal++;

  /*---------------------------------------------*/
////	restore_flags(flags);     // UnMask Interrupt

  /*---------------------------------------------*/
	return;
}
